@@ -1,4 +1,4 @@
-// Copyright (c) 1997, 1998, 1999, 2001, 2002  Per M.A. Bothner.
+// Copyright (c) 1997, 1998, 1999, 2001, 2002, 2004, 2005  Per M.A. Bothner.
 // This is free software;  for terms and warranty disclaimer see ./COPYING.
 
 package gnu.bytecode;
@@ -8,8 +8,25 @@ import java.util.Vector;
 public class ClassType extends ObjectType 
   implements AttrContainer, Externalizable
 {
-  public static final int minor_version = 3;
-  public static final int major_version = 45;
+  // An old but generally valid default value.
+  int classfileFormatVersion = 45 * 0x10000 + 3;
+
+  public short getClassfileMajorVersion ()
+  {
+    return (short) (classfileFormatVersion >> 16);
+  }
+  public short getClassfileMinorVersion ()
+  {
+    return (short) (classfileFormatVersion & 0xFFFF);
+  }
+  public void setClassfileVersion (int major, int minor)
+  {
+    classfileFormatVersion = (major & 0xFFFF) * 0x10000 + (minor * 0xFFFF);
+  }
+  public void setClassfileVersionJava5 ()
+  {
+    setClassfileVersion(49, 0);
+  }
 
   /** Find a ClassType with the given name, or create a new one.
    * Use this for "library classes", where you need the field/method types,
@@ -66,7 +83,8 @@ public class ClassType extends ObjectType
   /** Return the modifiers (access flags) for this class. */
   public final int getModifiers()
   {
-    if (access_flags == 0 && getReflectClass() != null)
+    if (access_flags == 0
+	&& (flags & EXISTING_CLASS) != 0 && getReflectClass() != null)
       access_flags = reflectClass.getModifiers();
     return access_flags;
   }
@@ -74,19 +92,117 @@ public class ClassType extends ObjectType
   /** Set the modifiers (access flags) for this class. */
   public final void setModifiers(int flags) { access_flags = flags; }
 
+  public final boolean hasOuterLink ()
+  {
+    getFields();
+    return (flags & HAS_OUTER_LINK) != 0;
+  }
+
+  public ClassType getOuterLinkType ()
+  {
+    return ! hasOuterLink() ? null
+      : (ClassType) getDeclaredField("this$0").getType();
+  }
+
+  /** Note that this class needs an other link ("this$0") field.
+   * This is only allowed if !isExisting().
+   * Adjust any existing "<init>" methods to take the extra
+   * implicit parameter.
+   * @param outer the outer class
+   */
+  public final Field setOuterLink (ClassType outer)
+  {
+    if ((flags & EXISTING_CLASS) != 0)
+      throw new Error("setOuterLink called for existing class "+getName());
+    Field field = getDeclaredField("this$0");
+    if (field == null)
+      {
+        field = addField("this$0", outer);
+        flags |= HAS_OUTER_LINK;
+        for (Method meth = methods; meth != null;  meth = meth.getNext())
+          {
+            if ("<init>".equals(meth.getName()))
+              {
+                if (meth.code != null)
+                  throw new Error("setOuterLink called when "+meth+" has code");
+                Type[] arg_types = meth.arg_types;
+                Type[] new_types = new Type[arg_types.length+1];
+                System.arraycopy(arg_types, 0, new_types, 1, arg_types.length);
+                new_types[0] = outer;
+                meth.arg_types = new_types;
+                meth.signature = null;
+              }
+          }
+      }
+    else if (! outer.equals(field.getType()))
+      throw new Error("inconsistent setOuterLink call for "+getName());
+    return field;
+  }
+
+  /** Check if a component is accessible from this class.
+   * @param declaring the class containing the component (a field, method,
+   *   or inner class)
+   * @param modifiers the access flags of the component
+   * @return true if the specified component can be accessed from this class.
+   */
+  public boolean isAccessible (ClassType declaring, int modifiers)
+  {
+    int cmods = declaring.getModifiers();
+    if ((modifiers & Access.PUBLIC) != 0 && (cmods & Access.PUBLIC) != 0)
+      return true;
+    String callerName = getName();
+    String className = declaring.getName();
+    if (callerName.equals(className))
+      return true;
+    if ((modifiers & Access.PRIVATE) != 0)
+      return false;
+    int dot = callerName.lastIndexOf('.');
+    String callerPackage = dot >= 0 ? callerName.substring(0, dot) : "";
+    dot = className.lastIndexOf('.');
+    String classPackage = dot >= 0 ? className.substring(0, dot) : "";
+    if (callerPackage.equals(classPackage))
+      return true;
+    if ((modifiers & Access.PROTECTED) != 0
+        && this.isSubclass(declaring))
+      return true;
+    return false;
+  }
+
   /** Sets the name of the class being defined in this classfile.
    * @param name the name to give to the class
    */
   public void setName (String name)
   {
     this_name = name;
-    name = name.replace ('.', '/');
-    setSignature("L"+name+";");
+    setSignature("L"+name.replace('.', '/')+";");
+  }
+
+  SourceDebugExtAttr sourceDbgExt;
+
+  /** Create a <code>SourceDebugExtAttr</code>, if needed, and
+   * set the "stratum".  The stratum is typically a programming language
+   * such as "JSP", "Scheme", or "Java" (the default). */
+  public void setStratum (String stratum)
+  {
+    if (sourceDbgExt == null)
+      sourceDbgExt = new SourceDebugExtAttr(this);
+    sourceDbgExt.addStratum(stratum);
   }
 
   /** Set the name of the SourceFile associated with this class. */
   public void setSourceFile (String name)
   {
+    if (sourceDbgExt != null)
+      {
+	sourceDbgExt.addFile(name);
+	if (sourceDbgExt.fileCount > 1)
+	  return;
+      }
+
+    name = SourceFileAttr.fixSourceFile(name);
+    int slash = name.lastIndexOf('/');
+    if (slash >= 0)
+      name = name.substring(slash+1);
     SourceFileAttr.setSourceFile(this, name);
   }
 
@@ -109,7 +225,7 @@ public class ClassType extends ObjectType
     if (superClass == null
 	&& ! isInterface()
 	&& ! ("java.lang.Object".equals(getName()))
- 	&& getReflectClass() != null)
+	&& (flags & EXISTING_CLASS) != 0 && getReflectClass() != null)
       {
 	superClass = (ClassType) make(reflectClass.getSuperclass());
       }
@@ -119,17 +235,18 @@ public class ClassType extends ObjectType
   public String getPackageName()
   {
     String name = getName();
-    int index = name.indexOf('.');
-    return index < 0 ? name : name.substring(0, index);
+    int index = name.lastIndexOf('.');
+    return index < 0 ? "" : name.substring(0, index);
   }
 
   /**
    * @return the interfaces this class is declared to implement
    * (not those inherited from its superclass/superinterfaces).
    */
-  public ClassType[] getInterfaces()
+  public synchronized ClassType[] getInterfaces()
   {
-    if (interfaces == null && getReflectClass() != null)
+    if (interfaces == null
+	&& (flags & EXISTING_CLASS) != 0 && getReflectClass() != null)
       {
 	Class[] reflectInterfaces = reflectClass.getInterfaces();
 	int numInterfaces = reflectInterfaces.length;
@@ -172,8 +289,8 @@ public class ClassType extends ObjectType
 
   public final void setInterface(boolean val)
   {
-    if (val) access_flags |= Access.INTERFACE;
-    else access_flags &= ~Access.INTERFACE;
+    if (val) access_flags |= Access.INTERFACE|Access.ABSTRACT;
+    else access_flags &= ~Access.INTERFACE|Access.ABSTRACT;
   }
 
   public ClassType () { }
@@ -200,7 +317,7 @@ public class ClassType extends ObjectType
   int LineNumberTable_name_index;
 
   /** Get the fields of this class. */
-  public final Field getFields()
+  public final synchronized Field getFields()
   {
     if ((flags & (ADD_FIELDS_DONE|EXISTING_CLASS)) == EXISTING_CLASS)
       addFields();
@@ -217,10 +334,7 @@ public class ClassType extends ObjectType
    */
   public Field getDeclaredField(String name)
   {
-    if ((flags & (ADD_FIELDS_DONE|EXISTING_CLASS)) == EXISTING_CLASS)
-      addFields();
-    Field field;
-    for (field = fields;   field != null;  field = field.next)
+    for (Field field = getFields();   field != null;  field = field.next)
       {
 	if (name.equals(field.name))
 	  return field;
@@ -229,22 +343,44 @@ public class ClassType extends ObjectType
   }
 
   /** Find a field with the given name declared in this class or its ancestors.
+   * @param name the name of the field.
+   * @param mask of match a field whose modifiers has one of these bits set.
+   *   Howeve, if mask is -1, ignore the access flags.
    * @return the matching field, or null if there is no such field.
    */
-  public Field getField(String name)
+  public Field getField(String name, int mask)
   {
     ClassType cl = this;
     for (;;)
       {
         Field field = cl.getDeclaredField(name);
-        if (field != null)
+        if (field != null
+            && (mask == -1 || (field.getModifiers() & mask) != 0))
           return field;
+        ClassType[] interfaces = cl.getInterfaces();
+        if (interfaces != null)
+          {
+            for (int i = 0;  i < interfaces.length;  i++)
+              {
+                field = interfaces[i].getField(name, mask);
+                if (field != null)
+                  return field;
+              }
+          }
         cl = cl.getSuperclass();
         if (cl == null)
           return null;
       }
   }
 
+  /** Find a field with the given name declared in this class or its ancestors.
+   * @return the matching field, or null if there is no such field.
+   */
+  public Field getField(String name)
+  {
+    return getField(name, Access.PUBLIC);
+  }
+
   /**
    * Add a new field to this class.
    */
@@ -276,7 +412,8 @@ public class ClassType extends ObjectType
 
   /** Use reflection to add all the declared fields of this class.
    * Does not add private or package-private fields.
-   * Does not check for duplicate (already-known) fields. */
+   * Does not check for duplicate (already-known) fields.
+   * Is not thread-safe if another thread may access this ClassType. */
   public void addFields()
   {
     Class clas = getReflectClass();
@@ -293,12 +430,10 @@ public class ClassType extends ObjectType
     for (int i = 0;  i < count;  i++)
       {
         java.lang.reflect.Field field = fields[i];
-        if (! field.getDeclaringClass().equals(clas))
-          continue;
-        int modifiers = field.getModifiers();
-        if ((modifiers & (Access.PUBLIC|Access.PROTECTED)) == 0)
-          continue;
-        addField(field.getName(), Type.make(field.getType()), modifiers);
+        if ("this$0".equals(field.getName()))
+          flags |= HAS_OUTER_LINK;
+        addField(field.getName(), Type.make(field.getType()),
+                 field.getModifiers());
       }
     flags |= ADD_FIELDS_DONE;
   }
@@ -323,9 +458,7 @@ public class ClassType extends ObjectType
   }
 
   public Method addMethod (String name) {
-    Method method = new Method (this, 0);
-    method.setName(name);
-    return method;
+    return addMethod(name, 0);
   }
 
   public Method addMethod (String name, int flags) {
@@ -353,8 +486,7 @@ public class ClassType extends ObjectType
         && return_type.equals(method.getReturnType())
         && (flags & method.access_flags) == flags)
       return method;
-    method = new Method (this, flags);
-    method.setName(name);
+    method = addMethod(name, flags);
     method.arg_types = arg_types;
     method.return_type = return_type;
     return method;
@@ -367,7 +499,21 @@ public class ClassType extends ObjectType
     return meth;
   }
 
-  public Method getDeclaredMethods()
+  /** Add a method to this ClassType.
+    * If an existing method matches, return that.  Otherwise, create
+    * a new one. */
+  public Method getMethod (java.lang.reflect.Method method)
+  {
+    String name = method.getName();
+    Class[] parameterClasses = method.getParameterTypes();
+    Type[] parameterTypes = new Type[parameterClasses.length];
+    for (int i = parameterClasses.length;  --i >= 0; )
+      parameterTypes[i] = Type.make(parameterClasses[i]);
+    return addMethod(name, method.getModifiers(),
+                     parameterTypes, Type.make(method.getReturnType()));
+  }
+
+  public final synchronized Method getDeclaredMethods()
   {
     if ((flags & (ADD_METHODS_DONE|EXISTING_CLASS)) == EXISTING_CLASS)
       addMethods(getReflectClass());
@@ -395,7 +541,7 @@ public class ClassType extends ObjectType
    * @param filter to select methods to return
    * @param searchSupers 0 if only current class should be searched,
    *   1 if superclasses should also be searched,
-   *   2 if super-interfaces should also be search
+   *   2 if super-interfaces should also be searched
    * @return a fresh array containing the methods satisfying the filter
    */
   public Method[] getMethods (Filter filter, int searchSupers)
@@ -410,7 +556,7 @@ public class ClassType extends ObjectType
    * @param filter to select methods to return
    * @param searchSupers 0 if only current class should be searched,
    *   1 if superclasses should also be searched,
-   *   2 if super-interfaces should also be search
+   *   2 if super-interfaces should also be searched
    * @param result array to place selected methods in
    * @param offset start of where in result to place result
    * @return number of methods placed in result array
@@ -452,7 +598,7 @@ public class ClassType extends ObjectType
    * @param filter to select methods to return
    * @param searchSupers 0 if only current class should be searched,
    *   1 if superclasses should also be searched,
-   *   2 if super-interfaces should also be search
+   *   2 if super-interfaces should also be searched
    * @param result Vector to add selected methods in
    * @param context If non-null, skip if class not visible in named package.
    * @return number of methods placed in result array
@@ -494,24 +640,37 @@ public class ClassType extends ObjectType
     return count;
   }
 
+  /** Look for a matching method.
+   * @param name method name
+   * @param arg_types parameter types that must match.
+   *  Can also be null, to match any parameter type list.
+   *  Otherwise, an element of arg_types must be the same type (equals),
+   *  though a null element of arg_types is a wildcard that matches any type.
+   */
   public Method getDeclaredMethod(String name, Type[] arg_types)
   {
-    if ((flags & (ADD_METHODS_DONE|EXISTING_CLASS)) == EXISTING_CLASS)
-      addMethods(getReflectClass());
-    Method method;
-    for (method = methods;  method != null;  method = method.next)
+    int needOuterLinkArg = "<init>".equals(name) && hasOuterLink() ? 1 : 0;
+    for (Method method = getDeclaredMethods();
+	 method != null;  method = method.next)
       {
 	if (! name.equals(method.getName()))
 	  continue;
 	Type[] method_args = method.getParameterTypes();
-	if (arg_types == null || arg_types == method_args)
+	if (arg_types == null
+            || (arg_types == method_args && needOuterLinkArg==0))
 	  return method;
 	int i = arg_types.length;
-	if (i != method_args.length)
+	if (i != method_args.length-needOuterLinkArg)
 	  continue;
 	while (-- i >= 0)
 	  {
-	    if (arg_types[i] != method_args[i])
+	    Type meth_type = method_args[i+needOuterLinkArg];
+	    Type need_type = arg_types[i];
+	    if (meth_type == need_type || need_type == null)
+	      continue;
+	    String meth_sig = meth_type.getSignature();
+	    String need_sig = need_type.getSignature();
+	    if (! meth_sig.equals(need_sig))
 	      break;
 	  }
 	if (i < 0)
@@ -523,13 +682,13 @@ public class ClassType extends ObjectType
   /** Get a method with matching name and number of arguments. */
   public Method getDeclaredMethod(String name, int argCount)
   {
-    if ((flags & (ADD_METHODS_DONE|EXISTING_CLASS)) == EXISTING_CLASS)
-      addMethods(getReflectClass());
     Method result = null;
-    for (Method method = methods;  method != null;  method = method.next)
+    int needOuterLinkArg = "<init>".equals(name) && hasOuterLink() ? 1 : 0;
+    for (Method method = getDeclaredMethods();
+	 method != null;  method = method.next)
       {
 	if (name.equals(method.getName())
-	    && argCount == method.getParameterTypes().length)
+	    && argCount + needOuterLinkArg == method.getParameterTypes().length)
 	  {
 	    if (result != null)
 	      throw new Error("ambiguous call to getDeclaredMethod(\""
@@ -547,12 +706,31 @@ public class ClassType extends ObjectType
     for (;;)
       {
         Method method = cl.getDeclaredMethod(name, arg_types);
-        if (method != null)
+	if (method != null)
           return method;
         cl = cl.getSuperclass();
         if (cl == null)
-          return null;
+          break;
       }
+    cl = this;
+    for (;;)
+      {
+        ClassType[] interfaces = cl.getInterfaces();
+        if (interfaces != null)
+          {
+            for (int i = 0;  i < interfaces.length;  i++)
+              {
+                Method method
+                  = interfaces[i].getDeclaredMethod(name, arg_types);
+                if (method != null)
+                  return method;
+              }
+          }
+        cl = cl.getSuperclass();
+        if (cl == null)
+          break;
+      }
+    return null;
   }
 
   /** Use reflection to add all the declared methods of this class.
@@ -561,6 +739,12 @@ public class ClassType extends ObjectType
    * @param clas should be the same as getReflectClass(). */
   public void addMethods(Class clas)
   {
+    // Set this flag BEFORE the actual addition.
+    // This prevents this method to be called indirectly for the same class
+    // while it is executed, which would result in methods being listed
+    // twice in this class.
+    flags |= ADD_METHODS_DONE;
+
     java.lang.reflect.Method[] methods;
     try
       {
@@ -577,15 +761,12 @@ public class ClassType extends ObjectType
         if (! method.getDeclaringClass().equals(clas))
           continue;
         int modifiers = method.getModifiers();
-        if ((modifiers & (Access.PUBLIC|Access.PROTECTED)) == 0)
-          continue;
         Class[] paramTypes = method.getParameterTypes();
         int j = paramTypes.length;
         Type[] args = new Type[j];
         while (--j >= 0)
           args[j] = Type.make(paramTypes[j]);
-        Method meth = new Method (this, modifiers);
-        meth.setName(method.getName());
+        Method meth = addMethod(method.getName(), modifiers);
         meth.arg_types = args;
         meth.return_type = Type.make(method.getReturnType());
       }
@@ -613,18 +794,14 @@ public class ClassType extends ObjectType
         Type[] args = new Type[j];
         while (--j >= 0)
           args[j] = Type.make(paramTypes[j]);
-        Method meth = new Method (this, modifiers);
-        meth.setName("<init>");
+        Method meth = addMethod("<init>", modifiers);
         meth.arg_types = args;
         meth.return_type = Type.void_type;
       }
-
-    flags |= ADD_METHODS_DONE;
   }
 
   public Method[] getMatchingMethods(String name, Type[] paramTypes, int flags)
   {
-    int i = getMethodCount();
     int nMatches = 0;
     java.util.Vector matches = new java.util.Vector(10);
     for (Method method = methods;  method != null;  method = method.getNext())
@@ -684,8 +861,8 @@ public class ClassType extends ObjectType
     doFixups ();
 
     dstr.writeInt (0xcafebabe);  // magic
-    dstr.writeShort (minor_version);
-    dstr.writeShort (major_version);
+    dstr.writeShort(getClassfileMinorVersion());
+    dstr.writeShort(getClassfileMajorVersion());
 
     // Write out the constant pool.
     if (constants == null)
@@ -735,10 +912,16 @@ public class ClassType extends ObjectType
   }
 
   public byte[] writeToArray ()
-    throws java.io.IOException
   {
     ByteArrayOutputStream stream = new ByteArrayOutputStream (500);
-    writeToStream(stream);
+    try
+      {
+	writeToStream(stream);
+      }
+    catch (java.io.IOException ex)
+      {
+	throw new InternalError(ex.toString());
+      }
     return stream.toByteArray ();    
   }
 
@@ -800,6 +983,23 @@ public class ClassType extends ObjectType
     return false;
   }
 
+  /** A more efficient version of isSubclass(ClassType.make(cname)).
+   * Does not cause the named class be loaded if it hasn't been.
+   * @param cname a class name - cannot be an interface name
+   */
+  public final boolean isSubclass (String cname)
+  {
+    ClassType ctype = this;
+    for (;;)
+      {
+        if (cname.equals(ctype.getName()))
+          return true;
+        ctype = ctype.getSuperclass();
+        if (ctype == null)
+          return false;
+      }
+  }
+
   public final boolean isSubclass(ClassType other)
   {
     if (other.isInterface())
@@ -821,12 +1021,8 @@ public class ClassType extends ObjectType
   {
     if (other == nullType)
       return 1;
-    if (other instanceof PrimType)
-      return swappedCompareResult(((PrimType) other).compare(this));
-    if (other instanceof ArrayType)
-      return swappedCompareResult(((ArrayType) other).compare(this));
     if (! (other instanceof ClassType))
-      return -3;
+      return swappedCompareResult(other.compare(this));
     String name = getName();
     if (name != null && name.equals(other.getName()))
       return 0;
@@ -836,9 +1032,9 @@ public class ClassType extends ObjectType
     if (cother.isSubclass(this))
       return 1;
     if (this == tostring_type)
-      return 1;
+      return cother == Type.pointer_type ? -1 : 1;
     if (cother == tostring_type)
-      return -1;
+      return this == Type.pointer_type ? 1 : -1;
     if (this.isInterface() || cother.isInterface())
       return -2;
     return -3;
@@ -854,12 +1050,7 @@ public class ClassType extends ObjectType
    */
   public void writeExternal(ObjectOutput out) throws IOException
   {
-    String name = getName();
-    Type found = lookupType(name);
-    if (found != this)
-      throw new IOException("serializing ClassType not in class table not impleemented");
-    else
-      out.writeUTF(name);
+    out.writeUTF(getName());
   }
 
   public void readExternal(ObjectInput in)
@@ -872,11 +1063,29 @@ public class ClassType extends ObjectType
   public Object readResolve() throws ObjectStreamException
   {
     String name = getName();
-    Type found = lookupType(name);
-    if (found != null)
-      return found;
-    mapNameToType.put(name, this);
+    /* #ifdef JAVA5 */
+    // java.util.HashMap<String,Type> map = mapNameToType;
+    /* #else */
+    java.util.Hashtable map = mapNameToType;
+    /* #endif */
+    synchronized (map)
+      {
+        Type found = (Type) map.get(name);
+        if (found != null)
+          return found;
+        map.put(name, this);
+      }
     return this;
   }
 
+  /** Clear various object references, to help garbage collection. */
+  public void cleanupAfterCompilation ()
+  {
+    for (Method meth = methods;  meth != null;  meth = meth.getNext())
+      meth.cleanupAfterCompilation();
+
+    constants = null;
+    attributes = null;
+    sourceDbgExt = null;
+  }
 }
