@@ -16,7 +16,7 @@ import java.io.*;
   * in <code>CodeAttr</code>instead.
   */
 
-public class Method implements AttrContainer {
+public class Method implements AttrContainer, Member {
   private String name;
   Type[] arg_types;
   Type return_type;
@@ -100,14 +100,13 @@ public class Method implements AttrContainer {
   }
 
   /** True if control could reach here. */
-  public boolean reachableHere () { return !code.unreachable_here; }
+  public final boolean reachableHere () { return code.reachableHere(); }
 
   public Scope popScope () { return code.popScope(); }
 
   /**
    * Allocate slots for a local variable (or parameter).
    * @param local the variable we need to allocate
-   * @return the index of the (first) slot.
    * @deprecated
    */
   public void allocate_local (Variable local)
@@ -115,22 +114,34 @@ public class Method implements AttrContainer {
     local.allocateLocal(code);
   }
 
-  /** Allocate a Code attribute, and prepare to generate code. */
-
+  /** Allocate a Code attribute, and prepare to generate code.
+   * Most code generators should use the startCode convenience method. */
   public void initCode ()
   {
     if (classfile.constants == null)
       classfile.constants = new ConstantPool();
     prepareCode(0);
+    code.sourceDbgExt = classfile.sourceDbgExt;
     code.pushScope();
   }
 
+  /**
+   * @deprecated Use startCode instead
+   */
   public void init_param_slots ()
   {
     initCode ();
     code.addParamLocals();
   }
 
+  /** Recommended method to create a new CodeAttr for this Method. */
+  public CodeAttr startCode ()
+  {
+    initCode();
+    code.addParamLocals();
+    return code;
+  }
+
   void kill_local (Variable var) { var.freeLocal(code); }
 
   /** Method that must be called before we generate any instructions.
@@ -191,13 +202,6 @@ public class Method implements AttrContainer {
     code.emitPushThis();
   }
 
-  public void compile_linenumber (int linenumber)
-  {
-    if (code == null)
-      code = new CodeAttr(this);
-    code.putLineNumber(linenumber);
-  }
-
   void write (DataOutputStream dstr, ClassType classfile)
        throws java.io.IOException
   {
@@ -208,7 +212,7 @@ public class Method implements AttrContainer {
     Attribute.writeAll(this, dstr);
   }
 
-  private String signature;
+  String signature;
 
   public String getSignature ()
   {
@@ -232,7 +236,11 @@ public class Method implements AttrContainer {
     if (len < 3 || signature.charAt(0) != '(')
       throw new ClassFormatError("bad method signature");
     int pos = 1;
+    /* #ifdef JAVA5 */
+    // java.util.Stack<Type> types = new java.util.Stack<Type>();
+    /* #else */
     java.util.Stack types = new java.util.Stack();
+    /* #endif */
     for (;;)
       {
 	int arg_sig_len = Type.signatureLength(signature, pos);
@@ -310,25 +318,37 @@ public class Method implements AttrContainer {
     return next;
   }
 
+  public void listParameters (StringBuffer sbuf)
+  {
+    int args_count = arg_types.length; 
+    sbuf.append('(');
+    for (int i = 0; i < args_count; i++)
+      {
+        if (i > 0)
+          sbuf.append(',');
+        sbuf.append (arg_types[i].getName());
+      }
+    sbuf.append(')');
+  }
+
   public String toString()
   {
     StringBuffer sbuf = new StringBuffer(100);
     sbuf.append(getDeclaringClass().getName());
-    sbuf.append(' ');
+    sbuf.append('.');
     sbuf.append(name);
     if (arg_types != null)
       {
-	int args_count = arg_types.length; 
-	sbuf.append('(');
-	for (int i = 0; i < args_count; i++)
-	  {
-	    if (i > 0)
-	      sbuf.append(',');
-	    sbuf.append (arg_types[i].getName());
-	  }
-	sbuf.append(')');
+        listParameters(sbuf);
 	sbuf.append(return_type.getName());
       }
     return sbuf.toString();
   }
+
+  public void cleanupAfterCompilation ()
+  {
+    attributes = null;
+    exceptions = null;
+    code = null;
+  }
 };
