@@ -1,5 +1,4 @@
 package gnu.bytecode;
-import java.io.*;
 
 public class PrimType extends Type {
 
@@ -56,7 +55,7 @@ public class PrimType extends Type {
     char sig1 = getSignature().charAt(0);
     ClassType clas;
     Method method;
-    String cname = null;
+    String cname;
     Type[] args;
     switch (sig1)
       {
@@ -67,7 +66,7 @@ public class PrimType extends Type {
 	code.emitElse();
 	code.emitGetStatic(clas.getDeclaredField("FALSE"));
 	code.emitFi();
-	break;
+	return;
       case 'C':  cname = "java.lang.Character"; break;
       case 'B':  cname = "java.lang.Byte";      break;
       case 'S':  cname = "java.lang.Short";     break;
@@ -75,7 +74,7 @@ public class PrimType extends Type {
       case 'J':  cname = "java.lang.Long";      break;
       case 'F':  cname = "java.lang.Float";     break;
       case 'D':  cname = "java.lang.Double";    break;
-      default:   cname = null;
+      default:   cname = null; // Should never happen.
       }
     clas = ClassType.make(cname);
     args = new Type[1];
@@ -87,6 +86,22 @@ public class PrimType extends Type {
     code.emitInvokeSpecial(method);
   }
 
+  public void emitIsInstance (CodeAttr code)
+  {
+    char sig1 = (signature == null || signature.length() != 1) ? ' '
+      : signature.charAt(0);
+    if (sig1 == 'Z')  // boolean
+      boolean_ctype.emitIsInstance(code);
+    else if (sig1 == 'V')
+      {
+	code.emitPop(1);
+	code.emitPushInt(1);
+      }
+    // Have left out Character -> char, since not used by Kawa.
+    else
+      number_type.emitIsInstance(code);
+  }
+
   public void emitCoerceFromObject (CodeAttr code)
   {
     char sig1 = (signature == null || signature.length() != 1) ? ' '
@@ -172,14 +187,27 @@ public class PrimType extends Type {
       return -3;
     char sig1 = signature.charAt(0);
     String otherName = other.getName();
-    // This is very incomplete!  FIXME.
+    if (otherName == null)
+       return -1;
     switch (sig1)
       {
       case 'V':
         return 1;
+      case 'Z':
+      case 'C':
+        break;
       case 'D':
         if (otherName.equals("java.lang.Double")
-            || otherName.equals("gnu.math.DFloat"))
+            || otherName.equals("gnu.math.DFloNum"))
+          return 0; // Or maybe 1?
+        if (otherName.equals("java.lang.Float"))
+          return 1;
+        break;
+      case 'F':
+        if (otherName.equals("java.lang.Double")
+            || otherName.equals("gnu.math.DFloNum"))
+          return -1;
+        if (otherName.equals("java.lang.Float"))
           return 0; // Or maybe 1?
         break;
       case 'I':
@@ -187,7 +215,12 @@ public class PrimType extends Type {
           return 0; // Or maybe 1?
         if (otherName.equals("gnu.math.IntNum"))
           return -1;
-        break;
+        /* fall through */
+      default:
+        if (otherName.equals("java.lang.Double")
+            || otherName.equals("java.lang.Float")
+            || otherName.equals("gnu.math.DFloNum"))
+          return -1;
       }
     if (otherName.equals("java.lang.Object")
 	|| other == tostring_type)
