@@ -1,8 +1,8 @@
-// Copyright (c) 1997, 2000, 2003  Per M.A. Bothner.
+// Copyright (c) 1997, 2000, 2003, 2004, 2006  Per M.A. Bothner.
 // This is free software;  for terms and warranty disclaimer see ./COPYING.
 
 package gnu.bytecode;
-import java.io.*;
+import java.util.*;
 
 public abstract class Type {
   String signature;
@@ -13,12 +13,14 @@ public abstract class Type {
    */
   int size;
 
+  ArrayType array_type;
+
   protected Type () { }
 
   /** The type used to implement types not natively understood by the JVM.
 
    * Usually, the identity function.  However, a language might handle
-   * union types or template types or type expression scalculated at
+   * union types or template types or type expressions calculated at
    * run time.  In that case return the type used at the JVM level,
    * and known at compile time.
    */
@@ -28,14 +30,31 @@ public abstract class Type {
   }
 
   // Maps java.lang.Class to corresponding Type.
-  static java.util.Hashtable mapClassToType;
+  /* #ifdef JAVA2 */
+  /* #ifdef JAVA5 */
+  // static WeakHashMap<Class,Type> mapClassToType;
+  /* #else */
+  static WeakHashMap mapClassToType;
+  /* #endif */
+  /* #else */
+  // static Hashtable mapClassToType;
+  /* #endif */
 
   /** Maps Java type name (e.g. "java.lang.String[]") to corresponding Type. */
+  /* #ifdef JAVA5 */
+  // static java.util.HashMap<String,Type> mapNameToType;
+  /* #else */
   static java.util.Hashtable mapNameToType;
+  /* #endif */
 
   public static Type lookupType (String name)
   {
+    /* #ifdef JAVA5 */
+    // java.util.HashMap<String,Type> map = mapNameToType;
+    // synchronized (map) { return map.get(name); }
+    /* #else */
     return (Type) mapNameToType.get(name);
+    /* #endif */
   }
 
   /** Find an Type with the given name, or create a new one.
@@ -45,37 +64,56 @@ public abstract class Type {
    */
   public static Type getType (String name)
   {
-    Type type = lookupType(name);
-    if (type == null)
+    /* #ifdef JAVA5 */
+    // java.util.HashMap<String,Type> map = mapNameToType;
+    /* #else */
+    java.util.Hashtable map = mapNameToType;
+    /* #endif */
+    synchronized (map)
       {
-	if (name.endsWith("[]"))
-	  {
-	    type = getType(name.substring(0, name.length()-2));
-	    type = new ArrayType(type, name);
-	  }
-	else
-	  {
-	    ClassType cl = new ClassType(name);
-            cl.flags |= ClassType.EXISTING_CLASS;
-	    type = cl;
-	  }
-	mapNameToType.put(name, type);
+        Type type = (Type) map.get(name);
+        if (type == null)
+          {
+            if (name.endsWith("[]"))
+              type = ArrayType.make(name);
+            else
+              {
+                ClassType cl = new ClassType(name);
+                cl.flags |= ClassType.EXISTING_CLASS;
+                type = cl;
+              }
+            map.put(name, type);
+          }
+        return type;
       }
-    return type;
   }
 
   /** Register that the Type for class is type. */
-  public static void registerTypeForClass(Class clas, Type type)
+  public synchronized static void registerTypeForClass(Class clas, Type type)
   {
-    if (mapClassToType == null)
-      mapClassToType = new java.util.Hashtable(100);
-    mapClassToType.put(clas, type);
+    /* #ifdef JAVA2 */
+    /* #ifdef JAVA5 */
+    // WeakHashMap<Class,Type> map = mapClassToType;
+    // if (map == null)
+    //   mapClassToType = map = new WeakHashMap<Class,Type>(100);
+    /* #else */
+    WeakHashMap map = mapClassToType;
+    if (map == null)
+      mapClassToType = map = new WeakHashMap(100);
+    /* #endif */
+    /* #else */
+    // Hashtable map = mapClassToType;
+    // if (map == null)
+    //   mapClassToType = map = new Hashtable(100);
+    /* #endif */
+    map.put(clas, type);
     type.reflectClass = clas;
   }
 
-  public static Type make(Class reflectClass)
+  public synchronized static Type make(Class reflectClass)
   {
     Type type;
+    
     if (mapClassToType != null)
       {
 	Object t = mapClassToType.get(reflectClass);
@@ -89,18 +127,25 @@ public abstract class Type {
     else
       {
 	String name = reflectClass.getName();
-	type = lookupType(name);
-	if (type == null
-            || (type.reflectClass != reflectClass
-                && type.reflectClass != null))
-	  {
-	    ClassType cl = new ClassType(name);
-	    cl.flags |= ClassType.EXISTING_CLASS;
-	    type = cl;
-	    mapNameToType.put(name, type);
-	  }
+        /* #ifdef JAVA5 */
+        // java.util.HashMap<String,Type> map = mapNameToType;
+        /* #else */
+        java.util.Hashtable map = mapNameToType;
+        /* #endif */
+        synchronized (map)
+          {
+            type = (Type) map.get(name);
+            if (type == null
+                || (type.reflectClass != reflectClass
+                    && type.reflectClass != null))
+              {
+                ClassType cl = new ClassType(name);
+                cl.flags |= ClassType.EXISTING_CLASS;
+                type = cl;
+                mapNameToType.put(name, type);
+              }
+          }
       }
-    type.reflectClass = reflectClass;
     registerTypeForClass(reflectClass, type);
     return type;
   }
@@ -113,11 +158,20 @@ public abstract class Type {
     signature = sig;
   }
 
+  public Type (Type type)
+  {
+    this_name = type.this_name;
+    signature = type.signature;
+    size = type.size;
+    reflectClass = type.reflectClass;
+  }
+
   public Type promote () {
     return size < 4 ? int_type : this;
   }
 
   public final int getSize() { return size; }
+  public int getSizeInWords () { return size > 4 ? 2 : 1; }
 
   public final boolean isVoid () { return size == 0; }
 
@@ -156,7 +210,7 @@ public abstract class Type {
     if (c == '[')
       {
 	type = signatureToType(sig, off+1, len-1);
-	return type == null ? null : new ArrayType(type);
+	return type == null ? null : ArrayType.make(type);
       }
     if (c == 'L' && len > 2 && sig.indexOf(';', off) == len-1+off)
       return ClassType.make(sig.substring(off+1,len-1+off).replace('/', '.'));
@@ -239,6 +293,11 @@ public abstract class Type {
     return this_name;
   }
 
+  protected void setName (String name)
+  {
+    this_name = name;
+  }
+
   public static boolean isValidJavaTypeName (String name)
   {
     boolean in_name = false;
@@ -404,8 +463,13 @@ public abstract class Type {
   public static final PrimType void_type
     = new PrimType ("void", "V", 0, java.lang.Void.TYPE);
 
-  static {
-	mapNameToType = new java.util.Hashtable();
+  static
+  {
+    /* #ifdef JAVA5 */
+    // mapNameToType = new java.util.HashMap<String,Type>();
+    /* #else */
+    mapNameToType = new java.util.Hashtable();
+    /* #endif */
 	mapNameToType.put("byte",    byte_type);
 	mapNameToType.put("short",   short_type);
 	mapNameToType.put("int",     int_type);
@@ -426,30 +490,37 @@ public abstract class Type {
 
   static public ClassType string_type = ClassType.make("java.lang.String");
   /* The String type. but coercion is handled by toString. */
-  static public ClassType tostring_type = new ClassType("java.lang.String");
+  public static final ClassType tostring_type
+    = new ClassType("java.lang.String");
 
-  static public ClassType pointer_type = ClassType.make("java.lang.Object");
-  static public ClassType boolean_ctype = ClassType.make("java.lang.Boolean");
-  static public ClassType throwable_type = ClassType.make("java.lang.Throwable");
-  static public Type[] typeArray0 = new Type[0];
-  static public Method toString_method
-  = pointer_type.getDeclaredMethod("toString", 0);
-  static public ClassType number_type = ClassType.make("java.lang.Number");
-  static public Method intValue_method
-  = number_type.addMethod ("intValue", typeArray0,
+  public static final ClassType pointer_type
+    = ClassType.make("java.lang.Object");
+  public static final ClassType boolean_ctype
+    = ClassType.make("java.lang.Boolean");
+  public static final ClassType throwable_type
+    = ClassType.make("java.lang.Throwable");
+  public static final Type[] typeArray0 = new Type[0];
+  public static final Method toString_method
+    = pointer_type.getDeclaredMethod("toString", 0);
+  public static final ClassType number_type
+    = ClassType.make("java.lang.Number");
+  public static final Method intValue_method
+    = number_type.addMethod("intValue", typeArray0,
 			    int_type, Access.PUBLIC);
-  static public Method longValue_method
-  = number_type.addMethod ("longValue", typeArray0,
+  public static final Method longValue_method
+    = number_type.addMethod("longValue", typeArray0,
 			    long_type, Access.PUBLIC);
-  static public Method floatValue_method
-  = number_type.addMethod ("floatValue", typeArray0,
+  public static final Method floatValue_method
+    = number_type.addMethod("floatValue", typeArray0,
 			    float_type, Access.PUBLIC);
-  static public Method doubleValue_method
-  = number_type.addMethod ("doubleValue", typeArray0,
+  public static final Method doubleValue_method
+    = number_type.addMethod("doubleValue", typeArray0,
 			    double_type, Access.PUBLIC);
-  static public Method booleanValue_method
-  = boolean_ctype.addMethod ("booleanValue", typeArray0,
+  public static final Method booleanValue_method
+    = boolean_ctype.addMethod("booleanValue", typeArray0,
 			      boolean_type, Access.PUBLIC);
+  public static final ClassType java_lang_Class_type
+    = ClassType.make("java.lang.Class");
 
   protected Class reflectClass;
 
@@ -459,8 +530,19 @@ public abstract class Type {
     return reflectClass;
   }
 
+  public void setReflectClass(java.lang.Class rclass)
+  {
+    reflectClass = rclass;
+  }
+
   public String toString()
   {
     return "Type " + getName();
   }
+
+  public int hashCode()
+  {
+    String name = toString();
+    return name == null ? 0 : name.hashCode ();
+  }
 }
