@@ -9,30 +9,33 @@ public class ClassExp extends LambdaExp
   public boolean isSimple() { return simple; }
   public void setSimple(boolean value) { simple = value; }
 
+  /** True if there is at least one explicit "<init>" ("*init*"} method. */
+  boolean explicitInit;
+
   /** The class of instances of this class.
    * Same as super.type unless isMakingClassPair(), in which case super.type
    * is an interface, and instanceType is a class implementing the interface.
    * Using an interface plus a class gives us true multiple inheritance. */
   ClassType instanceType;
 
-  boolean makeClassPair;
-
-  public void setMakingClassPair(boolean val)
-  {
-    makeClassPair = val;
-  }
-
   /** True if we should make a pair of an interface and a class. */
   public boolean isMakingClassPair()
   {
-    return makeClassPair;
+    return ! simple;
   }
 
   /** List of base classes and implemented interfaces. */
   public Expression[] supers;
 
+  /** An artificial method named {@code "$finit$"} for evaluating
+   * non-static initializations.
+   * All constructors need to call this. */
   public LambdaExp initMethod;
 
+  /** An artificial method named {@code "$clinit$"} for evaluating 
+   * static initializations. */
+  public LambdaExp clinitMethod;
+
   public ClassExp ()
   {
     type = null;
@@ -40,43 +43,48 @@ public class ClassExp extends LambdaExp
     setCanRead(true);
   }
 
-  /*
-  public Object eval (Environment env)
+  public ClassExp (boolean simple)
   {
-    System.err.println("eval ClassExp");
-    Class clas = evalToClass();
-    return Type.make(clas);
-    }
-  */
+    this.simple = simple;
+    if (simple)
+      instanceType = type = new ClassType();
+    else
+      {
+        PairClassType ptype = new PairClassType();
+        type = ptype;
+        instanceType = new ClassType();
+        ptype.setInterface(true);
+        ptype.instanceType = instanceType;
+      }
+    setCanRead(true);
+  }
+
+  protected boolean mustCompile () { return true; }
 
   public void compile (Compilation comp, Target target)
   {
     if (target instanceof IgnoreTarget)
       return;
-    ClassType new_class = compile (comp);
+    compile (comp);
     compilePushClass(comp, target);
   }
 
   public void compilePushClass (Compilation comp, Target target)
   {
     ClassType new_class = type;
-    String className = new_class.getName();
     // Type.make(Class.forname)
 
-    ClassType typeClass = ClassType.make("java.lang.Class");
-    Method forNameClassMethod
-      = typeClass.addMethod("forName", comp.string1Arg,
-                            typeClass, Access.STATIC|Access.PUBLIC);
     gnu.bytecode.CodeAttr code = comp.getCode();
-    code.emitPushString(className);
-    code.emitInvokeStatic(forNameClassMethod);
+    comp.loadClassRef(new_class);
     ClassType typeType;
     int nargs;
     boolean needsLink = getNeedsClosureEnv();
     if (isMakingClassPair() || needsLink)
       {
-	code.emitPushString(instanceType.getName());
-	code.emitInvokeStatic(forNameClassMethod);
+        if (new_class == instanceType)
+          code.emitDup(instanceType);
+        else
+          comp.loadClassRef(instanceType);
 	typeType = ClassType.make("gnu.expr.PairClassType");
 	nargs = needsLink ? 3 : 2;
       }
@@ -88,9 +96,10 @@ public class ClassExp extends LambdaExp
     Type[] argsClass = new Type[nargs];
     if (needsLink)
       {
-	comp.curLambda.loadHeapFrame(comp);
+	getOwningLambda().loadHeapFrame(comp);
 	argsClass[--nargs] = Type.pointer_type;
       }
+    ClassType typeClass = ClassType.make("java.lang.Class");
     while (--nargs >= 0) argsClass[nargs] = typeClass;
     Method makeMethod
       = typeType.addMethod("make", argsClass,
@@ -102,41 +111,56 @@ public class ClassExp extends LambdaExp
 
   public String getJavaName ()
   {
+    String name = getName();
     return name == null ? "object" : Compilation.mangleNameIfNeeded (name);
   }
 
-  public ClassType getCompiledClassType(Compilation comp)
+  protected ClassType getCompiledClassType(Compilation comp)
+  {
+    return type;
+  }
+
+  public void setClassName (Compilation comp)
   {
-    if (! partsDeclared)
-      {
-	getType();
-	declareParts();
-      }
     if (type.getName() == null)
       {
 	String name = getName();
-	if (name == null)
-	  name = "object";
-	else
+	if (name != null)
 	  {
 	    int nlen = name.length();
 	    if (nlen > 2
 		&& name.charAt(0) == '<' && name.charAt(nlen-1) == '>')
 	      name = name.substring(1, nlen-1);
 	  }
-	if (! isSimple() || this instanceof ObjectExp)
+        if (name == null)
+          {
+	    StringBuffer nbuf = new StringBuffer(100);
+            comp.getModule().classFor(comp);
+            nbuf.append(comp.mainClass.getName());
+            nbuf.append('$');
+            int len = nbuf.length();
+            for (int i = 0;  ; i++)
+              {
+                nbuf.append(i);
+                name = nbuf.toString();
+                if (comp.findNamedClass(name) == null)
+                  break;
+                nbuf.setLength(len);
+              }
+          }
+	else if (! isSimple() || this instanceof ObjectExp)
 	  name = comp.generateClassName(name);
 	else
 	  {
 	    int start = 0;
-	    StringBuffer nbuf = new StringBuffer(100);;
+	    StringBuffer nbuf = new StringBuffer(100);
 	    for (;;)
 	      {
 		int dot = name.indexOf('.', start);
 		if (dot < 0)
 		  break;
-		nbuf.append(comp.mangleNameIfNeeded(name.substring(start,
-								  dot)));
+		nbuf.append(Compilation
+			    .mangleNameIfNeeded(name.substring(start, dot)));
 		nbuf.append('.');
 		start = dot + 1;
 	      }
@@ -151,15 +175,21 @@ public class ClassExp extends LambdaExp
 		  nbuf.append(comp.classPrefix);
 	      }
 	    if (start < name.length())
-	      nbuf.append(comp.mangleNameIfNeeded(name.substring(start)));
+	      nbuf.append(Compilation
+			  .mangleNameIfNeeded(name.substring(start)));
 	    name = nbuf.toString();
 	  }
 	type.setName(name);
+        comp.addClass(type);
+        if (isMakingClassPair())
+          {
+            instanceType.setName(type.getName()+"$class");
+            comp.addClass(instanceType);
+          }
       }
-    return type;
   }
 
-  void setTypes()
+  public void setTypes(Compilation comp)
   {
     int len = supers == null ? 0 : supers.length;
     ClassType[] superTypes = new ClassType[len];
@@ -167,44 +197,50 @@ public class ClassExp extends LambdaExp
     int j = 0;
     for (int i = 0;  i < len;  i++)
       {
-	Type st = Interpreter.getInterpreter().getTypeFor(supers[i]);
-	if (st == null || ! (st instanceof ClassType))
-	  throw new Error("invalid super type");
+        // setTypes may be called at name-resolution time (so we can can
+        // resolve against inherited field and method names).  Therefore do
+        // inlining now.  Needed (for example) for deprecated PREXIX:<> syntax.
+        supers[i] = new InlineCalls(comp).walk(supers[i]);
+
+	Type st = Language.getDefaultLanguage().getTypeFor(supers[i]);
+	if (! (st instanceof ClassType))
+          {
+            comp.setLine(supers[i]);
+            comp.error('e', "invalid super type");
+            continue;
+          }
 	ClassType t = (ClassType) st;
-	if ((t.getModifiers() & Access.INTERFACE) == 0)
+	int modifiers;
+	try
+	  {
+	    modifiers = t.getModifiers();
+	  }
+	catch (RuntimeException ex)
+	  {
+	    modifiers = 0;
+	    if (comp != null)
+	      comp.error('e', "unknown super-type "+t.getName());
+	  }
+	if ((modifiers & Access.INTERFACE) == 0)
 	  {
 	    if (j < i)
-	      throw new Error("duplicate superclass");
+              comp.error('e', "duplicate superclass for "+this);
 	    superType = t;
 	  }
 	else
 	  superTypes[j++] = t;
       }
-    if (superType == null)
+    if (! isSimple())
       {
-	if (! isSimple())
-	  {
-	    PairClassType ptype = new PairClassType();
-	    type = ptype;
-	    setMakingClassPair(true);
-	    instanceType = new ClassType();
-	    type.setInterface(true);
-	    ClassType[] interfaces = { type };
-	    // Can we better.  FIXME.
-	    instanceType.setSuper(Type.pointer_type);
-	    instanceType.setInterfaces(interfaces);
-	    ptype.instanceType = instanceType;
-	  }
-	else
-	  instanceType = type = new ClassType();
-	type.setSuper(Type.pointer_type);
+        if (superType != null)
+          comp.error('e', "non-simple class inherts from non-interface "+superType.getName());
+          
+        ClassType[] interfaces = { type };
+        // Can we better.  FIXME.
+        instanceType.setSuper(Type.pointer_type);
+        instanceType.setInterfaces(interfaces);
       }
-    else
-      {
-	instanceType = type = new ClassType();
-	type.setSuper(superType);
-      }
-    instanceType.setModifiers(Access.SUPER);
+    type.setSuper(superType == null ? Type.pointer_type : superType);
 
     ClassType[] interfaces;
     if (j == len)
@@ -217,16 +253,9 @@ public class ClassExp extends LambdaExp
     type.setInterfaces(interfaces);
   }
 
-  public Type getType()
-  {
-    if (type == null)
-      setTypes();
-    return type;
-  }
-
   boolean partsDeclared;
 
-  public void declareParts()
+  public void declareParts(Compilation comp)
   {
     if (partsDeclared)
       return;
@@ -237,7 +266,7 @@ public class ClassExp extends LambdaExp
 	// If the declaration derives from a method, don't create field.
 	if (decl.getCanRead())
 	  {
-	    int flags = Access.PUBLIC;
+	    int flags = decl.getAccessFlags(Access.PUBLIC);
 	    if (decl.getFlag(Declaration.STATIC_SPECIFIED))
 	      flags |= Access.STATIC;
 	    if (isMakingClassPair())
@@ -264,11 +293,17 @@ public class ClassExp extends LambdaExp
     for (LambdaExp child = firstChild;  child != null;
 	 child = child.nextSibling)
       {
-	if (child != initMethod || ! isMakingClassPair())
-	  child.addMethodFor(type, null, null);
+        if ("*init*".equals(child.getName()))
+          explicitInit = true;
+	if ((child != initMethod && child != clinitMethod)
+	    || ! isMakingClassPair())
+	  child.addMethodFor(type, comp, null);
 	if (isMakingClassPair())
-	  child.addMethodFor(instanceType, null, type);
+          // FIXME this is wrong if the method is static
+	  child.addMethodFor(instanceType, comp, type);
       }
+    if (! explicitInit)
+      Compilation.getConstructor(instanceType, this);
   }
 
   /** Return implementation method matching name and param types.
@@ -339,7 +374,7 @@ public class ClassExp extends LambdaExp
 	if (type != instanceType)
 	  usedSuperClasses(instanceType, comp);
 
-	String filename = getFile();
+	String filename = getFileName();
 	if (filename != null)
 	  new_class.setSourceFile (filename);
 
@@ -354,32 +389,101 @@ public class ClassExp extends LambdaExp
               : saveLambda.closureEnv;
             if (parentFrame != null)
               closureEnvField = staticLinkField
-                = instanceType.addField("this$0", parentFrame.getType());
+                = instanceType.setOuterLink((ClassType) parentFrame.getType());
 	  }
-	comp.generateConstructor(instanceType, this);
 	CodeAttr code;
 
 	for (LambdaExp child = firstChild;  child != null; )
 	  {
 	    Method save_method = comp.method;
 	    LambdaExp save_lambda = comp.curLambda;
+            String saveFilename = comp.getFileName();
+            int saveLine = comp.getLineNumber();
+            int saveColumn = comp.getColumnNumber();
+            comp.setLine(child);
 	    comp.method = child.getMainMethod();
 	    //comp.curClass = comp.method.getDeclaringClass();
-	    child.declareThis(comp.curClass);
+            Declaration childDecl = child.nameDecl;
+            if (childDecl == null
+                || ! childDecl.getFlag(Declaration.STATIC_SPECIFIED))
+              child.declareThis(comp.curClass);
 	    comp.curClass = instanceType;
 	    comp.curLambda = child;
 	    comp.method.initCode();
             child.allocChildClasses(comp);
 	    child.allocParameters(comp);
 	    child.enterFunction(comp);
-	    child.compileBody(comp);
+            if ("*init*".equals(child.getName()))
+              {
+                code = comp.getCode();
+
+                // Extract "first" expression to see if it is special.
+                Expression bodyFirst = child.body;
+                while (bodyFirst instanceof BeginExp)
+                  {
+                    BeginExp bbody = (BeginExp) bodyFirst;
+                    if (bbody.length == 0)
+                      bodyFirst = null;
+                    else
+                      bodyFirst = bbody.exps[0];
+                  }
+
+                // See if bodyFirst is a this(...) or super(...) call.
+                ClassType calledInit = null;
+                Object value;  Expression exp;
+                if (bodyFirst instanceof ApplyExp
+                    && (exp = ((ApplyExp) bodyFirst).func) instanceof QuoteExp
+                    && (value = ((QuoteExp) exp).getValue()) instanceof PrimProcedure)
+                  {
+                    PrimProcedure pproc = (PrimProcedure) value;
+                    if (pproc.isSpecial()
+                        && ("<init>".equals(pproc.method.getName())))
+                      calledInit = pproc.method.getDeclaringClass();
+                  }
+                ClassType superClass = instanceType.getSuperclass();
+                if (calledInit != null)
+                  {
+                    bodyFirst.compileWithPosition(comp, Target.Ignore);
+                    if (calledInit != instanceType && calledInit != superClass)
+                      comp.error('e', "call to <init> for not this or super class");
+                  }
+                else if (calledInit != superClass)
+                  {
+                    // Call default super constructor if there isn't an explicit
+                    // call to a super constructor.
+                    Method superConstructor
+                      = superClass.getDeclaredMethod("<init>", 0);
+                    if (superConstructor == null)
+                      comp.error('e', "super class does not have a default constructor");
+                    else
+                      {
+                        code.emitPushThis();
+                        code.emitInvokeSpecial(superConstructor);
+                      }
+                  }
+                if (calledInit != instanceType)
+                  comp.callInitMethods(getCompiledClassType(comp),
+                                       new Vector(10));
+                if (calledInit != null)
+                  // Skip bodyFirst since we already compiled it.
+                  Expression.compileButFirst(child.body, comp);
+                else
+                  child.compileBody(comp);
+              }
+            else
+              child.compileBody(comp);
 	    child.compileEnd(comp);
 	    child.compileChildMethods(comp);
 	    comp.method = save_method;
 	    comp.curClass = new_class;
 	    comp.curLambda = save_lambda;
+            comp.setLine(saveFilename, saveLine, saveColumn);
 	    child = child.nextSibling;
 	  }
+        if (! explicitInit)
+          comp.generateConstructor(instanceType, this);
+        else if (initChain != null)
+          initChain.reportError("unimplemented: explicit constructor cannot initialize ", comp);
 
 	Method[] methods = type.getMethods(AbstractMethodFilter.instance, 2);
 	for (int i = 0;  i < methods.length;  i++)
@@ -398,7 +502,7 @@ public class ClassExp extends LambdaExp
 		&& mname.charAt(2) == 't'
 		&& mname.charAt(1) == 'e'
 		&& ((ch = mname.charAt(0)) == 'g' || ch == 's'))
-	      {
+	      { // a "set" or "get" method is treated as a slot accessor.
 		Type ftype;
 		if (ch == 's' && rtype.isVoid() && ptypes.length == 1)
 		  ftype = ptypes[0];
@@ -436,7 +540,7 @@ public class ClassExp extends LambdaExp
 		    String msg = vec.size() == 0
 		      ? "missing implementation for "
 		      : "ambiguous implementation for ";
-		    comp.error('e', msg+meth);
+		    comp.error('e', msg+meth+" mname:"+mname);
 		  }
 		else
 		  {
@@ -453,6 +557,7 @@ public class ClassExp extends LambdaExp
 	      }
 	  }
 
+        generateApplyMethods(comp);
 	comp.curLambda = saveLambda;
 
 	return new_class;
@@ -466,7 +571,19 @@ public class ClassExp extends LambdaExp
 
   protected Expression walk (ExpWalker walker)
   {
-    return walker.walkClassExp(this);
+    Compilation comp = walker.getCompilation();
+    if (comp == null)
+      return walker.walkClassExp(this);
+    ClassType saveClass = comp.curClass;
+    try
+      {
+	comp.curClass = type;
+	return walker.walkClassExp(this);
+      }
+    finally
+      {
+	comp.curClass = saveClass;
+      }
   }
 
   protected void walkChildren(ExpWalker walker)
@@ -478,7 +595,15 @@ public class ClassExp extends LambdaExp
 	for (LambdaExp child = firstChild;
 	     child != null && walker.exitValue == null;
 	     child = child.nextSibling)
-	  walker.walkLambdaExp(child);
+          {
+            if (instanceType != null)
+              {
+                Declaration firstParam = child.firstDecl();
+                if (firstParam != null && firstParam.isThisParameter())
+                  firstParam.setType(type);
+              }
+            walker.walkLambdaExp(child);
+          }
       }
     finally
       {
@@ -489,38 +614,38 @@ public class ClassExp extends LambdaExp
   public void print (OutPort out)
   {
     out.startLogicalBlock("("+getExpClassName()+"/", ")", 2);
+    Object name = getSymbol();
     if (name != null)
       {
 	out.print(name);
 	out.print('/');
       }
     out.print(id);
-    out.print("/ (");
+    out.print("/fl:");  out.print(Integer.toHexString(flags));
+    out.print(" (");
     Special prevMode = null;
     int i = 0;
-    int opt_i = 0;
     int key_args = keywords == null ? 0 : keywords.length;
-    int opt_args = defaultArgs == null ? 0 : defaultArgs.length - key_args;
+    //int opt_args = defaultArgs == null ? 0 : defaultArgs.length - key_args;
     for (Declaration decl = firstDecl();  decl != null; decl = decl.nextDecl())
       {
 	if (i > 0)
 	  out.print(' ');
-	out.print(decl);
+        decl.printInfo(out);
 	i++;
       }
     out.print(") ");
     for (LambdaExp child = firstChild;  child != null;
 	 child = child.nextSibling)
       {
-	out.writeSpaceLinear();
-        out.print(" method: ");
+	out.writeBreakLinear();
         child.print(out);
       }
-    out.writeSpaceLinear();
-    if (body == null)
-      out.print("<null body>");
-    else
-      body.print (out);
+    if (body != null)
+      {
+        out.writeBreakLinear();
+        body.print (out);
+      }
     out.endLogicalBlock(")");
   }
 
@@ -529,25 +654,32 @@ public class ClassExp extends LambdaExp
     return (new ClassInitializer(this, comp)).field;
   }
 
-  /** Mangle a "slot" name to a get- or set- method name. */
+  /** Mangle a "slot" name to a get- or set- method name.
+   * @param prefix either "get" or "set"
+   * @param sname a "slot" (property) name.  This is mangled if needed.
+   */
   public static String slotToMethodName(String prefix, String sname)
   {
-    sname = Compilation.mangleNameIfNeeded(sname);
+    if (! Compilation.isValidJavaName(sname))
+      sname = Compilation.mangleName(sname, false);
     StringBuffer sbuf = new StringBuffer(sname.length()+3);
     sbuf.append(prefix);
     sbuf.append(Character.toTitleCase(sname.charAt(0)));
     sbuf.append(sname.substring(1));
     return sbuf.toString();
   }
-}
 
-class AbstractMethodFilter implements gnu.bytecode.Filter
-{
-  public static AbstractMethodFilter instance = new AbstractMethodFilter();
-
-  public boolean select(Object value)
+  /** Helper class uses by ClassExp.compile.
+   */
+  private static class AbstractMethodFilter implements gnu.bytecode.Filter
   {
-    gnu.bytecode.Method method = (gnu.bytecode.Method) value;
-    return method.isAbstract();
+    public static final AbstractMethodFilter instance
+      = new AbstractMethodFilter();
+
+    public boolean select(Object value)
+    {
+      gnu.bytecode.Method method = (gnu.bytecode.Method) value;
+      return method.isAbstract();
+    }
   }
 }
