@@ -26,24 +26,14 @@ public class IfExp extends Expression
     return Language.getDefaultLanguage(); // FIXME
   }
 
-  public Object eval (Environment env) throws Throwable
-  {
-    Language language = getLanguage();
-    if (language.isTrue((test.eval (env))))
-      return then_clause.eval (env);
-    else if (else_clause != null)
-      return else_clause.eval (env);
-    else
-      return language.noValue();
-  }
+  protected boolean mustCompile () { return false; }
 
   public void apply (CallContext ctx) throws Throwable
   {
-    Language language = getLanguage();
-    if (language.isTrue((test.eval(ctx))))
-      then_clause.match0(ctx);
+    if (getLanguage().isTrue(test.eval(ctx)))
+      then_clause.apply(ctx);
     else if (else_clause != null)
-      else_clause.match0(ctx);
+      else_clause.apply(ctx);
   }
 
   public void compile (Compilation comp, Target target)
@@ -108,6 +98,9 @@ public class IfExp extends Expression
     if (! trueInherited /* && trueLabel.hasFixups()*/)
       {
 	trueLabel.define(code);
+        // An alternative to saving and restoring callContextVar
+        // would be to surround the compile with a pushScope and popScope.
+        // That would also release any "dangling" callContextVar.
         Variable callContextSave = comp.callContextVar;
 	then_clause.compileWithPosition(comp, target);
         comp.callContextVar = callContextSave;
@@ -116,6 +109,7 @@ public class IfExp extends Expression
       {
 	code.emitElse();
 	falseLabel.define(code);
+        // See note above for then_clause.
         Variable callContextSave = comp.callContextVar;
 	if (else_clause == null)
 	  comp.compileConstant(Values.empty, target);
@@ -142,6 +136,16 @@ public class IfExp extends Expression
      else_clause = walker.walk(else_clause);
   }
 
+  public gnu.bytecode.Type getType()
+  {
+    Type t1 = then_clause.getType();
+    Type t2 = else_clause == null ? Type.void_type : else_clause.getType();
+    if (t1 == t2)
+      return t1;
+    // FIXME - implement union types
+    return Type.pointer_type;
+  }
+
   public void print (OutPort out)
   {
     out.startLogicalBlock("(If ", false, ")");
