@@ -1,12 +1,14 @@
 package gnu.expr;
 import gnu.mapping.*;
 import java.io.*;
-import gnu.lists.*;
+import gnu.text.Printable;
+import gnu.lists.Consumer;
 
-public class Keyword extends CpsProcedure implements Printable, Externalizable
+public class Keyword extends Symbol
+  implements Printable, Externalizable
 {
-  // Does not include final ':'.
-  private String name;
+  public static final Namespace keywordNamespace = new Namespace();
+  static { keywordNamespace.setName("(keywords)"); }
 
   public Keyword()
   {
@@ -14,12 +16,22 @@ public class Keyword extends CpsProcedure implements Printable, Externalizable
 
   private Keyword (String name)
   {
-    this.name = name;
+    super(keywordNamespace, name);
   }
 
-  private static java.util.Hashtable keywordTable = new java.util.Hashtable ();
+  /** Used for constructing literals (int gnu.expr.LitTable). */
+  public Keyword (Namespace namespace, String name)
+  {
+    super(namespace, name);
+  }
 
-  public int hashCode () { return name.hashCode (); }
+  /** Get the corresponding non-keyword symbol.
+   * Informally, the symbol corresponding to dropping the ':'.
+   */
+  public Symbol asSymbol ()
+  {
+    return Namespace.EmptyNamespace.getSymbol(getName());
+  }
 
   /**
    * Create or find a Keyword with a given name (without final ':').
@@ -28,10 +40,12 @@ public class Keyword extends CpsProcedure implements Printable, Externalizable
    */
   static public Keyword make (String name)
   {
-    Keyword keyword = (Keyword) keywordTable.get (name);
-    if (keyword == null) {
-      keyword = new Keyword (name);
-      keywordTable.put (name, keyword);
+    int hash = name.hashCode();
+    Keyword keyword = (Keyword) keywordNamespace.lookup(name, hash, false);
+    if (keyword == null)
+      {
+	keyword = new Keyword(name);
+	keywordNamespace.add(keyword, hash);
     }
     return keyword;
   }
@@ -50,13 +64,13 @@ public class Keyword extends CpsProcedure implements Printable, Externalizable
 
   public final String toString()
   {
-    return name+':';
+    return getName()+':';
   }
 
-  public void print(java.io.PrintWriter ps)
+  public void print (Consumer out)
   {
-    Symbols.print(name, ps);
-    ps.print(':');
+    Symbols.print(getName(), out);
+    out.write(':');
   }
 
   /**
@@ -102,44 +116,4 @@ public class Keyword extends CpsProcedure implements Printable, Externalizable
       }
     return dfault;
   }
-
-  public final String getName() { return name; }
-  
-  /**
-   * @serialData Write the keyword name (without colons) using writeUTF.
-   */
-
-  public void writeExternal(ObjectOutput out) throws IOException
-  {
-    out.writeUTF(name);
-  }
-
-  public void readExternal(ObjectInput in)
-    throws IOException, ClassNotFoundException
-  {
-    name = in.readUTF();
-  }
-
-  public Object readResolve() throws ObjectStreamException
-  {
-    return make(name);
-  }
-
-  public void apply (CallContext ctx)
-  {
-    Consumer out = ctx.consumer;
-    String tag = getName();
-    int nargs = ctx.count;
-    out.beginAttribute(tag, this);
-    for (int i = 0;  i < nargs;  i++)
-      {
-	Object arg = ctx.getArgAsObject(i);
-	if (arg instanceof Consumable)
-	  ((Consumable) arg).consume(out);
-	else
-	  out.writeObject(arg);
-      }
-    out.endAttribute();
-  }
-
 }
