@@ -5,6 +5,7 @@ import gnu.mapping.Location; // As opposed to gnu.bytecode.Location
 import gnu.text.*;
 import java.io.*;
 import gnu.kawa.reflect.StaticFieldLocation;
+import java.net.URL;
 
 /**
  * Class used to implement Scheme top-level environments.
@@ -14,19 +15,18 @@ import gnu.kawa.reflect.StaticFieldLocation;
 public class ModuleExp extends LambdaExp
 		       implements Externalizable
 {
-  public static boolean debugPrintExpr = false;
-
   public static final int EXPORT_SPECIFIED = LambdaExp.NEXT_AVAIL_FLAG;
   public static final int STATIC_SPECIFIED = EXPORT_SPECIFIED << 1;
   public static final int NONSTATIC_SPECIFIED = STATIC_SPECIFIED << 1;
   public static final int SUPERTYPE_SPECIFIED = NONSTATIC_SPECIFIED << 1;
   public static final int STATIC_RUN_SPECIFIED = SUPERTYPE_SPECIFIED << 1;
   public static final int LAZY_DECLARATIONS = STATIC_RUN_SPECIFIED << 1;
+  public static final int IMMEDIATE = LAZY_DECLARATIONS << 1;
 
   public String getJavaName ()
   {
     String name = getName();
-    return name == null ? "lambda" : Compilation.mangleName(name);
+    return name == null ? "lambda" : Compilation.mangleName(name, 0);
   }
 
   public ModuleExp ()
@@ -36,6 +36,8 @@ public class ModuleExp extends LambdaExp
   /** Used to control which .zip file dumps are generated. */
   public static String dumpZipPrefix;
 
+  static int lastZipCounter;
+
   /** Numeric identifier for this interactive "command".
    * Incremented by Shell.run, and used to set the module name,
    * and maybe the name of the --debug-dump-zip output file.
@@ -45,33 +47,36 @@ public class ModuleExp extends LambdaExp
    * if there is a race condition on the counter.) */
   public static int interactiveCounter;
 
-  ///** A cache if this has already been evaluated. */
-  //Procedure thisValue;
-
-  public static Class evalToClass (Compilation comp)
+  public static Class evalToClass (Compilation comp, URL url)
   {
     ModuleExp mexp = comp.getModule();
     SourceMessages messages = comp.getMessages();
     try
       {
-	String class_name = mexp.getJavaName ();
 	ArrayClassLoader loader = new ArrayClassLoader ();
+        if (url == null)
+          {
+            CallContext ctx = CallContext.getInstance();
+            String base = ctx.getBaseUri();
+            url = URI_utils.toURL(base);
+          }
+        loader.setResourceContext(url);
 	comp.loader = loader;
 
-	comp.compile(mexp, class_name, null);
-	// FIXME - doesn't emit warnings.
+        comp.minfo.loadByStages(Compilation.COMPILED);
+
 	if (messages.seenErrors())
 	  return null;
 
-	byte[][] classes = new byte[comp.numClasses][];
-
 	java.util.zip.ZipOutputStream zout = null;
 	if (dumpZipPrefix != null)
 	  {
 	    StringBuffer zipname = new StringBuffer(dumpZipPrefix);
-	    if (interactiveCounter >= 0)
-	      // Incremented by Shell.run.
-	      zipname.append(interactiveCounter);
+            
+            lastZipCounter++;
+	    if (interactiveCounter > lastZipCounter)
+	      lastZipCounter = interactiveCounter;
+            zipname.append(lastZipCounter);
 	    zipname.append(".zip");
 	    java.io.FileOutputStream zfout
 	      = new java.io.FileOutputStream(zipname.toString());
@@ -84,6 +89,8 @@ public class ModuleExp extends LambdaExp
 	    String className = clas.getName ();
 	    byte[] classBytes = clas.writeToArray ();
 	    loader.addClass(className, classBytes);
+            // This reduces memory leaks if we do lots of evalToClass.
+            clas.cleanupAfterCompilation();
 
 	    if (zout != null)
 	      {
@@ -109,59 +116,109 @@ public class ModuleExp extends LambdaExp
 	  ClassTypeWriter.print(comp.classes[iClass], System.out, 0);
 	*/
 
-        CallContext ctx = CallContext.getInstance();
-        ctx.value1 = comp;
-
+        Class clas = null;
 	for (int iClass = 0;  iClass < comp.numClasses;  iClass++)
 	  {
 	    ClassType ctype = comp.classes[iClass];
-            ctype.setReflectClass(loader.loadClass(ctype.getName(), false));
+            Class cclass = loader.loadClass(ctype.getName(), false);
+            ctype.setReflectClass(cclass);
             ctype.setExisting(true);
+            if (iClass == 0)
+              clas = cclass;
           }
-	return loader.loadClass (class_name, true);
+
+        ModuleInfo minfo = comp.minfo;
+        minfo.moduleClass = clas;
+        int ndeps = minfo.numDependencies;
+
+        for (int idep = 0;  idep < ndeps;  idep++)
+          {
+            ModuleInfo dep = minfo.dependencies[idep];
+            if (dep.moduleClass == null)
+              dep.moduleClass = evalToClass(dep.comp, null);
+            comp.loader.addClass(dep.moduleClass);
+          }
+
+        return clas;
       }
     catch (java.io.IOException ex)
       {
-	ex.printStackTrace(OutPort.errDefault());
-	throw new RuntimeException ("I/O error in lambda eval: "+ex);
+	throw new WrappedException("I/O error in lambda eval", ex);
       }
     catch (ClassNotFoundException ex)
       {
-	throw new RuntimeException("class not found in lambda eval");
+	throw new WrappedException("class not found in lambda eval", ex);
+      }
+    catch (Throwable ex)
+      {
+	comp.error('f', "internal compile error - caught "+ex);
+	throw WrappedException.wrapIfNeeded(ex);
       }
   }
 
   /** Flag to force compilation, even when not required. */
   public static boolean alwaysCompile = false;
 
-  public final static void evalModule (Environment env, CallContext ctx, Compilation comp) throws Throwable
+  public final static boolean evalModule (Environment env, CallContext ctx,
+                                       Compilation comp, URL url,
+                                       OutPort msg)
+    throws Throwable
   {
+    comp.getLanguage().resolve(comp);
     ModuleExp mexp = comp.getModule();
     Environment orig_env = Environment.getCurrent();
+    Compilation orig_comp = Compilation.getCurrent();
+    SourceMessages messages = comp.getMessages();
+    ClassLoader savedLoader = null;
+    Thread thread = null; // Non-null if we need to restore context ClassLoader.
     try
       {
 	if (env != orig_env)
 	  Environment.setCurrent(env);
+        if (comp != orig_comp)
+          Compilation.setCurrent(comp);
+
+        if (alwaysCompile || comp.mustCompile)
+          comp.addMainClass(mexp);
+
+        comp.walkModule(mexp);
+        comp.setState(Compilation.WALKED);
+
+        if (msg != null ? messages.checkErrors(msg, 20) : messages.seenErrors())
+          return false;
 
 	if (! alwaysCompile && ! comp.mustCompile)
 	  { // optimization - don't generate unneeded Class.
-	    if (debugPrintExpr)
+	    if (Compilation.debugPrintFinalExpr)
 	      {
-		OutPort dout = OutPort.outDefault();
-		dout.println ("[Evaluating module \""+mexp.getName()+"\":");
-		mexp.print(dout);
-		dout.println(']');
-		dout.flush();
+		msg.println ("[Evaluating final module \""+mexp.getName()+"\":");
+		mexp.print(msg);
+		msg.println(']');
+		msg.flush();
 	      }
 	    mexp.body.apply(ctx);
 	  }
 	else
 	  {
+            if (comp.mainClass == null)
+              comp.addMainClass(mexp);
+
 	    try
 	      {
-		Class clas = evalToClass(comp);
+		Class clas = evalToClass(comp, url);
 		if (clas == null)
-		  return;
+		  return false;
+                try
+                  {
+                    thread = Thread.currentThread();
+                    savedLoader = thread.getContextClassLoader();
+                    thread.setContextClassLoader(clas.getClassLoader());
+                  }
+                catch (Throwable ex)
+                  {
+                    thread = null;
+                  }
+
                 Object inst;
                 try
                   {
@@ -176,7 +233,7 @@ public class ModuleExp extends LambdaExp
 		for (Declaration decl = mexp.firstDecl();
 		     decl != null;  decl = decl.nextDecl())
 		  {
-		    Object dname = decl.getName();
+		    Object dname = decl.getSymbol();
 		    if (decl.isPrivate() || dname == null)
 		      continue;
 		    Field fld = decl.field;
@@ -187,9 +244,11 @@ public class ModuleExp extends LambdaExp
 		    // Would it be better to check if fld is FINAL?
                     // If it is, gets its value; otherwise create
                     // a FieldLocation to access it?  FIXME.
-		    if (decl.getFlag(Declaration.PROCEDURE|Declaration.IS_CONSTANT|Declaration.INDIRECT_BINDING))
+                    Expression dvalue = decl.getValue();
+		    if (decl.getFlag(Declaration.PROCEDURE|Declaration.IS_CONSTANT|Declaration.INDIRECT_BINDING)
+                        && ! (dvalue instanceof ReferenceExp
+                              && ((ReferenceExp) dvalue).getBinding().needsContext()))
 		      {
-			Expression dvalue = decl.getValue();
 			Object value;
 			if (dvalue instanceof QuoteExp
 			    && dvalue != QuoteExp.undefined_exp)
@@ -204,13 +263,16 @@ public class ModuleExp extends LambdaExp
 		      }
 		    else
 		      {
-                        Location loc
+                        StaticFieldLocation loc
                           = new StaticFieldLocation(fld.getDeclaringClass(),
                                                     fld.getName());
-			// Perhaps set loc.decl = decl?
+			loc.setDeclaration(decl);
 			env.addLocation(sym, property, loc);
 		      }
 		  }
+                if (msg != null ? messages.checkErrors(msg, 20)
+                    : messages.seenErrors())
+                  return false;
                 if (inst instanceof ModuleBody)
                   ((ModuleBody) inst).run(ctx);
 	      }
@@ -225,46 +287,12 @@ public class ModuleExp extends LambdaExp
       {
 	if (env != orig_env)
 	  Environment.setCurrent(orig_env);
+	if (comp != orig_comp)
+	  Compilation.setCurrent(orig_comp);
+        if (thread != null)
+          thread.setContextClassLoader(savedLoader);
       }
-  }
-
-  /** Call-back from compiled code to initialize literals in immediate mode.
-   * In non-immediate mode (i.e. generating class files) the compiler emits
-   * code to "re-construct" literal values.  However, in immediate mode
-   * that would be wasteful, plus we would get values that are similar (equals)
-   * to but not necessarily identical (eq) to the compile-time literal.
-   * So we need to pass the literal values to the compiled code, by using
-   * reflectiion to initialize various static fields.  This method does that.
-   * It is called from start of the the generated static initializer, which
-   * helps makes things more consistent between immediate and non-immediate
-   * mode.
-   */
-  public static void setupLiterals ()
-  {
-    CallContext ctx = CallContext.getInstance();
-    Compilation comp = (Compilation) ctx.value1;
-    try
-      {
-        Class clas = comp.loader.loadClass(comp.mainClass.getName(), true);
-
-	/* Pass literal values to the compiled code. */
-	for (Literal init = comp.litTable.literalsChain;  init != null;
-	     init = init.next)
-	  {
-	    /* DEBUGGING:
-	    OutPort out = OutPort.errDefault();
-	    out.print("init["+init.index+"]=");
-	    out.print(init.value);
-	    out.println();
-	    */
-            clas.getDeclaredField(init.field.getName())
-              .set(null, init.value);
-	  }
-      }
-    catch (Throwable ex)
-      {
-        throw new WrappedException("internal error", ex);
-      }
+    return true;
   }
 
   ClassType superType;
@@ -279,8 +307,11 @@ public class ModuleExp extends LambdaExp
 
   public final boolean isStatic ()
   {
+    // In immediate mode there is no point in a non-static module:
+    // a static module is simpler and more efficient.
     return (getFlag(STATIC_SPECIFIED)
-	    || (gnu.expr.Compilation.moduleStatic > 0
+	    || ((gnu.expr.Compilation.moduleStatic > 0
+                 || getFlag(IMMEDIATE))
 		&& ! getFlag(SUPERTYPE_SPECIFIED)
 		&& ! getFlag(NONSTATIC_SPECIFIED)));
   }
@@ -293,6 +324,14 @@ public class ModuleExp extends LambdaExp
                 || Compilation.moduleStatic == 2));
   }
 
+  public void allocChildClasses (Compilation comp)
+  {
+    declareClosureEnv();
+    if (! comp.usingCPStyle())
+      return;
+    allocFrame(comp);
+  }
+
   void allocFields (Compilation comp)
   {
     // We want the create the loc$XXX Symbol fields for unknowns first,
@@ -306,7 +345,10 @@ public class ModuleExp extends LambdaExp
       {
 	if ((decl.isSimple() && ! decl.isPublic()) || decl.field != null)
 	  continue;
-	if (decl.getFlag(Declaration.IS_UNKNOWN))
+	if (decl.getFlag(Declaration.IS_UNKNOWN)
+            // We might have an unrefered unknown if the reference gets
+            // optimized away. For example references to <CLASSNAME>.
+            && decl.getFlag(Declaration.CAN_READ|Declaration.CAN_CALL))
 	  decl.makeField(comp, null);
       }
     for (Declaration decl = firstDecl();
@@ -315,20 +357,24 @@ public class ModuleExp extends LambdaExp
 	if (decl.field != null)
 	  continue;
 	Expression value = decl.getValue();
-	if (((decl.isSimple() && ! decl.isPublic()) || decl.ignorable())
+	if (((decl.isSimple() && ! decl.isPublic()))
             // Kludge - needed for macros - see Savannah bug #13601.
             && ! decl.isNamespaceDecl()
+            && ! (decl.getFlag(Declaration.IS_CONSTANT)
+                  && decl.getFlag(Declaration.CAN_READ|Declaration.CAN_CALL))
 	    && ! (value instanceof ClassExp))
 	  continue;
 	if (decl.getFlag(Declaration.IS_UNKNOWN))
 	  continue;
+        if (value instanceof ModuleExp) // if decl set by a module-name command.
+          continue;
 	if (value instanceof LambdaExp && ! (value instanceof ClassExp))
 	  {
 	    ((LambdaExp) value).allocFieldFor(comp);
 	  }
 	else
 	  {
-            if ((! decl.getFlag(Declaration.IS_CONSTANT) && ! decl.isAlias())
+            if (! (decl.getFlag(Declaration.IS_CONSTANT) || decl.isAlias())
 		|| value == QuoteExp.undefined_exp)
 	      value = null;
 	    decl.makeField(comp, value);
@@ -383,6 +429,71 @@ public class ModuleExp extends LambdaExp
     return decls;
   }
 
+  /** Return the class this module.
+   * If not set yet, sets it now, based on the source file name.
+   */
+  public ClassType classFor (Compilation comp)
+  {
+    if (type != null && type != Compilation.typeProcedure)
+      return (ClassType) type;
+    String fileName = getFileName();
+    File file = null;
+    String mname = getName();
+    if (mname != null)
+      fileName = mname;
+    else if (fileName == null)
+      {
+        fileName = getName();
+        if (fileName == null)
+          fileName = "$unnamed_input_file$";
+      }
+    else if (filename.equals("-") || filename.equals("<stdin>"))
+      {
+        fileName = getName();
+        if (fileName == null)
+          fileName = "$stdin$";
+      }
+    else
+      {
+        file = new File(fileName);
+        fileName = file.getName();
+        int dotIndex = fileName.lastIndexOf('.');
+        if (dotIndex > 0)
+          fileName = fileName.substring (0, dotIndex);
+      }
+    String parent;
+    String className;
+    if (getName() == null)
+      setName(fileName);
+    fileName = Compilation.mangleNameIfNeeded(fileName);
+    if (comp.classPrefix.length() == 0
+        && file != null
+        && ! file.isAbsolute()
+        && (parent = file.getParent()) != null
+        && parent.length() > 0 // Probably redundant.
+        && parent.indexOf("..") < 0)
+      {
+        parent = parent.replaceAll(System.getProperty("file.separator"), "/");
+        if (parent.startsWith("./"))
+          parent = parent.substring(2);
+        className = parent.equals(".") ? fileName
+          : Compilation.mangleURI(parent) + "." + fileName;
+      }
+    else
+      className = comp.classPrefix + fileName;
+    ClassType clas = new ClassType(className);
+    setType(clas);
+    if (comp.mainLambda == this)
+      {
+        if (comp.mainClass == null)
+          comp.mainClass = clas;
+        else if (! className.equals(comp.mainClass.getName()))
+          comp.error('e', "inconsistent main class name: "+className
+                     +" - old name: "+comp.mainClass.getName());
+      }
+    return clas;
+  }
+
   public void writeExternal(ObjectOutput out) throws IOException
   {
     String name = null;
@@ -397,7 +508,7 @@ public class ModuleExp extends LambdaExp
 	if (name == null)
 	  name = getName();
 	if (name == null)
-	  name = getFile();
+	  name = getFileName();
 	out.writeObject(name);
       }
   }
