@@ -1,20 +1,15 @@
 package gnu.jemacs.buffer;
-import javax.swing.text.*;
 import gnu.mapping.*;
 import java.io.*;
+import gnu.lists.CharBuffer;
 
 public class BufferReader extends InPort
 {
-  BufferContent content;
+  CharBuffer content;
   int rangeStart;
   int rangeLength;
 
-  public BufferReader(Buffer buffer, int start, int count)
-  {
-    this(buffer.content, buffer.getName(), start, count);
-  }
-
-  public BufferReader(BufferContent content, String name, int start, int count)
+  public BufferReader(CharBuffer content, String name, int start, int count)
   {
     super(gnu.text.NullReader.nullReader, name);
     this.content = content;
