@@ -2,12 +2,11 @@ package gnu.jemacs.lang;
 import gnu.mapping.*;
 import gnu.lists.*;
 import gnu.expr.*;
-import gnu.text.Char;
+import gnu.text.*;
 import kawa.standard.Scheme;
 import gnu.bytecode.Type;
-import gnu.kawa.lispexpr.LangPrimType;
+import gnu.kawa.lispexpr.*;
 import gnu.commonlisp.lang.*;
-import gnu.kawa.functions.DisplayFormat;
 
 public class ELisp extends Lisp2
 {
@@ -47,88 +46,68 @@ public class ELisp extends Lisp2
     return (char) i;
   }
 
-  public gnu.text.Lexer getLexer(InPort inp, gnu.text.SourceMessages messages)
-  {
-    return new ELispReader(inp, messages);
-  }
-
   public String getName()
   {
     return "Emacs-Lisp";
   }
 
-  static ELisp instance;
+  static final ELisp instance;
 
-  protected void defun(String name, Object value)
+  public static final Environment elispEnvironment
+    = Environment.make("elisp-environment");
+
+  static
   {
-    gnu.commonlisp.lang.Symbols.setFunctionBinding(environ, name, value);
-    if (value instanceof Named)
+    instance = new ELisp();
+
+    instance.define("t", TRUE);
+    instance.define("nil", FALSE);
+    CallContext ctx = CallContext.getInstance();
+    Environment saveEnv = ctx.getEnvironmentRaw();
+    try
       {
-	Named n = (Named) value;
-	if (n.getName() == null)
-	  n.setName(name);
+        ctx.setEnvironmentRaw(elispEnvironment);
+        instance.initELisp();
+      }
+    finally
+      {
+        ctx.setEnvironmentRaw(saveEnv);
       }
   }
 
-  private void defun(Procedure proc)
+  ELisp ()
   {
-    defun(proc.getName(), proc);
+    environ = elispEnvironment;
   }
 
-  static int elispCounter = 0;
-
-  public ELisp()
+  private void initELisp ()
   {
-    Environment scmEnv = Scheme.builtin();
-    environ = SymbolTable.make("interaction-environment."+(++elispCounter));
-    Environment.setCurrent(environ);
-
-    TRUE = environ.getSymbol("t");
-    TRUE.set(TRUE);
-    define("nil", FALSE);
-
-    SymbolEnumeration e
-      = Scheme.getInstance().builtin().enumerateAllSymbols();
-    while (e.hasMoreElements())
-      {
-	Symbol b = e.nextSymbol();
-	Object val = b.get(null);
-	if (val != null)
-	  {
-	    String name = b.getName();
-	    if (val instanceof Procedure || val instanceof kawa.lang.Syntax)
-	      defun(name, val);
-	    else
-	      define(name, val);
-	  }
-      }
-
-    if (instance == null)
-      instance = this;
-
     try
       {
 	// Force it to be loaded now, so we can over-ride let* length etc.
-	loadClass("kawa.lib.std_syntax");
-	loadClass("kawa.lib.lists");
-	loadClass("kawa.lib.strings");
 	loadClass("gnu.commonlisp.lisp.PrimOps");
 	loadClass("gnu.jemacs.lang.NumberOps");
 	loadClass("gnu.jemacs.lang.MiscOps");
+
+        defProcStFld("emacs", "gnu.jemacs.buffer.emacs");
       }
     catch (java.lang.ClassNotFoundException ex)
       {
 	// Ignore - happens while building this directory.
       }
 
-    defun(AddOp.$Pl); // "+"
-    defun(AddOp.$Mn); // "-"
-    defun(DivideOp.$Sl); // "/"
-    defun(NumberCompare.$Eq);
-    defun(NumberCompare.$Ls);
-    defun(NumberCompare.$Gr);
-    defun(NumberCompare.$Ls$Eq);
-    defun(NumberCompare.$Gr$Eq);
+    defSntxStFld("if", "gnu.jemacs.lang.MiscOps", "if");
+    defProcStFld("invoke", "gnu.kawa.reflect.Invoke", "invoke");
+
+    defProcStFld("+", "gnu.jemacs.lang.AddOp", "$Pl");
+    defProcStFld("-", "gnu.jemacs.lang.AddOp", "$Mn");
+    defProcStFld("/", "gnu.jemacs.lang.DivideOp", "$Sl");
+    defProcStFld("=", "gnu.jemacs.lang.NumberCompare", "$Eq");
+    defProcStFld("<", "gnu.jemacs.lang.NumberCompare", "$Ls");
+    defProcStFld(">", "gnu.jemacs.lang.NumberCompare", "$Gr");
+    defProcStFld("<=", "gnu.jemacs.lang.NumberCompare", "$Ls$Eq");
+    defProcStFld(">=", "gnu.jemacs.lang.NumberCompare", "$Gr$Eq");
+
     defun("self-insert-command", new gnu.jemacs.buffer.SelfInsertCommand());
 
     lambda lambda = new gnu.jemacs.lang.lambda();
@@ -140,6 +119,8 @@ public class ELisp extends Lisp2
     defun("defun", new gnu.commonlisp.lang.defun(lambda));
     defun("function", new gnu.commonlisp.lang.function(lambda));
 
+    defun(gnu.kawa.lispexpr.LispLanguage.quote_sym,
+	  kawa.lang.Quote.plainQuote);
     defun("defgroup", new defgroup());
     defun("defcustom", new defcustom());
     defun("defvar", new gnu.commonlisp.lang.defvar(false));
@@ -149,16 +130,14 @@ public class ELisp extends Lisp2
     defun("prog1", gnu.commonlisp.lang.prog1.prog1);
     defun("prog2", gnu.commonlisp.lang.prog1.prog2);
     defun("progn", new kawa.standard.begin());
-    defun("if", new kawa.standard.ifp());
-    defun("or", new kawa.standard.and_or(false, this));
-    defun("and", new kawa.standard.and_or(true, this));
     defun("while", new gnu.jemacs.lang.While());
     defun("unwind-protect", new gnu.commonlisp.lang.UnwindProtect());
     defun("save-excursion", new gnu.jemacs.lang.SaveExcursion(false));
     defun("save-current-buffer", new gnu.jemacs.lang.SaveExcursion(true));
     defun("let", new kawa.standard.fluid_let(false, nilExpr));
+    defun("%let", kawa.standard.let.let);
     defun("let*", new kawa.standard.fluid_let(true, nilExpr));
-    define_field ("concat", "kawa.lib.strings", "string$Mnappend");
+    defProcStFld("concat", "kawa.lib.strings", "string$Mnappend");
     Procedure not = new kawa.standard.not(this);
     defun("not", not);
     defun("null", not);
@@ -167,6 +146,11 @@ public class ELisp extends Lisp2
     defun("typep", new gnu.kawa.reflect.InstanceOf(this));
     defun("princ", displayFormat);
     defun("prin1", writeFormat);
+    LocationEnumeration e = Scheme.builtin().enumerateAllLocations();
+    while (e.hasMoreElements())
+      {
+        importLocation(e.nextLocation());
+      }
     try
       {
 	loadClass("gnu.jemacs.lisp.primitives");
@@ -185,39 +169,19 @@ public class ELisp extends Lisp2
 
   public static ELisp getInstance()
   {
-    if (instance == null)
-      {
-        Environment saveEnv = Environment.getCurrent();
-        try
-          {
-            instance = new ELisp();
-          }
-        finally
-          {
-            Environment.setCurrent(saveEnv);
-          }
-      }
     return instance;
   }
 
   /** The compiler insert calls to this method for applications and applets. */
   public static void registerEnvironment()
   {
-    ELisp interp = new ELisp();
-    Interpreter.defaultInterpreter = interp;
-    Environment.setCurrent(interp.getEnvironment());
+    Language.setDefaults(instance);
   }
 
-  public Object read (InPort in)
-    throws java.io.IOException, gnu.text.SyntaxException
-  {
-    return ELispReader.readObject(in);
-  }
+  static final AbstractFormat writeFormat = new Print(true);
+  static final AbstractFormat displayFormat = new Print(false);
 
-  static final DisplayFormat writeFormat = new Print(true);
-  static final DisplayFormat displayFormat = new Print(false);
-
-  public FormatToConsumer getFormat(boolean readable)
+  public AbstractFormat getFormat(boolean readable)
   {
     return readable ? writeFormat : displayFormat;
   }
@@ -253,27 +217,14 @@ public class ELisp extends Lisp2
     return Type.make(clas);
   }
 
-  public void defineFromFieldValue(String name, Object part)
-    throws Throwable
+  public ReadTable createReadTable ()
   {
-    if (part instanceof Named)
-      name = ((Named) part).getName();
-    else
-      name = name.intern();
-    if (part instanceof Symbol)
-      environ.addSymbol((Symbol) part);
-    else if (part instanceof Procedure
-	     || part instanceof kawa.lang.Syntax)
-      gnu.commonlisp.lang.Symbols.setFunctionBinding(environ, name, part);
-    else
-      environ.define(name, part);
+    ReadTable rt = super.createReadTable();
+    rt.set('[', new ReaderVector(']'));
+    rt.remove(']');
+    rt.set('?', new ELispReadTableEntry('?'));
+    return rt;
   }
-  /* If non-final field:  FIXME
-    Binding2 binding = new Binding2(name);
-    setValue(binding, object);
-    setConstraint(binding, new gnu.kawa.reflect.ClassMemberConstraint(field));
-    environ.addBinding(binding);
-  */
 
   public static void readableChar(char ch, StringBuffer buf, boolean quote)
   {
@@ -317,14 +268,45 @@ public class ELisp extends Lisp2
 
   public static void main(String[] args)
   {
-    registerEnvironment();
+    kawa.repl.processArgs(new String[] { "--elisp" }, 0, 1);
     if (args.length == 0)
-      {
-        args = new String[3];
-        args[0] = "-e";
-        args[1] = "(emacs)";
-        args[2] = "--";
-      }
+      args = new String[] { "-e", "(emacs)", "--" };
     kawa.repl.main(args);
   }
 }
+
+class ELispReadTableEntry extends ReaderDispatchMisc
+{
+  public ELispReadTableEntry(int code)
+  {
+    super(code);
+  }
+
+  public Object read (Lexer in, int ch, int count)
+    throws java.io.IOException, SyntaxException
+  {
+    LispReader reader = (LispReader) in;
+    if (code >= 0)
+      ch = code;
+    switch (ch)
+      {
+      case '?':
+	ch = reader.read();
+	if (ch == '\\')
+	  {
+	    ch = reader.read();
+	    if (ch != ' ' && ch >= 0)
+	      ch = reader.readEscape(ch);
+	  }
+	if (ch < 0)
+	  {
+	    reader.error("unexpected EOF in character literal");
+	    ch = '?';
+	  }
+	return ELisp.getCharacter(ch);
+      }
+    reader.error("unexpected dispatch character");
+    return null;
+  }
+}
+
