@@ -5,7 +5,12 @@ import gnu.lists.*;
 import gnu.math.*;
 import gnu.expr.*;
 
-/** A Lexer to reading S-expressions in generic Lisp-like syntax. */
+/** A Lexer for reading S-expressions in generic Lisp-like syntax.
+ * This class may have outlived its usefulness: It's mostly just a
+ * wrapper around a LineBufferedReader plus a helper token-buffer.
+ * The functionality should be moved to ReadTable, though it is
+ * unclear what to do about the tokenBuffer.
+ */
 
 public class LispReader extends Lexer
 {
@@ -28,7 +33,7 @@ public class LispReader extends Lexer
       {
 	return Environment.getCurrent().getChecked(name);
       }
-    catch (UnboundSymbol ex)
+    catch (UnboundLocationException ex)
       {
 	return name;
       }
@@ -38,9 +43,11 @@ public class LispReader extends Lexer
     * Assumes the initial "#|" has already been read.
     */
   final public void readNestedComment (char c1, char c2)
-       throws java.io.IOException
+       throws java.io.IOException, SyntaxException
   {
     int commentNesting = 1;
+    int startLine = port.getLineNumber();
+    int startColumn = port.getColumnNumber();
     do
       {
 	int c = read ();
@@ -58,7 +65,9 @@ public class LispReader extends Lexer
 	  }
 	if (c < 0)
 	  {
-	    error("unexpected eof in " + c1 + c2 + " comment.");
+            eofError("unexpected end-of-file in " + c1 + c2
+                     + " comment starting here",
+                     startLine + 1, startColumn - 1);
 	    return;
 	  }
       } while (commentNesting > 0);
@@ -68,13 +77,13 @@ public class LispReader extends Lexer
     * @return Either 'P' (means preserve case), 'U' (upcase),
     * 'D' (downcase, or 'I' (invert case).
     */
-  public char getReadCase()
+  static char getReadCase()
   {
     char read_case;
     try
       {
 	String read_case_string
-	  = Environment.lookup_global("symbol-read-case").toString();
+	  = Environment.getCurrent().get("symbol-read-case", "P").toString();
 	read_case = read_case_string.charAt(0);
 	if (read_case == 'P') ;
 	else if (read_case == 'u')
@@ -91,16 +100,14 @@ public class LispReader extends Lexer
     return read_case;
   }
 
-  protected ReadTable getReadTable () { return ReadTable.getCurrent(); }
-
-  public Object readValues (int ch)
+  public Object readValues (int ch,  ReadTable rtable)
       throws java.io.IOException, SyntaxException
   {
-    return readValues(ch, getReadTable().lookup(ch));
+    return readValues(ch, rtable.lookup(ch), rtable);
   }
 
   /** May return zero or multiple values. */
-  public Object readValues (int ch, ReadTableEntry entry)
+  public Object readValues (int ch, ReadTableEntry entry, ReadTable rtable)
       throws java.io.IOException, SyntaxException
   {
     // Step numbers refer to steps in section 2.2 of the HyperSpec.
@@ -110,11 +117,12 @@ public class LispReader extends Lexer
     if (entry == null)
       {
 	// Step 2:
-	error("invalid character #\\"+((char) ch));  // FIXME
+	String err = ("invalid character #\\"+((char) ch));  // FIXME
+	if (interactive) fatal(err);
+	else error(err);
 	return Values.empty;
       }
     int kind = entry.getKind();
-    boolean inEscapes = false;
     seenEscapes = false;
     switch (kind)
       {
@@ -125,32 +133,24 @@ public class LispReader extends Lexer
       case ReadTable.NON_TERMINATING_MACRO:
 	Object value = entry.read(this, ch, -1);
 	return value;
-      case ReadTable.SINGLE_ESCAPE:
-	// Step 5:
-	ch = read();
-	if (ch < 0)
-	  eofError("unexpected EOF after single escape");
-	tokenBufferAppend(TOKEN_ESCAPE_CHAR);
-	tokenBufferAppend(ch);
-	seenEscapes = true;
-	ch = read();
-	break;
-      case ReadTable.MULTIPLE_ESCAPE:
-	// Step 6:
-	inEscapes = true;
-	seenEscapes = true;
-	ch = read();
-	break;
-      default:  // case ReadTable.CONSTITUENT:
+      case ReadTable.CONSTITUENT:
+        if (ch == rtable.postfixLookupOperator)
+          { // Force an initial ':' to be treated as a CONSTITUENT.
+            tokenBufferAppend(ch);
+            ch = read();
+          }
+      case ReadTable.SINGLE_ESCAPE: // Step 5:
+      case ReadTable.MULTIPLE_ESCAPE: // Step 6:
+      default:  // 
 	break;
       }
 
-    readToken(ch, inEscapes, getReadCase());
+    readToken(ch, getReadCase(), rtable);
     int endPos = tokenBufferLength;
     if (seenEscapes)
-      return returnSymbol(startPos, endPos);
+      return returnSymbol(startPos, endPos, rtable);
     else
-      return handleToken(startPos, endPos);
+      return handleToken(startPos, endPos, rtable);
   }
 
   public static final char TOKEN_ESCAPE_CHAR = '\uffff';
@@ -166,9 +166,10 @@ public class LispReader extends Lexer
   /** True if "IDENTIFIER:" should be treated as a keyword. */
   protected boolean finalColonIsKeyword = true;
 
-  public void readToken(int ch, boolean inEscapes, char readCase)
+  void readToken(int ch, char readCase, ReadTable rtable)
       throws java.io.IOException, SyntaxException
   {
+    boolean inEscapes = false;
     for (;; ch = read())
       {
 	if (ch < 0)
@@ -178,13 +179,23 @@ public class LispReader extends Lexer
 	    else
 	      break;
 	  }
-	ReadTableEntry entry = getReadTable().lookup(ch);
+	ReadTableEntry entry = rtable.lookup(ch);
 	if (entry == null)
 	  {
+	    if (inEscapes)
+	      {
+		tokenBufferAppend(TOKEN_ESCAPE_CHAR);
+		tokenBufferAppend(ch);
+		continue;
+	      }
 	    unread(ch);
 	    break;
 	  }
 	int kind = entry.getKind();
+        if (ch == rtable.postfixLookupOperator && ! inEscapes
+            && validPostfixLookupStart(rtable))
+          kind = ReadTable.TERMINATING_MACRO;
+                  
 	if (kind == ReadTable.SINGLE_ESCAPE)
 	  {
 	    ch = read();
@@ -247,17 +258,18 @@ public class LispReader extends Lexer
     ((InPort) port).readState = ' ';
     try
       {
+        ReadTable rtable = ReadTable.getCurrent();
 	for (;;)
 	  {
+	    int line = port.getLineNumber();
+	    int column = port.getColumnNumber();
 	    int ch = port.read();
 	    if (ch < 0)
 	      return Sequence.eofValue; // FIXME
-	    Object value = readValues(ch);
+            Object value = readValues(ch, rtable);
 	    if (value == Values.empty)
 	      continue;
-	    if (value == QuoteExp.voidExp)
-	      value = Values.empty;
-	    return value;
+	    return handlePostfix(value, rtable, line, column);
 	  }
       }
     finally
@@ -267,48 +279,105 @@ public class LispReader extends Lexer
       }
   }
 
-  /** If tokenBuffer.substring(startPos,endPos) is a number, return it.
-   * If it is a symbol, return null.
-   * Otherwise, (i.e. if it is a "potential number") throw an Exception.
-   * or return a String - for an error message
-   */
-  protected Object checkNumber(int startPos, int endPos)
+  protected boolean validPostfixLookupStart (ReadTable rtable)
+      throws java.io.IOException
   {
-    return parseNumber(tokenBuffer, startPos, endPos - startPos, 0, SCM_NUMBERS);
-    /*
-    int sawDigits = 0;
-    int sawDots = 0;
-    for (int i = startPos;  i < endPos;  i++)
+    int ch = port.peek();
+    ReadTableEntry entry;
+    if (ch < 0 || ch == ':' || (entry = rtable.lookup(ch)) == null
+        || ch == rtable.postfixLookupOperator)
+      return false;
+    int kind = entry.getKind();
+    return kind == ReadTable.CONSTITUENT
+      || kind == ReadTable.NON_TERMINATING_MACRO
+      || kind == ReadTable.MULTIPLE_ESCAPE
+      || kind == ReadTable.SINGLE_ESCAPE;
+  }
+
+  Object handlePostfix (Object value, ReadTable rtable, int line, int column)
+      throws java.io.IOException, SyntaxException
+  {
+    if (value == QuoteExp.voidExp)
+      value = Values.empty;
+    for (;;)
       {
-	char ch = tokenBuffer[i];
+        int ch = port.peek();
+        if (ch < 0 || ch != rtable.postfixLookupOperator)
+          break;
+        // A kludge to map PreOpWord to ($lookup$ Pre 'Word).
+        port.read();
+        if (! validPostfixLookupStart(rtable))
+          {
+            unread();
+            break;
+          }
+        ch = port.read();
+        Object rightOperand = readValues(ch, rtable.lookup(ch), rtable);
+        value = LList.list2(value,
+                            LList.list2(LispLanguage.quote_sym, rightOperand));
+        value = PairWithPosition.make(LispLanguage.lookup_sym, value,
+                                      port.getName(), line+1, column+1);
+      }
+    return value;
+  }
+
+  private boolean isPotentialNumber (char[] buffer, int start, int end)
+  {
+    int sawDigits = 0;
+    for (int i = start;  i < end;  i++)
+      {
+	char ch = buffer[i];
 	if (Character.isDigit(ch))
 	  sawDigits++;
+	else if (ch == '-' || ch == '+')
+	  {
+	    if (i + 1 == end)
+	      return false;
+	  }
+	else if (ch == '#')
+	  return true;
+	else if (Character.isLetter(ch) || ch == '/'
+		 || ch == '_' || ch == '^')
+ 	  {
+	    // CommonLisp defines _123 (and ^123) as a "potential number";
+	    // most implementations seem to define it as a symbol.
+	    // Scheme does defines it as a symbol.
+	    if (i == start)
+	      return false;
+	  }
+	else if (ch != '.')
+	  return false;
       }
-    if (sawDigits == endPos - startPos)
-      gnu.math.IntNum.valueOf (tokenBuffer, startPos, endPos - startPos, 10, false);
-    return null;
-    */
+    return sawDigits > 0;
   }
 
   static final int SCM_COMPLEX = 1;
   public static final int SCM_NUMBERS = SCM_COMPLEX;
 
   /** Parse a number.
-   * If not a number returns a String (error message) or null.
+   * @param buffer contains the characters of the number
+   * @param start startinging index of the number in the buffer
+   * @param count number of characters in buffer to use
+   * @param exactness either 'i' or 'I' force an inexact result,
+   *   either 'e' or 'E' force an exact result,
+   *   '\0' yields an inact or inexact depending on the form of the literal,
+   *   while ' ' is like '\0' but does not allow more exactness specifiers.
    * @param radix the number base to use or 0 if unspecified
+   * @return the number if a valid number; null or a String-valued error
+   *   message if if there was some error parsing the number.
    */
-  public static Object parseNumber(char[] buffer, int start, int count, int radix, int flags)
+  public static Object parseNumber(char[] buffer, int start, int count,
+				   char exactness, int radix, int flags)
   {
-    char exactness = ' ';
     int end = start + count;
     int pos = start;
     if (pos >= end)
-      return null;
+      return "no digits";
     char ch = buffer[pos++];
-    for (; ch == '#';  ch = buffer[pos++])
+    while (ch == '#')
       {
 	if (pos >= end)
-	  return null;
+	  return "no digits";
 	ch = buffer[pos++];
 	switch (ch)
 	  {
@@ -334,8 +403,13 @@ public class LispReader extends Lexer
 	    break;
 	  case 'e':  case 'E':
 	  case 'i':  case 'I':
-	    if (exactness != ' ')
-	      return "duplicate exactness specifier";
+	    if (exactness != '\0')
+	      {
+		if (exactness == ' ')
+		  return "non-prefix exactness specifier";
+		else
+		  return "duplicate exactness specifier";
+	      }
 	    exactness = ch;
 	    break;
 	  default:
@@ -347,7 +421,7 @@ public class LispReader extends Lexer
 		  break;
 		value = 10 * value + dig;
 		if (pos >= end)
-		  return null;
+		  return "missing letter after '#'";
 		ch = buffer[pos++];
 	      }
 	    if (ch == 'R' || ch == 'r')
@@ -361,7 +435,12 @@ public class LispReader extends Lexer
 	      }
 	    return "unknown modifier '#" + ch + '\'';
 	  }
+	if (pos >= end)
+	  return "no digits";
+	ch = buffer[pos++];
       }
+    if (exactness == '\0')
+      exactness = ' ';
     if (radix == 0)
       {
 	for (int i = count;  ; )
@@ -382,10 +461,11 @@ public class LispReader extends Lexer
       }
 
     boolean negative = ch == '-';
+    boolean numeratorNegative = negative;
     if (ch == '-' || ch == '+')
       {
 	if (pos >= end)
-	  return null;
+	  return "no digits following sign";
 	ch = buffer[pos++];
       }
 
@@ -395,7 +475,7 @@ public class LispReader extends Lexer
       {
 	char sign = buffer[start];
 	if (sign != '+' && sign != '-')
-	  return null;
+	  return "no digits";
 	if (exactness == 'i' || exactness == 'I')
 	  return new DComplex(0, negative ? -1 : 1);
 	return negative ? Complex.imMinusOne() : Complex.imOne();
@@ -486,11 +566,8 @@ public class LispReader extends Lexer
 		  return "no digits before fraction symbol '/'";
 		if (exp_seen != '\000' || decimal_point >= 0)
 		  return "fraction symbol '/' following exponent or '.'";
-		if (pos - digits_start < 18)
-		  numerator = IntNum.make(negative ? - lvalue : lvalue);
-		else
-		  numerator = IntNum.valueOf(buffer, digits_start,
-					     pos - digits_start, radix, negative);
+		numerator = valueOf(buffer, digits_start, pos - digits_start,
+				    radix, negative, lvalue);
 		digits_start = -1;
 		lvalue = 0;
 		negative = false;
@@ -508,7 +585,7 @@ public class LispReader extends Lexer
       }
 
     if (digits_start < 0)
-      return "not a number - no digits";
+      return "no digits";
 
     if (hash_seen || underscore_seen)
       {
@@ -530,10 +607,8 @@ public class LispReader extends Lexer
       }
     else
       {
-	IntNum iresult = pos - digits_start < 18
-	  ? IntNum.make(negative ? - lvalue : lvalue)
-	  : IntNum.valueOf(buffer, digits_start, pos - digits_start,
-			   radix, negative);
+	IntNum iresult = valueOf(buffer, digits_start, pos - digits_start,
+				 radix, negative, lvalue);
 	if (numerator == null)
 	  number = iresult;
 	else
@@ -545,7 +620,7 @@ public class LispReader extends Lexer
 		boolean numeratorZero = numerator.isZero();
 		if (inexact)
 		  number =  new DFloNum ((numeratorZero ? Double.NaN
-					  : negative ? Double.NEGATIVE_INFINITY
+					  : numeratorNegative ? Double.NEGATIVE_INFINITY
 					  : Double.POSITIVE_INFINITY));
 		else if (numeratorZero)
 		  return "0/0 is undefined";
@@ -559,7 +634,7 @@ public class LispReader extends Lexer
 	  }
 	if (inexact && number.isExact())
 	  // We want #i-0 or #i-0/1 to be -0.0, not 0.0.
-	  number = new DFloNum(negative && number.isZero() ? -0.0
+	  number = new DFloNum(numeratorNegative && number.isZero() ? -0.0
 			       : number.doubleValue());
       }
 
@@ -572,7 +647,8 @@ public class LispReader extends Lexer
 
 	if (ch == '@')
 	  { /* polar notation */
-	    Object angle = parseNumber(buffer, pos, end - pos, 10, flags);
+	    Object angle = parseNumber(buffer, pos, end - pos,
+				       exactness, 10, flags);
 	    if (angle instanceof String)
 	      return angle;
 	    if (! (angle instanceof RealNum))
@@ -589,11 +665,12 @@ public class LispReader extends Lexer
 	if (ch == '-' || ch == '+')
 	  {
 	    pos--;
-	    Object imag = parseNumber(buffer, pos, end - pos, 10, flags);
+	    Object imag = parseNumber(buffer, pos, end - pos,
+				      exactness, 10, flags);
 	    if (imag instanceof String)
 	      return imag;
 	    if (! (imag instanceof Complex))
-	      return "invalid numeric constant";
+	      return "invalid numeric constant ("+imag+")";
 	    Complex cimag = (Complex) imag;
 	    RealNum re = cimag.re();
 	    if (! re.isZero())
@@ -642,13 +719,17 @@ public class LispReader extends Lexer
 		    boolean neg = ch == '-';
 		    if ((ch == '-' || ch == '+') && ++pos < end)
 		      ch = buffer[pos];
-		    power = 0;
+		    power = -1;
 		    for (;;)
 		      {
 			int d = Character.digit(ch, 10);
 			if (d < 0)
-			  return "junk after unit name";
-			power = 10 * power + d;
+			  {
+			    if (power < 0)
+			      return "junk after unit name";
+			    break;
+			  }
+			power = power < 0 ? d  : 10 * power + d;
 			if (++pos == end)
 			  break;
 			if (power > 1000000)
@@ -669,7 +750,7 @@ public class LispReader extends Lexer
 		if (unit == null)
 		  unit = u;
 		else if (u instanceof Unit && unit instanceof Unit)
-		  unit = Unit.mul((Unit) unit, (Unit) u);
+		  unit = Unit.times((Unit) unit, (Unit) u);
 		else
 		  unit = LList.list3("*", unit, u);
 		if (pos >= end)
@@ -699,7 +780,7 @@ public class LispReader extends Lexer
 	      }
 
 	    if (unit == null)
-	      return null; // ??
+	      return "expected unit";
 	    else if (unit instanceof Unit)
 	      return Quantity.make(number, (Unit) unit);
 	    else
@@ -712,7 +793,22 @@ public class LispReader extends Lexer
     return number;
   }
 
-  protected Object returnSymbol(int startPos, int endPos)
+  private static IntNum valueOf (char[] buffer, int digits_start,
+				 int number_of_digits,
+				 int radix, boolean negative,
+				 long lvalue)
+  {
+    // It turns out that if number_of_digits + radix <= 28
+    // then the value will fit in a long without overflow,
+    // so we can use the value calculated in lvalue.
+    if (number_of_digits + radix <= 28)
+      return IntNum.make(negative ? - lvalue : lvalue);
+    else
+      return IntNum.valueOf(buffer, digits_start, number_of_digits,
+			    radix, negative);
+  }
+
+  protected Object returnSymbol(int startPos, int endPos, ReadTable rtable)
   {
     char readCase = getReadCase();
     if (readCase == 'I')
@@ -771,33 +867,23 @@ public class LispReader extends Lexer
 	String str = new String(tokenBuffer, startPos, len - 1);
 	return Keyword.make(str.intern());
       }
-    return makeSymbol(new String(tokenBuffer, startPos, len));
+    return rtable.makeSymbol(new String(tokenBuffer, startPos, len));
   }
 
-  /*
-  public boolean isPotentialNumber(int startPos, int endPos)
-  {
-    int len = endPos - startPos;
-    if (len <= 0)
-      return false;
-    int prevLetter = 0, curLetter;
-    int seenDigit = 0;
-    int seenDot = 0;
-    if ((string[0] < '0' || string[0] > '9')
-        && string[0] != '+' && string[0] != '-'
-        && string[0] != '.' && string[0] != '^' && string[0] != '_')
-        return 0;
-    if (string[len-1] == '+' || string[len-1] == '-') return 0;
-  }
-  */
-
   /** Classify and return a token in tokenBuffer from startPos to endPos. */
-  public Object handleToken(int startPos, int endPos)
+  public Object handleToken(int startPos, int endPos,  ReadTable rtable)
   {
-    Object value = checkNumber(startPos, endPos);
+    Object value = parseNumber(tokenBuffer, startPos, endPos - startPos,
+			       '\0', 0, SCM_NUMBERS);
     if (value != null && ! (value instanceof String))
       return value;
-    return returnSymbol(startPos, endPos);
+    if (isPotentialNumber(tokenBuffer, startPos, endPos))
+      {
+	error(value == null ? "not a valid number"
+	      : "not a valid number: " + value);
+	return IntNum.zero();
+      }
+    return returnSymbol(startPos, endPos, rtable);
   }
 
   /** Reads a C-style String escape sequence.
@@ -944,16 +1030,19 @@ public class LispReader extends Lexer
     return c;
   }
 
-   public final Object readObject (int c)
+  public final Object readObject (int c)
       throws java.io.IOException, SyntaxException
   {
     unread(c);
     return readObject();
   }
 
-  protected Object makeSymbol (String name)
+  /** Read a "command" - a top-level expression or declaration.
+   * Return Sequence.eofValue of end of file. */
+  public Object readCommand ()
+      throws java.io.IOException, SyntaxException
   {
-    return name.intern();
+    return readObject();
   }
 
   protected Object makeNil ()
@@ -969,7 +1058,7 @@ public class LispReader extends Lexer
 
   public Object makePair (Object car, Object cdr)
   {
-    Object pair = makePair(car, 0, 0);
+    Object pair = makePair(car, -1, -1);
     setCdr(pair, cdr);
     return pair;
   }
@@ -988,7 +1077,7 @@ public class LispReader extends Lexer
     throws java.io.IOException, SyntaxException
   {
     int startPos = reader.tokenBufferLength - previous;
-    reader.readToken(reader.read(), false, 'P');
+    reader.readToken(reader.read(), 'P', ReadTable.getCurrent());
     int endPos = reader.tokenBufferLength;
     if (startPos == endPos)
       {
@@ -996,7 +1085,7 @@ public class LispReader extends Lexer
 	return IntNum.zero();
       }
     Object result = LispReader.parseNumber(reader.tokenBuffer, startPos,
-					   endPos - startPos, radix, 0);
+					   endPos - startPos, '\0', radix, 0);
     if (result instanceof String)
       {
 	reader.error((String) result);
@@ -1019,7 +1108,7 @@ public class LispReader extends Lexer
       reader.eofError("unexpected EOF in character literal");
     int startPos = reader.tokenBufferLength;
     reader.tokenBufferAppend(ch);
-    reader.readToken(reader.read(), false, 'D');
+    reader.readToken(reader.read(), 'D', ReadTable.getCurrent());
     int length = reader.tokenBufferLength - startPos;
     if (length == 1)
       return Char.make(reader.tokenBuffer[startPos]);
@@ -1063,7 +1152,7 @@ public class LispReader extends Lexer
 
     int startPos = reader.tokenBufferLength;
     reader.tokenBufferAppend(ch);
-    reader.readToken(reader.read(), false, 'D');
+    reader.readToken(reader.read(), 'D', ReadTable.getCurrent());
     int length = reader.tokenBufferLength - startPos;
     String name = new String(reader.tokenBuffer, startPos, length);
     if (name.equals("optional"))
@@ -1080,7 +1169,7 @@ public class LispReader extends Lexer
     if (name.equals("default"))
       return Special.dfault;
     if (name.equals("undefined"))
-      return Interpreter.undefinedObject;
+      return Special.undefined;
     if (name.equals("null"))
       return null;
     reader.error("unknown named constant #!"+name);
@@ -1112,7 +1201,7 @@ public class LispReader extends Lexer
       }
     Object list = ReaderParens.readList(reader, '(', -1, ')');
     int len = LList.listLength(list, false);
-    if (len <= 0)
+    if (len < 0)
       {
         reader.error("invalid elements in uniform vector syntax");
         return null;
