@@ -2,8 +2,11 @@ package gnu.kawa.reflect;
 import gnu.bytecode.*;
 import gnu.mapping.*;
 import gnu.expr.*;
+import java.io.*;
 
-public class ArrayLength extends Procedure1 implements Inlineable
+public class ArrayLength 
+  extends Procedure1
+  implements Inlineable, Externalizable
 {
   Type element_type;
 
@@ -19,14 +22,25 @@ public class ArrayLength extends Procedure1 implements Inlineable
 
   public void compile (ApplyExp exp, Compilation comp, Target target)
   {
-    exp.getArgs()[0].compile(comp, new ArrayType(element_type));
+    exp.getArgs()[0].compile(comp, ArrayType.make(element_type));
     CodeAttr code = comp.getCode();
     code.emitArrayLength();
-    target.compileFromStack(comp, kawa.standard.Scheme.intType);
+    target.compileFromStack(comp, gnu.kawa.lispexpr.LangPrimType.intType);
   }
 
   public gnu.bytecode.Type getReturnType (Expression[] args)
   {
-    return kawa.standard.Scheme.intType;
+    return gnu.kawa.lispexpr.LangPrimType.intType;
+  }
+
+  public void writeExternal(ObjectOutput out) throws IOException
+  {
+    out.writeObject(element_type);
+  }
+
+  public void readExternal(ObjectInput in)
+    throws IOException, ClassNotFoundException
+  {
+    element_type = (Type) in.readObject();
   }
 }
