@@ -8,42 +8,22 @@ import java.util.Vector;
 
 public class ClassMethods extends Procedure2
 {
+  public static final ClassMethods classMethods = new ClassMethods();
+  static { classMethods.setName("class-methods"); }
+
+  /** Create a method or generic of the matching methods.
+   * @param arg0 a Class, ClassType, or a String, FString or Symbol
+   *  that names a class.
+   * @param arg1 a method name (a String, FString, or Symbol)
+   * Loosely the same as GetNamedPart.make(arg0, arg1),
+   * but with some extra conversions and checks.
+   */
   public Object apply2 (Object arg0, Object arg1)
   {
-    return apply(this, arg0, arg1, null, null, 0, 0);
+    return apply(this, arg0, arg1);
   }
 
-  public static MethodProc apply (String className, String methodName)
-  {
-    String mname = methodName.equals("new") ? "<init>"
-      : Compilation.mangleName(methodName);
-    ClassType methodClass = ClassType.make(className);
-    MethodProc proc
-      = ClassMethods.apply(methodClass, mname, null, null, 0, 0);
-    if (proc == null)
-      {
-	StringBuffer sbuf = new StringBuffer();
-	if (! methodClass.isExisting())
-	  {
-	    sbuf.append("no class named ");
-	    sbuf.append(className);
-	  }
-	else
-	  {
-	    sbuf.append("no method named `");
-	    sbuf.append(mname);
-	    sbuf.append("' in class ");
-	    sbuf.append(className);
-	  }
-	throw new RuntimeException(sbuf.toString());
-      }
-    return proc;
-  }
-
-  public static MethodProc apply(Procedure thisProc,
-                                 Object arg0, Object arg1,
-                                 Type rtype, Type[] atypes,
-                                 int modifiers, int modmask)
+  public static MethodProc apply(Procedure thisProc, Object arg0, Object arg1)
   {
     ClassType dtype;
     String mname;
@@ -63,23 +43,13 @@ public class ClassMethods extends Procedure2
       throw new WrongType(thisProc, 1, null);
     if (! ("<init>".equals(mname)))
       mname = Compilation.mangleName(mname);
-    MethodProc result = apply(dtype, mname, rtype, atypes, modifiers, modmask);
+    MethodProc result = apply(dtype, mname, '\0', Language.getDefaultLanguage());
     if (result == null)
       throw new RuntimeException("no applicable method named `"+mname+"' in "
                                  +dtype.getName());
     return result;
   }
 
-  /** Return the methods of a class with the specified name and flag.
-   * @return an array containing the methods.
-   */
-  public static PrimProcedure[] getMethods(ClassType dtype, String mname,
-                                           int modifiers, int modmask,
-                                           Interpreter interpreter)
-  {
-    return getMethods(dtype,mname,modifiers,modmask,false,interpreter);
-  }
-
   private static int removeRedundantMethods(Vector methods)
   {
     // Remove over-ridden methods.
@@ -118,19 +88,23 @@ public class ClassMethods extends Procedure2
   }
     
   /** Return the methods of a class with the specified name and flag.
+   * @param caller if non-null, check that methods are accessible in it.
    * @return an array containing the methods.
    */
   public static PrimProcedure[] getMethods(ClassType dtype, String mname,
-                                           int modifiers, int modmask,
-                                           boolean is_special,
-                                           Interpreter interpreter)
+                                           char mode,
+                                           ClassType caller,
+                                           Language language)
   {
-    MethodFilter filter = new MethodFilter(mname, modifiers, modmask);
-    boolean named_class_only = is_special || "<init>".equals(mname);
+    MethodFilter filter = new MethodFilter(mname, 0, 0, caller);
+    // FIXME kludge until we handle "language types".
+    if (dtype == Type.tostring_type)
+      dtype = Type.string_type;
+    boolean named_class_only = mode == 'P' || "<init>".equals(mname);
     Vector methods = new Vector();
     dtype.getMethods(filter, named_class_only ? 0 : 2,
 		     methods,
-		     "-"); // Should be more specific - FIXME
+                     caller == null ? "-" : caller.getPackageName());
 
     int mlength = (named_class_only ? methods.size()
 		   : removeRedundantMethods(methods));
@@ -140,9 +114,7 @@ public class ClassMethods extends Procedure2
     for (int i = mlength;  --i >= 0; )
     {
       Method method = (Method) methods.elementAt(i);
-      if ((method.getModifiers() & modmask) != modifiers)
-	continue;
-      PrimProcedure pproc = new PrimProcedure(method, is_special, interpreter);
+      PrimProcedure pproc = new PrimProcedure(method, mode, language);
       result[count++] = pproc;
     }
     return result;
@@ -191,34 +163,21 @@ public class ClassMethods extends Procedure2
     return (((long) numDefApplicable) << 32) + (long) numPosApplicable;
   }
 
+  /** Find methods.
+   * @param dtype class to search
+   * @param mname method name (already mangled, if need be).
+   * @param mode one of 'P' (use invokespecial). 'V' (require this argument
+   *  even if method is static), or '\0' (otherwise).
+   */
   public static MethodProc apply(ClassType dtype, String mname,
-                                 Type rtype, Type[] atypes,
-                                 int modifiers, int modmask)
+                                 char mode, Language language)
   {
-    Interpreter interpreter = Interpreter.defaultInterpreter; // FIXME
-    PrimProcedure[] methods = getMethods(dtype, mname,
-                                         modifiers, modmask, interpreter);
+    PrimProcedure[] methods = getMethods(dtype, mname, mode, null, language);
     GenericProc gproc = null;
     PrimProcedure pproc = null;
     for (int i = 0;  i < methods.length;  i++)
       {
         PrimProcedure cur = methods[i];
-        if (atypes != null)
-          {
-            int applicable = cur.isApplicable(atypes);
-            if (applicable == -1)
-              continue;
-            if (pproc != null)
-              {
-                MethodProc best = MethodProc.mostSpecific(pproc, cur);
-                if (best != null)
-                  {
-                    if (cur == best)
-                      pproc = cur;
-                    continue;
-                  }
-              }
-          }
         if (pproc != null && gproc == null)
           {
             gproc = new GenericProc();
@@ -236,6 +195,10 @@ public class ClassMethods extends Procedure2
     return pproc;
   }
 
+  /** Convert an expression to a name.
+   * @return a String if the expression has the form of a symbol or
+   *   string literal, mangled as needed; otherwise null
+   */
   static String checkName(Expression exp, boolean reversible)
   {
     if (exp instanceof QuoteExp)
@@ -254,6 +217,25 @@ public class ClassMethods extends Procedure2
       }
     return null;
   }
+
+  /** Convert an expression to a name.
+   * @return a String if the expression has the form of a symbol or
+   *   string literal, with no mangling; otherwise null
+   */
+  static String checkName(Expression exp)
+  {
+    if (exp instanceof QuoteExp)
+      {
+        Object name = ((QuoteExp) exp).getValue();
+        if (name instanceof FString || name instanceof String)
+	  return name.toString();
+	else if (name instanceof Symbol)
+	  return ((Symbol) name).getName();
+	else
+	  return null;
+      }
+    return null;
+  }
 }
 
 class MethodFilter implements gnu.bytecode.Filter
@@ -262,32 +244,36 @@ class MethodFilter implements gnu.bytecode.Filter
   int nlen;
   int modifiers;
   int modmask;
+  ClassType caller;
 
-  public MethodFilter(String name, int modifiers, int modmask)
+  public MethodFilter(String name, int modifiers, int modmask,
+		      ClassType caller)
   {
     this.name = name;
     this.nlen = name.length();
     this.modifiers = modifiers;
     this.modmask = modmask;
+    this.caller = caller;
   }
 
   public boolean select(Object value)
   {
     gnu.bytecode.Method method = (gnu.bytecode.Method) value;
     String mname = method.getName();
-    if ((method.getModifiers() & modmask) == modifiers
-	&& mname.startsWith(name))
-      {
-	int mlen = mname.length();
-	char c;
-	if (mlen == nlen
-	    || (mlen == nlen + 2
-		&& mname.charAt(nlen) == '$'
-		&& ((c = mname.charAt(nlen+1)) == 'V' || c == 'X'))
-	    || (mlen == nlen + 4
-		&& mname.endsWith("$V$X")))
-	  return true;
-      }
-    return false;
+    int mmods = method.getModifiers();
+    if ((mmods & modmask) != modifiers
+	|| ! mname.startsWith(name))
+      return false;
+    int mlen = mname.length();
+    char c;
+    if (mlen != nlen
+	&& (mlen != nlen + 2
+	    || mname.charAt(nlen) != '$'
+	    || ((c = mname.charAt(nlen+1)) != 'V' && c != 'X'))
+	&& (mlen != nlen + 4
+	    || ! mname.endsWith("$V$X")))
+      return false;
+    return caller == null
+      || caller.isAccessible(method.getDeclaringClass(), mmods);
   }
 }
