@@ -1,4 +1,4 @@
-// Copyright (c) 2002  Per M.A. Bothner and Brainfood Inc.
+// Copyright (c) 2002. 2006  Per M.A. Bothner and Brainfood Inc.
 // This is free software;  for terms and warranty disclaimer see ./COPYING.
 
 package gnu.kawa.xml;
@@ -15,12 +15,22 @@ import java.util.Vector;
 public class HttpPrinter extends FilterConsumer
 {
   Vector headers = new Vector();
+  /* #ifdef JAVA5 */
+  // StringBuilder sbuf = new StringBuilder(100);
+  /* #else */
   StringBuffer sbuf = new StringBuffer(100);
-  String currentHeader;
+  /* #endif */
+  Object currentHeader;
+
+  /** 1 - implicit; 2: explicit. */
+  private int seenBeginDocument;
 
   protected String sawContentType;
 
-  OutputStream ostream;
+  /** Difference between number of beginGroup and endGroup calls so far. */
+  private int groupNesting;
+
+  protected OutputStream ostream;
   OutPort writer;
 
   public HttpPrinter(OutputStream out)
@@ -79,23 +89,23 @@ public class HttpPrinter extends FilterConsumer
   {
     if (label.equalsIgnoreCase("Content-type"))
       sawContentType = value;
-    headers.add(label);
-    headers.add(value);
+    headers.addElement(label);
+    headers.addElement(value);
   }
 
-  public void beginAttribute(String attrName, Object attrType)
+  public void beginAttribute(Object attrType)
   {
     if (base == null)
-      currentHeader = attrName;
+      currentHeader = attrType;
     else
-      base.beginAttribute(attrName, attrType);
+      base.beginAttribute(attrType);
   }
 
   public void endAttribute()
   {
     if (currentHeader != null)
       {
-	addHeader(currentHeader, sbuf.toString());
+	addHeader(currentHeader.toString(), sbuf.toString());
 	sbuf.setLength(0);
 	currentHeader = null;
       }
@@ -121,6 +131,11 @@ public class HttpPrinter extends FilterConsumer
 	else if ("text/plain".equalsIgnoreCase(sawContentType))
 	  style = "plain";
 	base = XMLPrinter.make(writer, style);
+        if (seenBeginDocument == 0)
+          {
+            base.beginDocument();
+            seenBeginDocument = 1;
+          }
 	try
 	  {
 	    printHeaders();
@@ -130,43 +145,110 @@ public class HttpPrinter extends FilterConsumer
 	    throw new RuntimeException(ex.toString());
 	  }
       }
+    /* #ifdef use:java.lang.CharSequence */
+    append(sbuf);
+    /* #else */
+    // append(sbuf.toString());
+    /* #endif */
+    sbuf.setLength(0);
   }
 
-  public void beginGroup(String typeName, Object type)
+  public void beginGroup(Object type)
   {
     if (sawContentType == null)
       {
 	String mimeType;
 	if (! seenXmlHeader)
 	  mimeType = "text/html";
-	else if (typeName.equals("html"))
+	else if (type instanceof Symbol
+                 && "html".equals(((Symbol) type).getLocalPart()))
 	  mimeType = "text/xhtml";
 	else
 	  mimeType = "text/xml";
 	addHeader("Content-type", mimeType);
       }
     beginData();
-    base.beginGroup(typeName, type);
+    base.beginGroup(type);
+    groupNesting++;
+  }
+
+  public void endGroup()
+  {
+    super.endGroup();
+    groupNesting--;
+    if (groupNesting == 0 && seenBeginDocument == 1)
+      endDocument();
   }
 
   public void writeObject(Object v)
   {
-    beginData();
-    super.writeObject(v);
+    if (v instanceof Consumable && ! (v instanceof UnescapedData))
+      ((Consumable) v).consume(this);
+    else
+      {
+	beginData();
+	super.writeObject(v);
+      }
   }
 
-  public void writeChars(String str)
+  /* #ifdef use:java.lang.CharSequence */
+  public Consumer append (CharSequence csq, int start, int end)
   {
     if (base == null)
-      sbuf.append(str);
+      {
+        /* #ifdef JAVA5 */
+        // sbuf.append(csq, start, end);
+        /* #else */
+        if (csq == null)
+          csq = "null";
+        sbuf.append(csq.subSequence(start, end).toString());
+        /* #endif */
+      }
     else
-      base.writeChars(str);
+      base.write(csq, start, end);
+    return this;
+  }
+
+  public Consumer append (CharSequence csq)
+  {
+    if (base == null)
+      {
+        /* #ifdef JAVA5 */
+        // sbuf.append(csq);
+        /* #else */
+        sbuf.append(csq.toString());
+        /* #endif */
+      }
+    else if (csq == null)
+      base.write("null");
+    else
+      base.write(csq, 0, csq.length());
+    return this;
+  }
+  /* #else */
+  // public Consumer append (String str)
+  // {
+  //   if (base == null)
+  //     sbuf.append(str);
+  //   else
+  //     base.append(str);
+  //   return this;
+  // }
+  /* #endif */
+
+  public void write(char[] buf, int off, int len)
+  {
+    if (base == null)
+      sbuf.append(buf, off, len);
+    else
+      base.write(buf, off, len);
   }
 
   public void beginDocument()
   {
     if (base != null)
       base.beginDocument();
+    seenBeginDocument = 2;
   }
 
   public void endDocument()
@@ -177,7 +259,7 @@ public class HttpPrinter extends FilterConsumer
       {
 	// else ???;
 	if (writer != null)
-	  writer.flush();
+	  writer.close();
 	if (ostream != null)
 	  ostream.flush();
       }
