@@ -14,6 +14,26 @@ public class MakeElement extends NodeConstructor
 
   public int numArgs() { return 0xFFFFF001; }
 
+  /** Optional tag.  If non-null, the element tag is this value,
+   * rather than the first parameter. */
+  public Object tag;
+
+  public int copyNamespacesMode = XMLFilter.COPY_NAMESPACES_PRESERVE;
+
+  private boolean handlingKeywordParameters;
+
+  /** Should {@code KEYWORD: EXPRESSION} be mapped to an
+   * attribute constructor? */
+  public boolean isHandlingKeywordParameters ()
+  {
+    return handlingKeywordParameters;
+  }
+
+  public void setHandlingKeywordParameters (boolean value)
+  {
+    handlingKeywordParameters = value;
+  }
+
   NamespaceBinding namespaceNodes;
 
   public NamespaceBinding getNamespaceNodes ()
@@ -26,7 +46,7 @@ public class MakeElement extends NodeConstructor
     namespaceNodes = bindings;
   }
 
-  public static SName getTagName (ApplyExp exp)
+  public static Symbol getTagName (ApplyExp exp)
   {
     Expression[] args = exp.getArgs();
     if (args.length > 0)
@@ -35,66 +55,43 @@ public class MakeElement extends NodeConstructor
 	if (arg0 instanceof QuoteExp)
 	  {
 	    Object val = ((QuoteExp) arg0).getValue();
-	    if (val instanceof SName)
-	      return (SName) val;
+	    if (val instanceof Symbol)
+	      return (Symbol) val;
 	  }
       }
     return null;
   }
 
   public static void beginGroup(Consumer out, Object qname,
+                                int copyNamespacesMode,
 				NamespaceBinding namespaceNodes)
   {
-    String name;
-    SName type;
-    if (qname instanceof SName)
-      {
-	type = new XName((SName) qname, namespaceNodes);
-	name = type.toString();
-      }
-    else if (qname instanceof Symbol)
-      {
-	type = new XName((Symbol) qname, "", namespaceNodes);
-	name = type.getLocalPart();
-      }
+    XName type;
+    if (qname instanceof Symbol)
+      type = new XName((Symbol) qname, namespaceNodes);
     else
-      {
-	name = qname.toString();
-	type = new XName(Symbol.make("", name), "", namespaceNodes);
-      }
-    out.beginGroup(name, type);
+      type = new XName(Symbol.make("", qname.toString(), ""), namespaceNodes);
+    if (out instanceof XMLFilter)
+      ((XMLFilter) out).copyNamespacesMode = copyNamespacesMode;
+    out.beginGroup(type);
   }
 
-  public static void beginGroup(Consumer out, Object qname)
+  public static void beginGroup(Consumer out, Object qname,
+                                int copyNamespacesMode)
   {
-    String name;
-    SName type;
-    if (qname instanceof SName)
-      {
-	type = (SName) qname;
-	name = type.toString();
-      }
-    else if (qname instanceof Symbol)
-      {
-	type = new SName((Symbol) qname, "");
-	name = type.getLocalPart();
-      }
+    Symbol type;
+    if (qname instanceof Symbol)
+      type = (Symbol) qname;
     else
-      {
-	name = qname.toString();
-	type = SName.make("", name, "");
-      }
-    out.beginGroup(name, type);
+      type = Symbol.make("", qname.toString(), "");
+    if (out instanceof XMLFilter)
+      ((XMLFilter) out).copyNamespacesMode = copyNamespacesMode;
+    out.beginGroup(type);
   }
 
-  public static void endGroup(Consumer out, Object type)
+  public static void endGroup(Consumer out, Object type/*FIXME:unused*/)
   {
-    String name;
-    if (type instanceof Symbol)
-      name = ((Symbol) type).getLocalName();
-    else
-      name = type.toString();
-    out.endGroup(name);
+    out.endGroup();
   }
 
   public void apply (CallContext ctx)
@@ -103,11 +100,11 @@ public class MakeElement extends NodeConstructor
     Consumer out = pushNodeContext(ctx);
     try
       {
-	Object type = ctx.getNextArg();
+	Object type = tag != null ? tag : ctx.getNextArg();
 	if (namespaceNodes != null)
-	  beginGroup(out, type, namespaceNodes);
+	  beginGroup(out, type, copyNamespacesMode, namespaceNodes);
 	else
-	  beginGroup(out, type);
+	  beginGroup(out, type, copyNamespacesMode);
 	Object endMarker = Special.dfault;
 	for (;;)
 	  {
@@ -118,6 +115,9 @@ public class MakeElement extends NodeConstructor
 	      ((Consumable) arg).consume(out);
 	    else
 	      ctx.writeValue(arg);
+            // Handling Keyword values is actually done by the Consumer.
+            if (isHandlingKeywordParameters())
+              out.endAttribute();
 	  }
 	endGroup(out, type);
       }
@@ -136,18 +136,36 @@ public class MakeElement extends NodeConstructor
     CodeAttr code = comp.getCode();
     code.emitLoad(consumer);
     code.emitDup();
-    args[0].compile(comp, Target.pushObject);
+    int i;
+    if (tag == null)
+      {
+        args[0].compile(comp, Target.pushObject);
+        i = 1;
+      }
+    else
+      {
+        comp.compileConstant(tag, Target.pushObject);
+        i = 0;
+      }
     code.emitDup(1, 1); // dup_x1
     // Stack:  consumer, tagtype, consumer, tagtype
+    code.emitPushInt(copyNamespacesMode);
     if (namespaceNodes != null)
       {
 	comp.compileConstant(namespaceNodes, Target.pushObject);
-	code.emitInvokeStatic(beginGroupMethod3);
+	code.emitInvokeStatic(beginGroupMethod4);
       }
     else
-      code.emitInvokeStatic(beginGroupMethod2);
-    for (int i = 1;  i < nargs;  i++)
-      compileChild(args[i], comp, target);
+      code.emitInvokeStatic(beginGroupMethod3);
+    for (;  i < nargs;  i++)
+      {
+        compileChild(args[i], comp, target);
+        if (isHandlingKeywordParameters())
+          {
+            code.emitLoad(consumer);
+            code.emitInvokeInterface(MakeAttribute.endAttributeMethod);
+          }
+      }
     code.emitInvokeStatic(endGroupMethod);
   }
 
@@ -158,10 +176,10 @@ public class MakeElement extends NodeConstructor
 
   static final ClassType typeMakeElement
     = ClassType.make("gnu.kawa.xml.MakeElement");
-  static final Method beginGroupMethod2
-    = typeMakeElement.getDeclaredMethod("beginGroup", 2);
   static final Method beginGroupMethod3
     = typeMakeElement.getDeclaredMethod("beginGroup", 3);
+  static final Method beginGroupMethod4
+    = typeMakeElement.getDeclaredMethod("beginGroup", 4);
   static final Method endGroupMethod
     = typeMakeElement.getDeclaredMethod("endGroup", 2);
 }
