@@ -7,7 +7,10 @@ import java.io.*;
 /** Simple adjustable-length vector whose elements are Object references. */
 
 public class FVector extends SimpleVector
-implements Externalizable, Consumable
+  implements Externalizable, Consumable
+  /* #ifdef JAVA2 */
+  , Comparable
+  /* #endif */
 {
   public Object[] data;
 
@@ -27,8 +30,11 @@ implements Externalizable, Consumable
   public FVector (int num, Object o)
   {
     Object[] data = new Object[num];
-    for (int i = 0;  i < num;  i++)
-      data[i] = o;
+    if (o != null)
+      {
+	for (int i = 0;  i < num;  i++)
+	  data[i] = o;
+      }
     this.data = data;
     this.size = num;
   }
@@ -40,11 +46,20 @@ implements Externalizable, Consumable
     this.data = data;
   }
 
-  public FVector(Sequence seq)
+  /* #ifdef JAVA2 */
+  public FVector(java.util.List seq)
   {
     this.data = new Object[seq.size()];
     addAll(seq);
   }
+  /* #endif */
+  /* #ifndef JAVA2 */
+  // public FVector(Sequence seq)
+  // {
+  //   this.data = new Object[seq.size()];
+  //   addAll(seq);
+  // }
+  /* #endif */
 
   /** Get the allocated length of the data buffer. */
   public int getBufferLength()
@@ -113,6 +128,27 @@ implements Externalizable, Consumable
     return true;
   }
 
+  /* #ifdef JAVA2 */
+  public int compareTo(Object obj)
+  {
+    FVector vec2 = (FVector) obj;
+    Object[] d1 = data;
+    Object[] d2 = vec2.data;
+    int n1 = size;
+    int n2 = vec2.size;
+    int n = n1 > n2 ? n2 : n1;
+    for (int i = 0;  i < n;  i++)
+      {
+        Comparable v1 = (Comparable) d1[i];
+        Comparable v2 = (Comparable) d2[i];
+        int d = v1.compareTo(v2);
+        if (d != 0)
+          return d;
+      }
+    return n1 - n2;
+  }
+  /* #endif */
+
   /*
   public final void setElementAt (Object new_value, int index)
   {
@@ -129,7 +165,7 @@ implements Externalizable, Consumable
        data[i] = new_value;
   }
 
-  public boolean consumeNext(int ipos, Object xpos, Consumer out)
+  public boolean consumeNext(int ipos, Consumer out)
   {
     int index = ipos >>> 1;
     if (index >= size)
@@ -138,15 +174,25 @@ implements Externalizable, Consumable
     return true;
   }
 
+  public void consumePosRange (int iposStart, int iposEnd, Consumer out)
+  {
+    if (out.ignoring())
+      return;
+    int i = iposStart >>> 1;
+    int end = iposEnd >>> 1;
+    if (end > size)
+      end = size;
+    for (;  i < end;  i++)
+      out.writeObject(data[i]);
+  }
+
   public void consume(Consumer out)
   {
-    String typeName = "#vector"; 
-    String type = typeName;
-    out.beginGroup(typeName, type);
+    out.beginGroup("#vector");
     int len = size;
     for (int i = 0;  i < len;  i++)
       out.writeObject(data[i]);
-    out.endGroup(typeName);
+    out.endGroup();
   }
 
   /**
