@@ -199,7 +199,7 @@ public class LList extends ExtSequence
 
   public Object get (int index)
   {
-    throw new ArrayIndexOutOfBoundsException (index);
+    throw new IndexOutOfBoundsException();
   }
   
   /* Count the length of a list.
@@ -297,24 +297,22 @@ public class LList extends ExtSequence
   public void consume(Consumer out)
   {
     Object list = this;
-    String typeName = "list";
-    String type = typeName;
-    out.beginGroup(typeName, type);
+    out.beginGroup("list");
     while (list instanceof Pair)
       {
 	if (list != this)
-	  out.writeChar(' ');
+	  out.write(' ');
 	Pair pair = (Pair) list;
 	out.writeObject(pair.car);
 	list = pair.cdr;
       }
     if (list != Empty)
       {
-	out.writeChar(' ');
-	out.writeChars(". ");
+	out.write(' ');
+	out.write(". ");
 	out.writeObject(checkNonList(list));
       }
-    out.endGroup(typeName);
+    out.endGroup();
   }
 
   public void readExternal(ObjectInput in)
@@ -401,6 +399,26 @@ public class LList extends ExtSequence
     return list;
   }
 
+  /** SRFI-1's cons* and Common Lisp's list* function. */
+  public static Object consX (Object[] args)
+  {
+    // Error if args.length==0.
+    Object first = args[0];
+    int n = args.length - 1;
+    if (n <= 0)
+      return first;
+    Pair result = new Pair(first, null);
+    Pair prev = result;
+    for (int i = 1;  i < n;  i++)
+      {
+        Pair next = new Pair(args[i], null);
+        prev.cdr = next;
+        prev = next;
+      }
+    prev.cdr = args[n];
+    return result;
+  }
+
   public String toString ()
   {
     Object rest = this;
