@@ -1,12 +1,7 @@
-// -*-Java-*-
 // Copyright (c) 2001, 2003  Per M.A. Bothner and Brainfood Inc.
 // This is free software;  for terms and warranty disclaimer see ./COPYING.
 
 package gnu.lists;
-/* BEGIN JAVA2 */
-import java.util.*;
-/* END JAVA2 */
-import java.io.*;
 
 /**
  * A Sequence is an ordered list of elements.
@@ -23,13 +18,13 @@ import java.io.*;
 
 public interface Sequence
 extends
-    /* BEGIN JAVA2 */
+    /* #ifdef JAVA2 */
     java.util.List,
-    /* END JAVA2 */
+    /* #endif */
     Consumable
 {
   /** Special magic end-of-file marker. */
-  public static final Object eofValue = new EofClass();
+  public static final Object eofValue = EofClass.eofValue;
 
   /** True is this sequence contains no elements. */
   public boolean isEmpty();
@@ -40,6 +35,9 @@ extends
   /** See java.util.List. */
   public Object get (int index);
 
+  /** See java.util.List. */
+  public Object set (int index, Object value);
+
   public void fill(Object value);
 
   public java.util.Enumeration elements();
@@ -66,36 +64,15 @@ extends
    * Part of a char, in contrast with INT_S8_VALUE, which is an integer. */
   public static final int TEXT_BYTE_VALUE = PRIM_VALUE + 12;
   public static final int CHAR_VALUE = PRIM_VALUE + 13;
+  public static final int CDATA_VALUE = 31;
   public static final int OBJECT_VALUE = 32;
   public static final int GROUP_VALUE = 33;
   public static final int DOCUMENT_VALUE = 34;
   public static final int ATTRIBUTE_VALUE = 35;
+  public static final int COMMENT_VALUE = 36;
+  public static final int PROCESSING_INSTRUCTION_VALUE = 37;
   /*
   public static final int NAMESPACE_ATTRIBUTE_VALUE = 16;
-  public static final int COMMENT_VALUE = 16;
-  public static final int PROCESSING_INSTRUCTION_VALUE = 16;
   public static final int ENTITY_REFERENCE_VALUE = 16;
   */
 }
-
-class EofClass implements Externalizable
-{
-  public final String toString()
-  {
-    return "#!eof";
-  }
-
-  public void writeExternal(ObjectOutput out) throws IOException
-  {
-  }
-
-  public void readExternal(ObjectInput in)
-    throws IOException, ClassNotFoundException
-  {
-  }
-
-  public Object readResolve() throws ObjectStreamException
-  {
-    return Sequence.eofValue;
-  }
-}
