@@ -6,7 +6,11 @@ import java.io.*;
 
 /** Simple adjustable-length vector of unsigned 8-bit integers (bytes). */
 
-public class U8Vector extends SimpleVector implements Externalizable
+public class U8Vector extends SimpleVector
+  implements Externalizable
+  /* #ifdef JAVA2 */
+  , Comparable
+  /* #endif */
 {
   byte[] data;
 
@@ -144,6 +148,11 @@ public class U8Vector extends SimpleVector implements Externalizable
       out.writeInt(data[i] & 0xff);
   }
 
+  public int compareTo(Object obj)
+  {
+    return compareToInt(this, (U8Vector) obj);
+  }
+
   /**
    * @serialData Write 'size' (using writeInt),
    *   followed by 'size' elements in order (using writeByte).
