@@ -42,7 +42,7 @@ public class CallContext // implements Runnable
     currentContext.set(ctx);
     /* #else */
     // if (thread instanceof Future)
-    //   ((Future) thread).context = ctx;
+    //   ((Future) thread).closure.context = ctx;
     // else
     //   threadMap.put(thread, ctx);
     /* #endif */
@@ -419,6 +419,9 @@ public class CallContext // implements Runnable
   {
     pushedFluids[--pushedFluidsCount] = null;
   }
+
+  /** Current stack of evaluation frames for interpreter. */
+  public Object[][] evalFrames;
 }
 
 /* CPS:
