@@ -1,6 +1,5 @@
 package gnu.mapping;
 import java.io.*;
-import java.text.*;
 import gnu.text.*;
 import gnu.lists.*;
 
@@ -10,24 +9,39 @@ import gnu.lists.*;
 
 public class OutPort extends PrintConsumer implements Printable
 {
-  String name;
+  Object name;
   private Writer base;
 
   // To keep track of column-numbers, we use a helper class.
   // Otherwise, it is too painful, as there is no documented
   // interface that would allow PrintWriter to be cleanly extended ...
   // The helper class also lets us make transparent use of WriterManager.
-  PrettyWriter bout;
+  protected PrettyWriter bout;
 
-  /** An index into the WriterManager's internal table. */
+  /** An index into the WriterManager's internal table.
+   * The value zero means it is unregistered. */
   protected int index;
   
-  OutPort(Writer base, PrettyWriter out, boolean autoflush)
+  protected OutPort(Writer base, PrettyWriter out, boolean autoflush)
   {
     super(out, autoflush);
     this.bout = out;
     this.base = base;
-    index = WriterManager.instance.register(out);
+    if (closeOnExit())
+      index = WriterManager.instance.register(out);
+  }
+
+  protected OutPort (OutPort out, boolean autoflush)
+  {
+    this(out, out.bout, autoflush);
+  }
+
+  protected OutPort (Writer out, boolean autoflush)
+  {
+    this(out,
+         (out instanceof OutPort ? ((OutPort) out).bout
+          : new PrettyWriter(out, true)),
+         autoflush);
   }
 
   public OutPort(Writer base, boolean printPretty, boolean autoflush)
@@ -36,7 +50,7 @@ public class OutPort extends PrintConsumer implements Printable
   }
 
   public OutPort(Writer base, boolean printPretty,
-		 boolean autoflush, String name)
+		 boolean autoflush, Object name)
   {
     this(base, new PrettyWriter(base, printPretty), autoflush);
     this.name = name;
@@ -47,23 +61,26 @@ public class OutPort extends PrintConsumer implements Printable
     this (out, null);
   }
 
-  public OutPort (OutputStream out, String name)
+  public OutPort (OutputStream out, Object name)
   {
     this(new OutputStreamWriter(out), true, name);
   }
 
   public OutPort (Writer out)
   {
-    this(out, false, false);
+    this(out,
+         out instanceof OutPort ? ((OutPort) out).bout
+         : new PrettyWriter(out, false),
+         false);
   }
 
-  public OutPort (Writer base, String name)
+  public OutPort (Writer base, Object name)
   {
     this(base, false, false);
     this.name = name;
   }
 
-  public OutPort (Writer base, boolean autoflush, String name)
+  public OutPort (Writer base, boolean autoflush, Object name)
   {
     this (base, false, autoflush);
     this.name = name;
@@ -71,45 +88,52 @@ public class OutPort extends PrintConsumer implements Printable
 
   public boolean printReadable;
 
-  // For now, these are static.  They should probably be thread-local.
-  private static OutPort outInitial = new OutPort (new LogWriter (new BufferedWriter(new OutputStreamWriter(System.out))), true, true, "<stdout>");
-  private static OutPort out = outInitial;
+  static OutPort outInitial = new OutPort (new LogWriter (new BufferedWriter(new OutputStreamWriter(System.out))), true, true, "<stdout>");
 
   private static OutPort errInitial = new OutPort (new LogWriter(new OutputStreamWriter(System.err)), true, true, "<stderr>");
-  private static OutPort err = errInitial;
 
+  public static final ThreadLocation outLocation
+    = new ThreadLocation("out-default");
+  static { outLocation.setGlobal(outInitial); }
+  public static final ThreadLocation errLocation
+    = new ThreadLocation("err-default");
+  static { errLocation.setGlobal(errInitial); }
   static public OutPort outDefault ()
   {
-    Thread thread = Thread.currentThread ();
-    if (thread instanceof Future)
-      return ((Future) thread).out;
-    return out;
+    return (OutPort) outLocation.get();
   }
 
   static public void setOutDefault (OutPort o)
   {
-    Thread thread = Thread.currentThread ();
-    if (thread instanceof Future)
-      ((Future) thread).out = o;
-    else
-      out = o;
+    outLocation.set(o);
   }
 
   static public OutPort errDefault ()
   {
-    Thread thread = Thread.currentThread ();
-    if (thread instanceof Future)
-      return ((Future) thread).err;
-    return err;
+    return (OutPort) errLocation.get();
   }
 
   static public void setErrDefault (OutPort e)
   {
-    Thread thread = Thread.currentThread ();
-    if (thread instanceof Future)
-      ((Future) thread).err = e;
-    else
-      err = e;
+    errLocation.set(e);
+  }
+
+  public static OutPort openFile(Object fname)
+    throws java.io.IOException
+  {
+      Object conv = Environment.user().get("port-char-encoding");
+      java.io.OutputStream strm = URI_utils.getOutputStream(fname);
+      strm = new java.io.BufferedOutputStream(strm);
+      java.io.Writer wr;
+      if (conv == null || conv == Boolean.TRUE)
+	wr = new java.io.OutputStreamWriter(strm);
+      else
+	{
+	  if (conv == Boolean.FALSE)
+	    conv = "8859_1";
+	  wr = new java.io.OutputStreamWriter(strm, conv.toString());
+	}
+      return new OutPort(wr, fname);
   }
 
   public void echo (char[] buf, int off, int len)  throws java.io.IOException
@@ -158,84 +182,20 @@ public class OutPort extends PrintConsumer implements Printable
   }
   */
 
-  protected static final int WORD = -2;
-  protected int prev = '\n';
-
   protected static final boolean isWordChar(char ch)
   {
     return Character.isJavaIdentifierPart(ch) || ch == '-' || ch == '+';
   }
 
-  private void startWord()
-  {
-    /*
-    if (prev == WORD || isWordChar((char) prev))
-      {
-	super.write(' ');
-      }
-    */
-    prev = WORD;
-  }
-
-  public void write (int c)
-  {
-    // if (prev == WORD && isWordChar((char) c))    out.print(' ');
-    super.write(c);
-    prev = c;
-  }
-
-  public void write (char[] buffer, int start, int count)
-  {
-    if (count > 0)
-      {
-	// if (prev == WORD && isWordChar(buffer[start]))  out.print(' ');
-	super.write(buffer, start, count);
-	prev = buffer[start+count-1];
-      }
-  }
-
-  public void write(String v)
-  {
-    int len = v.length();
-    if (len == 0)
-      return;
-    //if (prev == WORD && isWordChar(v.charAt(0)))  out.write(' ');
-    prev = v.charAt(len-1);
-    super.write(v);
-  }
-
-  /**
-   * Write a character value to a byte-stream.
-   * The default transation generates UTF-8 multi-bytes.
-   * We support character values above 0xFFFF for future extension.
-   */
-  public void writeChar (int i)
-  {
-    write (i);
-  }
-
   //  java.text.FieldPosition fieldPosition;
 
   /** If non-null, use this to print numbers. */
   java.text.NumberFormat numberFormat;
 
-  public FormatToConsumer objectFormat;
-
-  public void print(char v)
-  {
-    /*
-    if (prev == WORD && isWordChar((char) v))
-      {
-	out.print(' ');
-      }
-    */
-    super.print(v);
-    prev = v;
-  }
+  public AbstractFormat objectFormat;
 
   public void print(int v)
   {
-    startWord();
     if (numberFormat == null)
       super.print(v);
     else
@@ -244,7 +204,6 @@ public class OutPort extends PrintConsumer implements Printable
 
   public void print(long v)
   {
-    startWord();
     if (numberFormat == null)
       super.print(v);
     else
@@ -253,7 +212,6 @@ public class OutPort extends PrintConsumer implements Printable
 
   public void print(double v)
   {
-    startWord();
     if (numberFormat == null)
       super.print(v);
     else
@@ -262,7 +220,6 @@ public class OutPort extends PrintConsumer implements Printable
 
   public void print(float v)
   {
-    startWord();
     if (numberFormat == null)
       super.print(v);
     else
@@ -276,7 +233,6 @@ public class OutPort extends PrintConsumer implements Printable
 
   public void print(Object v)
   {
-    startWord();
     if (objectFormat != null)
       objectFormat.writeObject(v, this);
     else if (v instanceof Consumable)
@@ -285,59 +241,72 @@ public class OutPort extends PrintConsumer implements Printable
       super.print(v == null ? "null" : v);
   }
 
-  public void print(java.io.PrintWriter ps)
+  public void print (Consumer out)
   {
-    ps.print ("#<output-port");
+    out.write("#<output-port");
     if (name != null)
       {
-	ps.print (' ');
-	ps.print (name);
+	out.write(' ');
+	out.write(name.toString());
       }
-    ps.print ('>');
+    out.write('>');
   }
 
-  //public void writeChar(int v);
-
-  public void beginGroup(String typeName, Object type)
+  public void beginGroup(Object type)
   {
     if (objectFormat != null)
-      objectFormat.beginGroup(typeName, type, this);
+      objectFormat.beginGroup(type, this);
     else
       {
 	print('(');
-	print(typeName);
+	print(type);
       }
   }
 
-  public void endGroup(String typeName)
+  public void endGroup ()
   {
     if (objectFormat != null)
-      objectFormat.endGroup(typeName,this);
+      objectFormat.endGroup(this);
     else
-      out.print(')');
-    prev = ')';
+      print(')');
   }
 
   /** Write a attribute for the current group.
    * This is only allowed immediately after a beginGroup. */
-  public void beginAttribute(String attrName, Object attrType)
+  public void beginAttribute(Object attrType)
   {
-    print(' ');
-    print(attrName);
-    print(": ");
-    prev = WORD;
+    if (objectFormat != null)
+      objectFormat.beginAttribute(attrType, this);
+    else
+      {
+        print(' ');
+        print(attrType);
+        print(": ");
+      }
   }
 
   /** No more attributes in this group. */
   public void endAttribute()
   {
-    prev = WORD;
-    print(' ');  // FIXME
+    if (objectFormat != null)
+      objectFormat.endAttribute(this);
+    else
+      print(' ');
   }
 
-  public void writeChars(String str)
+  /** Note the end of a "word".  See {@link #writeWordStart}. */
+  public void writeWordEnd ()
   {
-    print(str);
+    bout.writeWordEnd();
+  }
+
+  /** Maybe write a word-separating space.
+   * Specifically, write a space if the previous output
+   * was {@link #writeWordEnd}.  Otherwise, do nothing.
+   */
+  public void writeWordStart ()
+  {
+    bout.writeWordStart();
   }
 
   public void freshLine()
@@ -364,8 +333,26 @@ public class OutPort extends PrintConsumer implements Printable
 
   public void close()
   {
-    super.close();
-    WriterManager.instance.unregister(index);
+    try
+      {
+        if (base instanceof OutPort && ((OutPort) base).bout == bout)
+          base.close();
+        else
+          out.close();
+      }
+    catch (IOException ex)
+      {
+        setError();
+      }
+    if (index > 0)
+      WriterManager.instance.unregister(index);
+  }
+
+  /** True if the port should be automatically closed on exit.
+   * (If so, it will be registered by WriterManager. */
+  protected boolean closeOnExit ()
+  {
+    return true;
   }
 
   public static void runCleanups ()
