@@ -1,8 +1,10 @@
-// Copyright (c) 1997, 1998, 1999, 2000  Per M.A. Bothner.
+// Copyright (c) 1997, 1998, 1999, 2000, 2006  Per M.A. Bothner.
 // This is free software;  for terms and warranty disclaimer see ./COPYING.
 
 package gnu.math;
 import java.io.*;
+import java.math.BigDecimal;
+import java.math.BigInteger;
 
 /** A class for infinite-precision integers.
  * @author Per Bothner
@@ -180,8 +182,22 @@ public class IntNum extends RatNum implements Externalizable
 
   public int sign ()
   {
-    int top = words == null ? ival : words[ival-1];
-    return top > 0 ? 1 : top < 0 ? -1 : 0;
+    int n = ival;
+    int[] w = words;
+    if (w == null)
+      return n > 0 ? 1 : n < 0 ? -1 : 0;
+    int i = w[--n];
+    if (i > 0)
+      return 1;
+    if (i < 0)
+      return -1;
+    for (;;)
+      {
+	if (n == 0)
+	  return 0;
+	if (w[--n] != 0)
+	  return 1;
+      }
   }
 
   /** Return -1, 0, or 1, depending on which value is greater. */
@@ -227,8 +243,6 @@ public class IntNum extends RatNum implements Externalizable
   {
     if (obj instanceof IntNum)
       return compare (this, (IntNum) obj);
-    if (! (obj instanceof RealNum))
-      throw new IllegalArgumentException ();
     return ((RealNum)obj).compareReversed (this);
   }
 
@@ -612,7 +626,7 @@ public class IntNum extends RatNum implements Externalizable
 
   /** Divide two integers, yielding quotient and remainder.
    * @param x the numerator in the division
-   * @param t the denominator in the division
+   * @param y the denominator in the division
    * @param quotient is set to the quotient of the result (iff quotient!=null)
    * @param remainder is set to the remainder of the result
    *  (iff remainder!=null)
@@ -700,6 +714,8 @@ public class IntNum extends RatNum implements Externalizable
 	  }
       }
 
+    while (rlen > 1 && ywords[rlen-1] == 0)
+      rlen--;
     if (ywords[rlen-1] < 0)
       {
         ywords[rlen] = 0;
@@ -811,6 +827,8 @@ public class IntNum extends RatNum implements Externalizable
 
   public static IntNum remainder (IntNum x, IntNum y)
   {
+    if (y.isZero())
+      return x;
     IntNum rem = new IntNum ();
     divide (x, y, null, rem, TRUNCATE);
     return rem.canonicalize ();
@@ -818,6 +836,8 @@ public class IntNum extends RatNum implements Externalizable
 
   public static IntNum modulo (IntNum x, IntNum y)
   {
+    if (y.isZero())
+      return x;
     IntNum rem = new IntNum ();
     divide (x, y, null, rem, FLOOR);
     return rem.canonicalize ();
@@ -1080,7 +1100,6 @@ public class IntNum extends RatNum implements Externalizable
 	  work = words;
 	int len = ival;
 
-	int buf_size = len * (MPN.chars_per_word (radix) + 1);
 	if (radix == 16)
 	  {
 	    if (neg)
@@ -1217,9 +1236,9 @@ public class IntNum extends RatNum implements Externalizable
        throws NumberFormatException
   {
     int len = s.length ();
-    // Testing (len < MPN.chars_per_word(radix)) would be more accurate,
+    // Testing (len < 2 * MPN.chars_per_word(radix)) would be more accurate,
     // but slightly more expensive, for little practical gain.
-    if (len <= 15 && radix <= 16)
+    if (len + radix <= 28)
       return IntNum.make (Long.parseLong (s, radix));
     
     int byte_len = 0;
@@ -1236,7 +1255,7 @@ public class IntNum extends RatNum implements Externalizable
 	  {
 	    int digit = Character.digit (ch, radix);
 	    if (digit < 0)
-	      break;
+	      throw new NumberFormatException("For input string: \""+s+'"');
 	    bytes[byte_len++] = (byte) digit;
 	  }
       }
@@ -1488,7 +1507,7 @@ public class IntNum extends RatNum implements Externalizable
     return IntNum.neg (this);
   }
 
-  /** Calculates ceiling(log2(this < 0 ? -this : this+1))
+  /** Calculates {@code ceiling(log2(this < 0 ? -this : this+1))}.
    * See Common Lisp: the Language, 2nd ed, p. 361.
    */
   public int intLength ()
@@ -1556,4 +1575,19 @@ public class IntNum extends RatNum implements Externalizable
     return canonicalize();
   }
 
+  public BigInteger asBigInteger ()
+  {
+    if (words == null || ival <= 2)
+      return BigInteger.valueOf(longValue());
+    return new BigInteger(toString());
+  }
+
+  public BigDecimal asBigDecimal ()
+  {
+    if (words == null)
+      return new BigDecimal(ival);
+    if (ival <= 2)
+      return BigDecimal.valueOf(longValue());
+    return new BigDecimal(toString());
+  }
 }
