@@ -1,3 +1,6 @@
+// Copyright (c) 2004  Per M.A. Bothner.
+// This is free software;  for terms and warranty disclaimer see ./COPYING.
+
 package gnu.text;
 import java.io.*;
 
@@ -37,6 +40,13 @@ public class LineBufferedReader extends FilterReader
   /** The high-water mark for pos, at a reset or line start. */
   int highestPos;
 
+  public char readState = '\n';
+  /** Return a character that indicates what we are currently reading.
+    * Returns '\n' if we are not inside read; '\"' if reading a string;
+    * '|' if inside a comment; '(' if inside a list; and
+    * ' ' if otherwise inside a read. */
+  public char getReadState () { return readState; }
+
   private int flags;
 
   // Notice the asymmetry in how "X\r\nY" is handled (assuming convertCR)..
@@ -74,7 +84,7 @@ public class LineBufferedReader extends FilterReader
   }
 
   /** The position that marks the start of the current or marked line.
-    * If the readAheadLimit && markPos < pos, then it is the start of the
+    * If the readAheadLimit > 0 && markPos < pos, then it is the start of the
     * line containing the markPos.
     * If we are at the end of a line, and have not started reading the next
     * one (and are therefore allowed by unread back to the old line),
@@ -83,7 +93,7 @@ public class LineBufferedReader extends FilterReader
     * If lineStartPos < 0, it means we went beyond the buffer maximum. */
   int lineStartPos;
 
-  String name;
+  Object name;
 
   /** The current line number (at position of lineStartPos). */
   protected int lineNumber;
@@ -184,7 +194,7 @@ public class LineBufferedReader extends FilterReader
 	saveStart = pos;
 	if (readAheadLimit > 0 && markPos < pos)
 	  {
-	    if (pos - markPos < readAheadLimit
+	    if (pos - markPos > readAheadLimit
 		|| ((flags & USER_BUFFER) != 0
 		    && reserve - markPos > buffer.length))
 	      clearMark();
@@ -292,35 +302,69 @@ public class LineBufferedReader extends FilterReader
 
   public int read (char[] cbuf, int off, int len) throws java.io.IOException
   {
-    if (len <= 0)
-      return len;
-    int c;
-    int i = off;
+    // Same logic as in skip(n), when n>0.
+    int ch;
     if (pos >= limit)
+      ch = '\0';
+    else if (pos > 0)
+      ch = buffer[pos-1];
+    else if ((flags & PREV_WAS_CR) != 0 || lineStartPos >= 0)
+      ch = '\n';
+    else
+      ch = '\0';
+    int to_do = len;
+    while (to_do > 0)
       {
-	c = read();
-	if (c < 0)
-	  return -1;
-	cbuf[i++] = (char) c;
-      }
-    while (--len > 0 && pos < limit)
-      {
-	c = buffer[pos];
-	if (c == '\r' || c == '\n')
-	  c = read();
+	if (pos >= limit || ch == '\n' || ch == '\r')
+	  {
+	    // Return if there is no more in the input buffer, and we got
+	    // at least one char.  This is desirable for interactive input.
+	    if (pos >= limit && to_do < len)
+	      return len - to_do;
+	    ch = read();
+	    if (ch < 0)
+	      {
+		len -= to_do;
+		return len <= 0 ? -1 : len;
+	      }
+	    cbuf[off++] = (char) ch;
+	    to_do--;
+	  }
 	else
-	  pos++;
-	cbuf[i++] = (char) c;
+	  {
+	    int p = pos;
+	    int lim = limit;
+	    if (to_do < lim - p)
+	      lim = p + to_do;
+	    while (p < lim)
+	      {
+		ch = buffer[p];
+		// For simplicity and correctness we defer handling of
+		// newlines (including previous character) to read().
+		if (ch == '\n' || ch == '\r')
+		  break;
+		cbuf[off++] = (char) ch;
+		p++;
+	      }
+	    to_do -= p - pos;
+	    pos = p;
+	  }
       }
-    return i - off;
+    return len;
   }
 
-  public String getName ()
+  /** Should return a URI or a String. */
+  public Object getURI ()
   {
     return name;
   }
 
-  public void setName (String name)
+  public String getName ()
+  {
+    return name == null ? null : name.toString();
+  }
+
+  public void setName (Object name)
   {
     this.name = name;
   }
@@ -349,6 +393,7 @@ public class LineBufferedReader extends FilterReader
     this.lineNumber += lineNumber - getLineNumber();
   }
 
+  /** Return the current (zero-based) column number. */
   public int getColumnNumber ()
   {
     if (pos > 0)
@@ -493,26 +538,46 @@ public class LineBufferedReader extends FilterReader
       }
     else
       {
+	// Same logic as in read(char[],int,int).
 	int to_do = n;
+	int ch;
+	if (pos >= limit)
+	  ch = '\0';
+	else if (pos > 0)
+	  ch = buffer[pos-1];
+	else if ((flags & PREV_WAS_CR) != 0 || lineStartPos >= 0)
+	  ch = '\n';
+	else
+	  ch = '\0';
 	while (to_do > 0)
 	  {
-	    int ch;
-	    int i = pos;
-	    int count = limit - i;
-	    if (count > to_do)
-	      count = to_do;
-	    while (--count >= 0 
-		   && (ch = buffer[i]) != '\r' && ch != '\n')
-	      i++;
-	    to_do -= i - pos;
-	    if (to_do <= 0)
-	      break;
-	    ch = read();
-	    if (ch < 0)
-	      break;
-	    to_do--;
+	    if (ch == '\n' || ch == '\r' || pos >= limit)
+	      {
+		ch = read();
+		if (ch < 0)
+		  return n - to_do;
+		to_do--;
+	      }
+	    else
+	      {
+		int p = pos;
+		int lim = limit;
+		if (to_do < lim - p)
+		  lim = p + to_do;
+		while (p < lim)
+		  {
+		    ch = buffer[p];
+		    // For simplicity and correctness we defer handling of
+		    // newlines (including previous character) to read().
+		    if (ch == '\n' || ch == '\r')
+		      break;
+		    p++;
+		  }
+		to_do -= p - pos;
+		pos = p;
+	      }
 	  }
-	return n - to_do;
+	return n;
       }
   }
 
@@ -532,13 +597,13 @@ public class LineBufferedReader extends FilterReader
     read();
   }
 
-  static int countLines (char[] buffer, int off, int len)
+  static int countLines (char[] buffer, int start, int limit)
   {
     int count = 0;
     char prev = '\0';
-    for (int i = 0;  i < len;  i++)
+    for (int i = start;  i < limit;  i++)
       {
-	char ch = buffer[i+off];
+	char ch = buffer[i];
 	if ((ch == '\n' && prev != '\r') || ch == '\r')
 	  count++;
 	prev = ch;
@@ -546,6 +611,27 @@ public class LineBufferedReader extends FilterReader
     return count;
   }
 
+  /** Skips the rest of the current line, including the line terminator. */
+  public void skipRestOfLine ()
+       throws java.io.IOException
+  {
+    for (;;)
+      {
+        int c = read();
+        if (c < 0)
+          return;
+        if (c == '\r')
+          {
+            c = read();
+            if (c >= 0 && c != '\n')
+              unread();
+            break;
+          }
+        else if (c == '\n')
+          break;
+      }
+  }
+
   /* Move one character backwards. */
   public void unread ()
        throws java.io.IOException
