@@ -3,6 +3,7 @@ package gnu.text;
 /** Represents an error message from processing a "source" file. */
 
 public class SourceError
+  implements SourceLocator
 {
   /** Used to chain to the "next" message. */
   public SourceError next;
@@ -11,9 +12,12 @@ public class SourceError
    * 'e' (for error), or 'f' (for fatal error). */
   public char severity;
 
-  /* The name or URL of the file containing the error. */
+  /** The name or URL of the file containing the error. */
   public String filename;
 
+  /** If non-null, an error code, as might be specified by a standard. */
+  public String code;
+
   /** The line number of the error, with 1 being the top line.
    * The value 0 means unknown or not applicable (such as the entire file). */
   /** The (1-origin) location of the error. */
@@ -28,6 +32,11 @@ public class SourceError
    * It can contain multiple lines, separated by '\n'.*/
   public String message;
 
+  /** Provides optional stack trace.
+   * Filled when --debug-error-prints-stack-trace or
+   * --debug-warning-prints-stack-trace option is used.*/
+  public Throwable fakeException;
+
   public SourceError(char severity, String filename, int line, int column, 
 		     String message)
   {
@@ -38,6 +47,12 @@ public class SourceError
     this.message = message;
   }
 
+  public SourceError(char severity, SourceLocator location, String message)
+  {
+    this(severity, location.getFileName(), location.getLineNumber(),
+         location.getColumnNumber(), message);
+  }
+
   /** Create a new SourceError using the current line/column from
    * a <code>LineBufferedReader</code>. */
   public SourceError(LineBufferedReader port, char severity, String message)
@@ -56,7 +71,7 @@ public class SourceError
   {
     StringBuffer buffer = new StringBuffer ();
     buffer.append (filename == null ? "<unknown>" : filename);
-    if (line != 0 || column != 0)
+    if (line > 0 || column > 0)
       {
 	buffer.append (':');
 	buffer.append (line);
@@ -70,6 +85,33 @@ public class SourceError
     if (severity == 'w')
       buffer.append("warning - ");
     buffer.append (message);
+    if (code != null)
+      {
+        buffer.append(" [");
+        buffer.append(code);
+        buffer.append("]");
+      }
+    if (fakeException != null)
+      {
+        // We assume getStackTrace is evailable if getCause is,
+        // rather than add a new PreProcess parameter.
+        /* #ifdef use:java.lang.Throwable.getCause */
+        StackTraceElement[] stackTrace = fakeException.getStackTrace();
+        for (int i = 0; i < stackTrace.length; i++)
+          {
+            buffer.append("\n");
+            buffer.append("    ");
+            buffer.append(stackTrace[i].toString());
+          }
+        /* #else */
+        // java.io.StringWriter writer = new java.io.StringWriter();
+        // java.io.PrintWriter pwriter = new java.io.PrintWriter(writer);
+        // fakeException.printStackTrace(pwriter);
+        // pwriter.close();
+        // buffer.append("\n");
+        // buffer.append(writer.toString());
+        /* #endif */
+      }
     return buffer.toString ();
   }
 
@@ -105,4 +147,11 @@ public class SourceError
       }
     out.println(line);
   }
+
+  public int getLineNumber () { return line == 0 ? -1 : line; }
+  public int getColumnNumber () { return column == 0 ? -1 : column; }
+  public String getPublicId() { return null; }
+  public String getSystemId() { return filename; }
+  public String getFileName() { return filename; }
+  public boolean isStableSourceLocation() { return true; }
 }
