@@ -11,28 +11,33 @@ import gnu.bytecode.*;
 import gnu.kawa.lispexpr.LangPrimType;
 import gnu.xquery.util.*;
 import gnu.xml.*;
-import gnu.kawa.reflect.ClassMethodProc;
 import gnu.text.Lexer;
 import gnu.text.SourceMessages;
 import java.io.Reader;
 import java.util.Vector;
 import gnu.kawa.functions.ConstantFunction0;
+import gnu.kawa.reflect.ClassMethods;
 import gnu.math.IntNum;
+import gnu.kawa.xml.*;
 
 /** The XQuery language. */
 
 public class XQuery extends Language
 {
   public static final String XQUERY_FUNCTION_NAMESPACE
-    = "http://www.w3.org/2004/10/xpath-functions";
+    = "http://www.w3.org/2005/xpath-functions";
   public static final String KAWA_FUNCTION_NAMESPACE
     = "http://kawa.gnu.org/";
   public static final String QEXO_FUNCTION_NAMESPACE
     = "http://qexo.gnu.org/";
   public static final String LOCAL_NAMESPACE
-    = "http://www.w3.org/2004/10/xquery-local-functions";
+    = "http://www.w3.org/2005/xquery-local-functions";
   public static final String SCHEMA_NAMESPACE
     = "http://www.w3.org/2001/XMLSchema";
+  public static final String SCHEMA_INSTANCE_NAMESPACE
+    = "http://www.w3.org/2001/XMLSchema-instance";
+  public static final String XHTML_NAMESPACE
+    = "http://www.w3.org/1999/xhtml";
   public static final Namespace xqueryFunctionNamespace
     = Namespace.getInstance(XQUERY_FUNCTION_NAMESPACE);
   public static final Namespace kawaFunctionNamespace
@@ -47,9 +52,9 @@ public class XQuery extends Language
   static boolean charIsInt = false;
 
   /** Pseudo-namespace "prefix" for the default element namespace. */
-  public static final String DEFAULT_ELEMENT_PREFIX = "elements$";
+  public static final String DEFAULT_ELEMENT_PREFIX = null;
   /** Pseudo-namespace "prefix" for the default function namespace. */
-  public static final String DEFAULT_FUNCTION_PREFIX = "functions$";
+  public static final String DEFAULT_FUNCTION_PREFIX = "(functions)";
 
   Namespace defaultNamespace;
 
@@ -89,30 +94,26 @@ public class XQuery extends Language
     return new XQParser(inp, messages, this);
   }
 
+  public Compilation getCompilation (Lexer lexer, SourceMessages messages)
+  {
+    return new Compilation(this, messages, ((XQParser) lexer).lexical);
+  }
+
   /** Special parser flag used by <code>evalToFocusProc</code>. */
   public static final int PARSE_WITH_FOCUS = 0x10000;
 
-  public Compilation parse(Lexer lexer, int options)
+  public boolean parse (Compilation tr, int options)
     throws java.io.IOException, gnu.text.SyntaxException
   {
-    XQParser parser = (XQParser) lexer;
+    ModuleExp mexp = tr.mainLambda;
     Compilation.defaultCallConvention = Compilation.CALL_WITH_CONSUMER;
-    Compilation tr = new Compilation(this, parser.getMessages(),
-				     parser.lexical);
-    tr.immediate = (options & PARSE_IMMEDIATE) != 0;
-    XQResolveNames resolver = new XQResolveNames(tr);
-    resolver.functionNamespacePath = parser.functionNamespacePath;
-    ModuleExp mexp = new ModuleExp();
-    mexp.setFile(lexer.getName());
-    tr.push(mexp);
     tr.mustCompileHere();
-    ((XQParser) lexer).resolver = resolver;
-    resolver.parser =  (XQParser) lexer;
-    if ((options & PARSE_ONE_LINE) != 0)
+    XQParser parser = (XQParser) tr.lexer;
+    if (parser.isInteractive())
       {
-	Expression sexp = ((XQParser) lexer).parse(tr);
+	Expression sexp = parser.parse(tr);
 	if (sexp == null)
-	  return null;
+	  return false;
 	mexp.body = sexp;
       }
     else if ((options & PARSE_WITH_FOCUS) != 0)
@@ -124,18 +125,38 @@ public class XQuery extends Language
 	lexp.addDeclaration(XQParser.POSITION_VARNAME, Type.int_type);
 	lexp.addDeclaration(XQParser.LAST_VARNAME, Type.int_type);
 	tr.push(lexp);
-	lexp.body = ((XQParser) lexer).parse(tr);
+	lexp.body = parser.parse(tr);
 	tr.pop(lexp);
 	mexp.body = lexp;
       }
     else
       {
 	Vector exps = new Vector(10);
+        Expression sexp = mexp.body;
+        if (sexp instanceof BeginExp)
+          {
+            BeginExp bexp = (BeginExp) sexp;
+            int blen = bexp.getExpressionCount();
+            Expression[] bexps = bexp.getExpressions();
+            for (int i = 0;  i < blen;  i++)
+              exps.addElement(bexps[i]);
+          }
+        else if (sexp != null && sexp != QuoteExp.voidExp)
+          {
+            exps.addElement(sexp);
+          }
 	for (;;)
 	  {
-	    Expression sexp = ((XQParser) lexer).parse(tr);
+	    sexp = parser.parse(tr);
 	    if (sexp == null)
-	      break;
+              {
+                if (parser.parseCount == 0 && !parser.isInteractive())
+                  {
+                    parser.error('e', "empty module", "XPST0003");
+                    return false;
+                  }
+                break;
+              }
 	    exps.addElement(sexp);
 	  }
 	int nexps = exps.size();
@@ -150,8 +171,11 @@ public class XQuery extends Language
 	    mexp.body = new BeginExp(arr);
 	  }
       }
+    // It seems silly to pop all the declarations, and then push them
+    // in resolveModule.  The complication is that imported variable
+    // declarations are pushed eagerly in the first pass.
     tr.pop(mexp);
-    
+
     if (false)
       {
 	OutPort dout = OutPort.outDefault();
@@ -161,15 +185,67 @@ public class XQuery extends Language
 	dout.flush();
       }
 
-    resolver.resolveModule(mexp);
-    return tr;
+    XQResolveNames resolver = new XQResolveNames(tr);
+    resolver.functionNamespacePath = parser.functionNamespacePath;
+    resolver.parser = parser;
+    resolver.resolveModule(mexp); // FIXME should move to resolve(Compilation)
+    tr.setState(Compilation.BODY_PARSED);
+    return true;
   }
 
+  public void resolve (Compilation comp)
+  {
+  }
+
+  public static int namespaceForFunctions (int argCount)
+  {
+    return (argCount << 2) | FUNCTION_NAMESPACE;
+  }
+
+  public static final int VARIADIC_FUNCTION_NAMESPACE =
+    (-1 << 2) | FUNCTION_NAMESPACE;
+
   public int getNamespaceOf(Declaration decl)
   {
-    return decl.isProcedureDecl() ? FUNCTION_NAMESPACE
-      //: decl.isNamespaceDecl() ? NAMESPACE_PREFIX_NAMESPACE
-      : VALUE_NAMESPACE;
+    if (decl.isProcedureDecl())
+      {
+        if (decl.getCode() < 0)
+          return VARIADIC_FUNCTION_NAMESPACE;
+        Expression value = decl.getValue();
+        if (value instanceof LambdaExp)
+          {
+            LambdaExp lexp = (LambdaExp) value;
+            if (lexp.min_args == lexp.max_args)
+              return namespaceForFunctions(lexp.min_args);
+          }
+        else if (value instanceof QuoteExp)
+          {
+            Object val = ((QuoteExp) value).getValue();
+            if (val instanceof Procedure)
+              {
+                Procedure proc = (Procedure) val;
+                int min = proc.minArgs();
+                int max = proc.maxArgs();
+                if (min == max)
+                  return namespaceForFunctions(min);
+              }
+          }
+        else if (value instanceof ReferenceExp)
+          return getNamespaceOf(((ReferenceExp) value).getBinding());
+        // I believe we only get here after an error.
+        return VARIADIC_FUNCTION_NAMESPACE;
+      }
+    return VALUE_NAMESPACE;
+  }
+
+  public boolean hasNamespace (Declaration decl, int namespace)
+  {
+    int dnspace = getNamespaceOf(decl);
+    return (dnspace == namespace
+            || (dnspace == VARIADIC_FUNCTION_NAMESPACE
+                && (namespace & FUNCTION_NAMESPACE) != 0)
+            || (namespace == VARIADIC_FUNCTION_NAMESPACE
+                && (dnspace & FUNCTION_NAMESPACE) != 0));
   }
 
   public Symbol getSymbol (String name)
@@ -187,7 +263,10 @@ public class XQuery extends Language
   protected void define_method(String name, String cname, String mname)
   {
     Symbol sym = Symbol.make(defaultNamespace, name);
-    Procedure proc = ClassMethodProc.make(ClassType.make(cname), mname);
+    // This does require eager loading of the class, which takes
+    // extra time on startup.  FIXME.
+    ClassType ctype = ClassType.make(cname);
+    Procedure proc = ClassMethods.apply(ctype, mname, '\0', this);
     proc.setSymbol(sym);
     environ.define(sym, EnvironmentKey.FUNCTION, proc);
   }
@@ -376,7 +455,7 @@ public class XQuery extends Language
     int oldIndex = ctx.startFromContext();
     try
       {
-	ModuleExp.evalModule(Environment.getCurrent(), ctx, comp);
+	ModuleExp.evalModule(Environment.getCurrent(), ctx, comp, null, null);
 	return (Procedure) ctx.getFromContext(oldIndex);
       }
     catch (Throwable ex)
@@ -423,7 +502,7 @@ public class XQuery extends Language
 			       Object item, int position, int size)
     throws Throwable
   {
-   return applyWithFocus(evalToFocusProc(expr), item, position, size);
+    return applyWithFocus(evalToFocusProc(expr), item, position, size);
   }
 
   /** Evaluate an expression with a given focus (context).
@@ -474,7 +553,9 @@ public class XQuery extends Language
   public static final Environment xqEnvironment
     = Environment.make(XQUERY_FUNCTION_NAMESPACE);
 
-  static final XQuery instance = new XQuery();
+  // This field need to be public so that the findLiteral method in
+  // gnu.expr.LitTable can find it.
+  public static final XQuery instance = new XQuery();
   static { instance.initXQuery(); }
 
   public XQuery()
@@ -511,8 +592,8 @@ public class XQuery extends Language
       }
     catch (Throwable ex)
       {
-	// Ignore.  We get a ClassNotFoundException if gnu.kawa.slib.HTTP
-	// was not built.  We get a NoClassDefFoundError if gnu.kawa.slib.HTTP
+	// Ignore.  We get a ClassNotFoundException if gnu.kawa.servlet.HTTP
+	// was not built.  We get a NoClassDefFoundError if gnu.kawa.servlet.HTTP
 	// can't find servlets in the classpath.
       }
     finally
@@ -524,50 +605,80 @@ public class XQuery extends Language
     defProcStFld("unescaped-data", "gnu.kawa.xml.MakeUnescapedData", "unescapedData");
     defProcStFld("item-at", "gnu.xquery.util.ItemAt", "itemAt");
     defProcStFld("count", "gnu.kawa.functions.CountValues", "countValues");
-    defProcStFld("min", "gnu.xquery.util.MinMax", "min");
-    defProcStFld("max", "gnu.xquery.util.MinMax", "max");
-    defProcStFld("sum", "gnu.xquery.util.Reduce", "sum");
+    define_method("sum", "gnu.xquery.util.Reduce", "sum"); // Overloaded
     defProcStFld("avg", "gnu.xquery.util.Average", "avg");
-    defProcStFld("index-of", "gnu.xquery.util.IndexOf", "indexOf");
-    defProcStFld("last-index-of", "gnu.xquery.util.LastIndexOf", "lastIndexOf");
-    defProcStFld("sublist", "gnu.xquery.util.SubList", "subList");
+    defProcStFld("sublist", "gnu.xquery.util.SubList", "subList"); // deprecated
+    defProcStFld("subsequence", "gnu.xquery.util.SubList", "subList");
     define_method("empty", "gnu.xquery.util.SequenceUtils",
 		  "isEmptySequence");
     define_method("exists", "gnu.xquery.util.SequenceUtils",
 		  "exists");
+    define_method("insert-before", "gnu.xquery.util.SequenceUtils",
+		  "insertBefore$X");
+    define_method("remove", "gnu.xquery.util.SequenceUtils",
+		  "remove$X");
     define_method("reverse", "gnu.xquery.util.SequenceUtils",
-		  "reverse");
+		  "reverse$X");
     defProcStFld("false", "gnu.xquery.lang.XQuery", "falseFunction");
     defProcStFld("true", "gnu.xquery.lang.XQuery", "trueFunction");
-    defProcStFld("number", "gnu.xquery.util.NumberValue", "numberValue");
-    defProcStFld("string-value", "gnu.xquery.util.StringValue", "stringValue");
-    defProcStFld("string", "gnu.xquery.util.StringValue", "string");
+    defProcStFld("boolean", "gnu.xquery.util.BooleanValue", "booleanValue");
 
     define_method("trace", "gnu.xquery.util.Debug", "trace");
+    define_method("error", "gnu.xquery.util.XQException",
+                  "error"); // overloaded
     defProcStFld("write-to", "gnu.kawa.xml.WriteTo", "writeTo");
+    defProcStFld("write-to-if-changed", "gnu.kawa.xml.WriteTo",
+                 "writeToIfChanged");
     defProcStFld("iterator-items",
 		 "gnu.kawa.xml.IteratorItems", "iteratorItems");
     defProcStFld("list-items", "gnu.kawa.xml.ListItems", "listItems");
-    defProcStFld("base-uri", "gnu.kawa.functions.BaseUri", "baseUri");
     define_method("node-name", "gnu.kawa.xml.NodeName", "nodeName");
-    define_method("lower-case", "gnu.xquery.util.StringValue", "lowerCase");
-    define_method("upper-case", "gnu.xquery.util.StringValue", "upperCase");
-    define_method("substring", "gnu.xquery.util.StringValue", "substring");
+    define_method("nilled", "gnu.xquery.util.NodeUtils", "nilled");
+    define_method("data", "gnu.xquery.util.NodeUtils", "data$X");
+    define_method("lower-case", "gnu.xquery.util.StringUtils", "lowerCase");
+    define_method("upper-case", "gnu.xquery.util.StringUtils", "upperCase");
+    define_method("substring", "gnu.xquery.util.StringUtils", "substring");
     define_method("string-length",
-		  "gnu.xquery.util.StringValue", "stringLength");
+		  "gnu.xquery.util.StringUtils", "stringLength");
     define_method("substring-before",
-		  "gnu.xquery.util.StringValue", "substringBefore");
+		  "gnu.xquery.util.StringUtils", "substringBefore");
     define_method("substring-after",
-		  "gnu.xquery.util.StringValue", "substringAfter");
-    define_method("translate", "gnu.xquery.util.StringValue", "translate");
-    define_method("string-pad", "gnu.xquery.util.StringValue", "stringPad");
-    define_method("contains", "gnu.xquery.util.StringValue", "contains");
-    define_method("starts-with", "gnu.xquery.util.StringValue", "startsWith");
-    define_method("ends-with","gnu.xquery.util.StringValue", "endsWith");
-    define_method("string-join", "gnu.xquery.util.StringValue", "stringJoin");
-    define_method("concat", "gnu.xquery.util.StringValue", "concat");
+		  "gnu.xquery.util.StringUtils", "substringAfter");
+    define_method("translate", "gnu.xquery.util.StringUtils", "translate");
+    define_method("encode-for-uri", "gnu.xquery.util.StringUtils",
+                  "encodeForUri");
+    define_method("iri-to-uri", "gnu.xquery.util.StringUtils", "iriToUri");
+    define_method("escape-html-uri", "gnu.xquery.util.StringUtils",
+                  "escapeHtmlUri");
+    // Non-standard (in F&O example appendix).  Put in qexo namespace?
+    // define_method("string-pad", "gnu.xquery.util.StringUtils", "stringPad");
+    define_method("contains", "gnu.xquery.util.StringUtils", "contains");
+    define_method("starts-with", "gnu.xquery.util.StringUtils", "startsWith");
+    define_method("ends-with","gnu.xquery.util.StringUtils", "endsWith");
+    define_method("codepoint-equal", "gnu.xquery.util.StringUtils",
+                  "codepointEqual");
+    define_method("normalize-unicode", "gnu.xquery.util.StringUtils",
+                  "normalizeUnicode");
+    define_method("string-join", "gnu.xquery.util.StringUtils", "stringJoin");
+    define_method("concat", "gnu.xquery.util.StringUtils", "concat$V");
+    define_method("matches", "gnu.xquery.util.StringUtils", "matches");
+    define_method("replace", "gnu.xquery.util.StringUtils", "replace");
+    define_method("tokenize", "gnu.xquery.util.StringUtils", "tokenize$X");
+    define_method("string-to-codepoints", "gnu.xquery.util.StringUtils",
+		  "stringToCodepoints$X");
+    define_method("codepoints-to-string", "gnu.xquery.util.StringUtils",
+		  "codepointsToString");
+
+    define_method("abs", "gnu.xquery.util.NumberValue", "abs");
+    define_method("floor", "gnu.xquery.util.NumberValue", "floor");
+    define_method("ceiling", "gnu.xquery.util.NumberValue", "ceiling");
+    define_method("round", "gnu.xquery.util.NumberValue", "round");
+    define_method("round-half-to-even", "gnu.xquery.util.NumberValue",
+                  "roundHalfToEven");
 
     define_method("QName", "gnu.xquery.util.QNameUtils", "makeQName");
+    define_method("resolve-QName", "gnu.xquery.util.QNameUtils",
+                  "resolveQNameUsingElement");
     define_method("prefix-from-QName", "gnu.xquery.util.QNameUtils",
 		  "prefixFromQName");
     define_method("local-name-from-QName", "gnu.xquery.util.QNameUtils",
@@ -576,33 +687,90 @@ public class XQuery extends Language
 		  "namespaceURIFromQName");
     define_method("namespace-uri-for-prefix", "gnu.xquery.util.QNameUtils",
 		  "namespaceURIForPrefix");
+    define_method("in-scope-prefixes", "gnu.xquery.util.NodeUtils",
+                  "inScopePrefixes$X");
+    define_method("document-uri", "gnu.xquery.util.NodeUtils", "documentUri");
+
+    define_method("years-from-duration", "gnu.xquery.util.TimeUtils",
+                  "yearsFromDuration");
+    define_method("months-from-duration", "gnu.xquery.util.TimeUtils",
+                  "monthsFromDuration");
+    define_method("days-from-duration", "gnu.xquery.util.TimeUtils",
+                  "daysFromDuration");
+    define_method("hours-from-duration", "gnu.xquery.util.TimeUtils",
+                  "hoursFromDuration");
+    define_method("minutes-from-duration", "gnu.xquery.util.TimeUtils",
+                  "minutesFromDuration");
+    define_method("seconds-from-duration", "gnu.xquery.util.TimeUtils",
+                  "secondsFromDuration");
+    define_method("year-from-dateTime", "gnu.xquery.util.TimeUtils",
+                  "yearFromDateTime");
+    define_method("month-from-dateTime", "gnu.xquery.util.TimeUtils",
+                  "monthFromDateTime");
+    define_method("day-from-dateTime", "gnu.xquery.util.TimeUtils",
+                  "dayFromDateTime");
+    define_method("hours-from-dateTime", "gnu.xquery.util.TimeUtils",
+                  "hoursFromDateTime");
+    define_method("minutes-from-dateTime", "gnu.xquery.util.TimeUtils",
+                  "minutesFromDateTime");
+    define_method("seconds-from-dateTime", "gnu.xquery.util.TimeUtils",
+                  "secondsFromDateTime");
+    define_method("timezone-from-dateTime", "gnu.xquery.util.TimeUtils",
+                  "timezoneFromDateTime");
+    define_method("year-from-date", "gnu.xquery.util.TimeUtils",
+                  "yearFromDate");
+    define_method("month-from-date", "gnu.xquery.util.TimeUtils",
+                  "monthFromDate");
+    define_method("day-from-date", "gnu.xquery.util.TimeUtils",
+                  "dayFromDate");
+    define_method("timezone-from-date", "gnu.xquery.util.TimeUtils",
+                  "timezoneFromDate");
+    define_method("hours-from-time", "gnu.xquery.util.TimeUtils",
+                  "hoursFromTime");
+    define_method("minutes-from-time", "gnu.xquery.util.TimeUtils",
+                  "minutesFromTime");
+    define_method("seconds-from-time", "gnu.xquery.util.TimeUtils",
+                  "secondsFromTime");
+    define_method("timezone-from-time", "gnu.xquery.util.TimeUtils",
+                  "timezoneFromTime");
+    define_method("adjust-dateTime-to-timezone", "gnu.xquery.util.TimeUtils",
+                  "adjustDateTimeToTimezone"); // overloaded
+    define_method("adjust-date-to-timezone", "gnu.xquery.util.TimeUtils",
+                  "adjustDateToTimezone"); // overloaded
+    define_method("adjust-time-to-timezone", "gnu.xquery.util.TimeUtils",
+                  "adjustTimeToTimezone"); // overloaded
+    define_method("dateTime", "gnu.xquery.util.TimeUtils", "dateTime");
+
+    define_method("zero-or-one", "gnu.xquery.util.SequenceUtils", "zeroOrOne");
+    define_method("one-or-more", "gnu.xquery.util.SequenceUtils", "oneOrMore");
+    define_method("exactly-one", "gnu.xquery.util.SequenceUtils", "exactlyOne");
 
     defProcStFld("distinct-nodes", "gnu.kawa.xml.SortNodes", "sortNodes");
 
     // FIXME - should be imported?
     defProcStFld("children", "gnu.kawa.xml.Children", "children");
-    defProcStFld("not", "kawa.standard.Scheme");
+    define_method("not", "gnu.xquery.util.BooleanValue", "not");
 
     defaultNamespace = qexoFunctionNamespace;
-    defProcStFld("response-header", "gnu.kawa.slib.HTTP");
-    defProcStFld("response-content-type", "gnu.kawa.slib.HTTP");
-    defProcStFld("response-status", "gnu.kawa.slib.HTTP");
-    defProcStFld("error-response", "gnu.kawa.slib.HTTP");
-    defProcStFld("current-servlet", "gnu.kawa.slib.HTTP");
-    defProcStFld("current-servlet-context", "gnu.kawa.slib.HTTP");
-    defProcStFld("current-servlet-config", "gnu.kawa.slib.HTTP");
-    defProcStFld("servlet-context-realpath", "gnu.kawa.slib.HTTP");
-    defProcStFld("get-response", "gnu.kawa.slib.HTTP");
-    defProcStFld("get-request", "gnu.kawa.slib.HTTP");
-    defProcStFld("request-method", "gnu.kawa.slib.HTTP");
-    defProcStFld("request-uri", "gnu.kawa.slib.HTTP");
-    defProcStFld("request-url", "gnu.kawa.slib.HTTP");
-    defProcStFld("request-path-info", "gnu.kawa.slib.HTTP");
-    defProcStFld("request-path-translated", "gnu.kawa.slib.HTTP");
-    defProcStFld("request-servlet-path", "gnu.kawa.slib.HTTP");
-    defProcStFld("request-query-string", "gnu.kawa.slib.HTTP");
-    defProcStFld("request-parameter", "gnu.kawa.slib.HTTP");
-    defProcStFld("request-parameters", "gnu.kawa.slib.HTTP");
+    defProcStFld("response-header", "gnu.kawa.servlet.HTTP");
+    defProcStFld("response-content-type", "gnu.kawa.servlet.HTTP");
+    defProcStFld("response-status", "gnu.kawa.servlet.HTTP");
+    defProcStFld("error-response", "gnu.kawa.servlet.HTTP");
+    defProcStFld("current-servlet", "gnu.kawa.servlet.HTTP");
+    defProcStFld("current-servlet-context", "gnu.kawa.servlet.HTTP");
+    defProcStFld("current-servlet-config", "gnu.kawa.servlet.HTTP");
+    defProcStFld("servlet-context-realpath", "gnu.kawa.servlet.HTTP");
+    defProcStFld("get-response", "gnu.kawa.servlet.HTTP");
+    defProcStFld("get-request", "gnu.kawa.servlet.HTTP");
+    defProcStFld("request-method", "gnu.kawa.servlet.HTTP");
+    defProcStFld("request-uri", "gnu.kawa.servlet.HTTP");
+    defProcStFld("request-url", "gnu.kawa.servlet.HTTP");
+    defProcStFld("request-path-info", "gnu.kawa.servlet.HTTP");
+    defProcStFld("request-path-translated", "gnu.kawa.servlet.HTTP");
+    defProcStFld("request-servlet-path", "gnu.kawa.servlet.HTTP");
+    defProcStFld("request-query-string", "gnu.kawa.servlet.HTTP");
+    defProcStFld("request-parameter", "gnu.kawa.servlet.HTTP");
+    defProcStFld("request-parameters", "gnu.kawa.servlet.HTTP");
     defaultNamespace = xqueryFunctionNamespace;
   }
 
@@ -617,45 +785,65 @@ public class XQuery extends Language
     Language.setDefaults(new XQuery());
   }
 
+  static public QuoteExp falseExp =
+    new QuoteExp(Boolean.FALSE, XDataType.booleanType);
+  static public QuoteExp trueExp =
+    new QuoteExp(Boolean.TRUE, XDataType.booleanType);
+
   public static final ConstantFunction0 falseFunction
-    = new ConstantFunction0("false", Boolean.FALSE);
+    = new ConstantFunction0("false", falseExp);
   public static final ConstantFunction0 trueFunction
-    = new ConstantFunction0("true", Boolean.TRUE);
-
-  public static final XMLFormat writeFormat = new XMLFormat();
-
-  public AbstractFormat getFormat(boolean readable)
-  {
-    return writeFormat;
-  }
+    = new ConstantFunction0("true", trueExp);
 
   public Consumer getOutputConsumer(java.io.Writer out)
   {
     return new XMLPrinter(out, false);
   }
 
-  LangPrimType booleanType;
-
   static Object[] typeMap =
-    { "string", Type.string_type,
-      "boolean", Type.boolean_type,
-      "QName", "gnu.xml.SName",
-      "integer", "gnu.math.IntNum",
-      "positiveInteger", "gnu.math.IntNum",
-      "nonPositiveInteger", "gnu.math.IntNum",
-      "negativeInteger", "gnu.math.IntNum",
-      "nonNegativeInteger", "gnu.math.IntNum",
-      "decimal", "gnu.math.RealNum"
+    { "string", XDataType.stringType,
+      "untypedAtomic", XDataType.untypedAtomicType,
+      "boolean", XDataType.booleanType,
+      "integer", XIntegerType.integerType,
+      "long", XIntegerType.longType,
+      "int", XIntegerType.intType,
+      "short", XIntegerType.shortType,
+      "byte", XIntegerType.byteType,
+      "unsignedLong", XIntegerType.unsignedLongType,
+      "unsignedInt", XIntegerType.unsignedIntType,
+      "unsignedShort", XIntegerType.unsignedShortType,
+      "unsignedByte", XIntegerType.unsignedByteType,
+      "positiveInteger", XIntegerType.positiveIntegerType,
+      "nonPositiveInteger", XIntegerType.nonPositiveIntegerType,
+      "negativeInteger", XIntegerType.negativeIntegerType,
+      "nonNegativeInteger", XIntegerType.nonNegativeIntegerType,
+      "date", XTimeType.dateType,
+      "dateTime", XTimeType.dateTimeType,
+      "time", XTimeType.timeType,
+      "duration", XTimeType.durationType,
+      "yearMonthDuration", XTimeType.yearMonthDurationType,
+      "dayTimeDuration", XTimeType.dayTimeDurationType,
+      "gYearMonth", XTimeType.gYearMonthType,
+      "gYear", XTimeType.gYearType,
+      "gMonthDay", XTimeType.gMonthDayType,
+      "gDay", XTimeType.gDayType,
+      "gMonth", XTimeType.gMonthType,
+      "decimal", XDataType.decimalType,
+      "float", XDataType.floatType,
+      "double", XDataType.doubleType,
+      "anyURI", XDataType.anyURIType,
+      "hexBinary", XDataType.hexBinaryType,
+      "base64Binary", XDataType.base64BinaryType,
+      "NOTATION", XDataType.NotationType,
+      "QName", "gnu.mapping.Symbol",
+      "anyAtomicType", Type.pointer_type
     };
 
-  public Type getTypeFor(String name)
+  public static Type getStandardType (String name)
   {
-    if (name == "t")
-      name = "java.lang.Object";
-    String core = name.startsWith("xs:") ? name.substring(3) : name;
     for (int i = typeMap.length;  (i -= 2) >= 0; )
       {
-	if (typeMap[i].equals(core))
+	if (typeMap[i].equals(name))
 	  {
 	    Object t = typeMap[i+1];
 	    if (t instanceof String)
@@ -664,7 +852,16 @@ public class XQuery extends Language
 	      return (Type) t;
 	  }
       }
-    return Scheme.string2Type(name);
+    return null;
+  }
+
+  public Type getTypeFor(String name)
+  {
+    String core = name.startsWith("xs:") ? name.substring(3)
+      : name.startsWith("xdt:") ? name.substring(4)
+      : name;
+    Type t = getStandardType(core);
+    return t != null ? t : Scheme.string2Type(name);
   }
 
   public Type getTypeFor (Class clas)
@@ -673,13 +870,31 @@ public class XQuery extends Language
       {
 	String name = clas.getName();
 	if (name.equals("boolean"))
-	  {
-	    if (booleanType == null)
-	      booleanType = new LangPrimType(Type.boolean_type, this);
-	    return booleanType;
-	  }
+          return XDataType.booleanType;
 	return Scheme.getNamedType(name);
       }
+    else if (! clas.isArray())
+      {
+        String name = clas.getName();
+        if (name.equals("java.lang.String"))
+          return XDataType.stringType;
+        if (name.equals("gnu.kawa.xml.UntypedAtomic"))
+          return XDataType.untypedAtomicType;
+        if (name.equals("java.lang.Boolean"))
+          return XDataType.booleanType;
+        if (name.equals("java.lang.Float"))
+          return XDataType.floatType;
+        if (name.equals("java.lang.Double"))
+          return XDataType.doubleType;
+        if (name.equals("java.math.BigDecimal"))
+          return XDataType.decimalType;
+        if (name.equals("gnu.math.Duration"))
+          return XDataType.durationType;
+        /* #ifdef use:java.net.URI */
+        if (name.equals("java.net.URI"))
+          return  XDataType.anyURIType;
+        /* #endif */
+      }
     return Type.make(clas);
   }
 
@@ -769,24 +984,51 @@ public class XQuery extends Language
     return sbuf.toString();
   }
 
-  public static Object getExternal (SName name, Object type)
+  public static String makeClassName (String source)
+  {
+    source = source.replace(java.io.File.separatorChar, '/');
+    int sl = source.lastIndexOf('/');
+    if (sl >= 0)
+      source = source.substring(sl+1);
+    int dot = source.lastIndexOf('.');
+    if (dot >= 0)
+      source = source.substring(0, dot);
+    return Compilation.mangleNameIfNeeded(source);
+  }
+
+  public static Object getExternal (Symbol name, Object type)
   {
     Environment env = Environment.getCurrent();
-    Symbol symbol = name.getSymbol();
-    Object value = env.get(symbol, null, null);
+    Object value = env.get(name, null, null);
+    if (value == null)
+      value = env.get(Symbol.makeWithUnknownNamespace(name.getLocalName(),
+                                                      name.getPrefix()),
+                      null, null);
     if (value == null)
       throw new RuntimeException("unbound external "+name);
+    if (type == null)
+      return value;
+    if (type instanceof XDataType)
+      return ((XDataType) type).cast(value);
     if (type instanceof ClassType)
       {
         String cname = ((ClassType) type).getName();
         // KLUDGE - FIXME
         if ("gnu.math.IntNum".equals(cname))
-          value = IntNum.valueOf(value.toString());
-        else if ("gnu.math.RealNum".equals(cname))
-          value = gnu.math.DFloNum.make(Double.parseDouble(value.toString()));
-        else
-          throw new Error("cast to "+cname+" for external "
-                          +name+" not implemented");
+          return IntNum.valueOf(value.toString());
+        if ("gnu.math.RealNum".equals(cname))
+          return gnu.math.DFloNum.make(Double.parseDouble(value.toString()));
+      }
+    try
+      {
+        value = ((Type) type).coerceFromObject(value);
+      }
+    catch (ClassCastException ex)
+      {
+        throw new WrongType(name.toString(),
+                            WrongType.ARG_VARNAME,
+                            value,
+                            type.toString());
       }
     return value;
   }
