@@ -1,8 +1,5 @@
 package kawa;
 
-import kawa.lang.*;
-import gnu.kawa.lispexpr.LispReader;
-import kawa.standard.*;
 import gnu.mapping.*;
 import gnu.expr.*;
 import java.io.*;
@@ -17,7 +14,7 @@ public class Shell
   private static Class[] noClasses = { };
   private static  Class[] boolClasses = { Boolean.TYPE };
   private static  Class[] xmlPrinterClasses
-    = {gnu.lists.Consumer.class,  java.lang.Object.class };
+    = {gnu.mapping.OutPort.class,  java.lang.Object.class };
   private static  Class[] httpPrinterClasses
     = {gnu.mapping.OutPort.class };
   private static Object portArg = "(port)";
@@ -116,7 +113,7 @@ public class Shell
   }
 
   /** Return a Consumer that formats using the appropriate format.
-   * The format is choses depending on specified defaults.
+   * The format is chosen depending on specified defaults.
    * @param out The output where formatted output is sent to.
    */
   public static Consumer getOutputConsumer(OutPort out)
@@ -125,7 +122,7 @@ public class Shell
     if (out == null)
       return VoidConsumer.getInstance();
     else if (info == null)
-      return Interpreter.getInterpreter().getOutputConsumer(out);
+      return Language.getDefaultLanguage().getOutputConsumer(out);
     try
       {
 	Object args[] = new Object[info.length - 4];
@@ -134,9 +131,9 @@ public class Shell
 	  if (args[i] == portArg)
 	    args[i] = out;
 	Object format = defaultFormatMethod.invoke(null, args);
-	if (format instanceof FormatToConsumer)
+	if (format instanceof AbstractFormat)
 	  {
-	    out.objectFormat = (FormatToConsumer) format;
+	    out.objectFormat = (AbstractFormat) format;
 	    return out;
 	  }
 	else
@@ -149,35 +146,30 @@ public class Shell
       }
   }
 
-  public static void run (Interpreter interp)
-  {
-    run(interp, interp.getEnvironment());
-  }
-
-  public static void run (Interpreter interp, Environment env)
+  public static void run (Language language, Environment env)
   {
     InPort inp = InPort.inDefault ();
     if (inp instanceof TtyInPort)
       {
-	Procedure prompter = interp.getPrompter();
+	Procedure prompter = language.getPrompter();
 	if (prompter != null)
 	  ((TtyInPort)inp).setPrompter(prompter);
       }
 
-    run(interp, env, inp, OutPort.outDefault(), OutPort.errDefault());
+    run(language, env, inp, OutPort.outDefault(), OutPort.errDefault());
   }
 
-  public static void run (Interpreter interp,  Environment env,
+  public static void run (Language language,  Environment env,
 			  InPort inp, OutPort pout, OutPort perr)
   {
     Consumer out;
-    FormatToConsumer saveFormat = null;
+    AbstractFormat saveFormat = null;
     if (pout != null)
       saveFormat = pout.objectFormat;
     out = getOutputConsumer(pout);
     try
       {
-	run(interp, env, inp, out, perr);
+	run(language, env, inp, out, perr, null);
       }
     finally
       {
@@ -186,13 +178,16 @@ public class Shell
       }
   }
 
-  public static void run (Interpreter interp,  Environment env,
-			  InPort inp, Consumer out, OutPort perr)
+  public static void run (Language language,  Environment env,
+			  InPort inp, Consumer out, OutPort perr,
+                          java.net.URL url)
   {
     SourceMessages messages = new SourceMessages();
-    Lexer lexer = interp.getLexer(inp, messages);
-    if (inp instanceof TtyInPort)  // Wrong for the case of '-f' '-'.
-      lexer.setInteractive(true);
+    Language saveLanguage = Language.getDefaultLanguage();
+    Lexer lexer = language.getLexer(inp, messages);
+    // Wrong for the case of '-f' '-':
+    boolean interactive = inp instanceof TtyInPort;
+    lexer.setInteractive(interactive);
     CallContext ctx = CallContext.getInstance();
     Consumer saveConsumer = null;
     if (out != null)
@@ -200,20 +195,25 @@ public class Shell
 	saveConsumer = ctx.consumer;
 	ctx.consumer = out;
       }
+    if (language != saveLanguage)
+      Language.setDefaultLanguage(language);
     try
       {
 	for (;;)
 	  {
+	    int opts = Language.PARSE_IMMEDIATE|Language.PARSE_ONE_LINE;
 	    try
 	      {
-		ModuleExp mod = interp.parse(env, lexer);
-		if (mod == null) // end-of-file
+		Compilation comp = language.parse(lexer, opts, null);
+		boolean sawError = messages.checkErrors(perr, 20);
+		if (comp == null) // ??? end-of-file
 		  break;
-		mod.setName("atInteractiveLevel");  // FIXME
-		if (lexer.checkErrors(perr, 20))
+		if (sawError)
 		  continue;
+		comp.getModule().setName("atInteractiveLevel$"
+					 + (++ModuleExp.interactiveCounter));
 
-		// Skip whitespace, in case somebody calls (read-char) or similar.
+		// Skip whitespace, in case (read-char) or similar is called:
 		int ch;
 		for (;;)
 		  {
@@ -227,20 +227,23 @@ public class Shell
 		      }
 		  }
 
-
-		mod.evalModule(env, ctx);
-		ctx.runUntilDone();
+		if (! ModuleExp.evalModule(env, ctx, comp, url, perr))
+		  continue;
+                if (out instanceof Writer)
+                  ((Writer) out).flush();
 		if (ch < 0)
 		  break;
 	      }
 	    catch (WrongArguments e)
 	      {
+		messages.printAll(perr, 20);
 		if (e.usage != null)
 		  perr.println("usage: "+e.usage);
 		e.printStackTrace(perr);
 	      }
 	    catch (java.lang.ClassCastException e)
 	      {
+		messages.printAll(perr, 20);
 		perr.println("Invalid parameter, was: "+ e.getMessage());
 		e.printStackTrace(perr);
 	      }
@@ -248,10 +251,25 @@ public class Shell
 	      {
 		e.printAll(perr, 20);
 		e.clear();
+		if (! interactive)
+		  return;
+	      }
+	    catch (java.io.IOException e)
+	      {
+		messages.printAll(perr, 20);
+		String msg = new SourceError(inp, 'e', "").toString();
+		msg = msg.substring(0, msg.length() - 2);
+		perr.println(msg + " (or later): caught IOException");
+		e.printStackTrace(perr);
+		if (! interactive)
+		  return;
 	      }
 	    catch (Throwable e)
 	      {
+		messages.printAll(perr, 20);
 		e.printStackTrace(perr);
+		if (! interactive)
+		  return;
 	      }
 	  }
       }
@@ -259,25 +277,32 @@ public class Shell
       {
 	if (out != null)
 	  ctx.consumer = saveConsumer;
+        if (language != saveLanguage)
+          Language.setDefaultLanguage(saveLanguage);
       }
   }
 
-  public static void runString (String str, Interpreter interp, Environment env)
+  public static void runString (String str, Language language, Environment env)
   {
-    run(interp, env, new CharArrayInPort(str),
+    run(language, env, new CharArrayInPort(str),
 	ModuleBody.getMainPrintValues() ? OutPort.outDefault() : null,
 	OutPort.errDefault());
   }
 
-  public static void runFile (String fname)
+  public static void runFile (String fname, int skipLines)
   {
-    Environment env = Environment.user();
+    Environment env = Environment.getCurrent();
     try
       {
 	if (fname.equals ("-"))
-	  kawa.standard.load.loadSource(InPort.inDefault(), env);
+          {
+            InPort in = InPort.inDefault();
+            while (--skipLines >= 0)
+              in.skipRestOfLine();
+            kawa.standard.load.loadSource(in, env, null);
+          }
 	else
-	  kawa.standard.load.apply(fname,env);
+	  kawa.standard.load.apply(fname, env, false, skipLines);
       }
     catch (gnu.text.SyntaxException e)
       {
