@@ -1,7 +1,8 @@
 package kawa.lang;
 import gnu.mapping.*;
-import gnu.expr.*;
 import java.io.*;
+import gnu.text.*;
+import gnu.lists.Consumer;
 
 /**
  * A pattern that requires an exact match (using equal?).
@@ -18,17 +19,22 @@ public class EqualPat extends Pattern implements Printable, Externalizable
 
   static public EqualPat make (Object obj) { return new EqualPat (obj); }
 
-  public boolean match (Object obj, Object[] vars, int start_vars) {
+  public boolean match (Object obj, Object[] vars, int start_vars)
+  {
+    // We should be using Translator's matches routine, but the current
+    // Translator isn't available, so here is a special-purpose kludge.
+    if (value instanceof String && obj instanceof Symbol)
+      obj = ((Symbol) obj).getName();
     return value.equals (obj);
   }
 
   public int varCount () { return 0; }
 
-  public void print(java.io.PrintWriter ps)
+  public void print (Consumer out)
   {
-    ps.print ("#<equals: ");
-    SFormat.print (value, ps);
-    ps.print ('>');
+    out.write("#<equals: ");
+    ReportFormat.print(value, out);
+    out.write('>');
   }
 
   /**
