@@ -1,8 +1,8 @@
 package kawa.lang;
 import gnu.mapping.*;
-import gnu.expr.*;
 import gnu.lists.*;
 import java.io.*;
+import gnu.text.Printable;
 
 public class ListRepeatPat extends Pattern implements Printable, Externalizable
 {
@@ -22,11 +22,11 @@ public class ListRepeatPat extends Pattern implements Printable, Externalizable
     return new ListRepeatPat (element_pattern);
   }
 
-  public void print(java.io.PrintWriter ps)
+  public void print (Consumer out)
   {
-    ps.print ("#<list-repeat-pattern ");
-    SFormat.print (element_pattern, ps);
-    ps.print ('>');
+    out.write("#<list-repeat-pattern ");
+    element_pattern.print(out);
+    out.write('>');
   }
 
   public boolean match (Object obj, Object[] vars, int start_vars)
@@ -35,7 +35,7 @@ public class ListRepeatPat extends Pattern implements Printable, Externalizable
       System.err.print ("(match ");
       print (System.err);
       System.err.print (" on ");
-      SFormat.print (obj, System.err);
+      System.err.print(obj);
       System.err.print (")\n");
       */
     int length = LList.listLength(obj, false);
@@ -51,9 +51,9 @@ public class ListRepeatPat extends Pattern implements Printable, Externalizable
 	Pair pair = (Pair) obj;
 	/* DEBUGGING
 	   System.err.print ("(sub-match ["+j+"] ");
-	   SFormat.print (element_pattern, System.err);
+	   System.err.print(element_pattern);
 	   System.err.print (" on ");
-	   SFormat.print (pair.car, System.err);
+	   System.err.print(pair.car);
 	   */
 
 	if (! element_pattern.match (pair.car, element_vars, 0))
