@@ -1,8 +1,8 @@
 package kawa.lang;
 import gnu.mapping.*;
-import gnu.expr.*;
 import gnu.lists.*;
 import java.io.*;
+import gnu.text.Printable;
 
 public class PairPat extends Pattern implements Printable, Externalizable
 {
@@ -29,8 +29,6 @@ public class PairPat extends Pattern implements Printable, Externalizable
 
   public boolean match (Object obj, Object[] vars, int start_vars)
   {
-    if (obj instanceof SyntaxForm)
-      obj = ((SyntaxForm) obj).form;
     if (! (obj instanceof Pair))
       return false;
     Pair pair = (Pair) obj;
@@ -38,13 +36,13 @@ public class PairPat extends Pattern implements Printable, Externalizable
 	    && cdr.match (pair.cdr, vars, start_vars + car_count));
   }
 
-  public void print(java.io.PrintWriter ps)
+  public void print (Consumer out)
   {
-    ps.print ("#<pair-pattern car: ");
-    ps.print(car);
-    ps.print (" cdr: ");
-    ps.print(cdr);
-    ps.print ('>');
+    out.write("#<pair-pattern car: ");
+    car.print(out);
+    out.write(" cdr: ");
+    cdr.print(out);
+    out.write('>');
   }
 
   public int varCount () { return car_count + cdr_count; }
