@@ -7,7 +7,7 @@ import gnu.mapping.*;
 import gnu.lists.*;
 import gnu.expr.Compilation;
 
-public class Record extends NameMap
+public class Record
 {
   public String getTypeName()
   {
@@ -38,14 +38,29 @@ public class Record extends NameMap
     return hash;
   }
 
+  static java.lang.reflect.Field getField (Class clas, String fname)
+    throws NoSuchFieldException
+  {
+    ClassType ctype = (ClassType) Type.make(clas);
+    for (gnu.bytecode.Field fld = ctype.getFields();
+	 fld != null;  fld = fld.getNext())
+      {
+	if ((fld.getModifiers() & (Modifier.STATIC|Modifier.PUBLIC))
+	    != Modifier.PUBLIC)
+	  continue;
+	if (! fld.getSourceName().equals(fname))
+	  continue;
+	return fld.getReflectField();
+      }
+    throw new NoSuchFieldException();
+  }
+
   public Object get (String fname, Object defaultValue)
   {
     Class clas = getClass();
-    java.lang.reflect.Field fld;
     try
       {
-	fld = clas.getField (fname);
-	return fld.get(this);
+	return getField(clas, fname).get(this);
       }
     catch (NoSuchFieldException ex)
       {
@@ -66,17 +81,16 @@ public class Record extends NameMap
   public static Object set1 (Object record, String fname, Object value)
   {
     Class clas = record.getClass();
-    java.lang.reflect.Field fld;
     try
       {
-	fld = clas.getField (fname);
+	java.lang.reflect.Field fld = getField(clas, fname);
 	Object old = fld.get(record);
 	fld.set(record, value);
 	return old;
       }
     catch (NoSuchFieldException ex)
       {
-	//throw new UnboundSymbol(fname);
+	//throw new UnboundLocation(fname);
 	throw new GenericError("no such field "+fname+" in "+clas.getName());
       }
     catch (IllegalAccessException ex)
@@ -166,8 +180,7 @@ public class Record extends NameMap
     Method superConstructor
       = superClass.addMethod ("<init>", Type.typeArray0,
 			       Type.void_type, Access.PUBLIC);
-    constructor.init_param_slots ();
-    gnu.bytecode.CodeAttr code = constructor.getCode();
+    gnu.bytecode.CodeAttr code = constructor.startCode();
     code.emitPushThis();
     code.emitInvokeSpecial(superConstructor);
     code.emitReturn();
@@ -175,8 +188,7 @@ public class Record extends NameMap
       {
 	Method meth = clas.addMethod ("getTypeName", Type.typeArray0,
 				      Compilation.typeString, Access.PUBLIC);
-	meth.init_param_slots ();
-	code = meth.getCode();
+	code = meth.startCode();
 	code.emitPushString(name);
 	code.emitReturn();
       }
@@ -203,14 +215,7 @@ public class Record extends NameMap
     byte[][] arrays = new byte[1][];
     String[] names = new String[1];
     names[0] = mangledName;
-    try
-      {
-	arrays[0] = clas.writeToArray();
-      }
-    catch (java.io.IOException ex)
-      {
-	throw new InternalError (ex.toString());
-      }
+    arrays[0] = clas.writeToArray();
     ArrayClassLoader loader = new ArrayClassLoader(names, arrays);
     try
       {
