@@ -6,8 +6,10 @@ import gnu.bytecode.Type;
 import gnu.bytecode.ClassType;
 import gnu.text.SourceMessages;
 import gnu.lists.*;
-import gnu.kawa.lispexpr.LispLanguage;
+import gnu.kawa.lispexpr.*;
 import java.util.*;
+import gnu.kawa.functions.GetNamedPart;
+import gnu.text.SourceLocator;
 
 /** Used to translate from source to Expression.
  * The result has macros expanded, lexical names bound, etc, and is
@@ -30,6 +32,8 @@ public class Translator extends Compilation
    * from a <code>syntax-case</code>. */
   public PatternScope patternScope;
 
+  public Declaration templateScopeDecl;
+
   /** A variable to hold the matched values for syntax-case
    * pattern variables. */
   public Declaration matchArray;
@@ -38,6 +42,10 @@ public class Translator extends Compilation
   Stack renamedAliasStack;
 
   public Stack formStack = new Stack();
+  public int firstForm;
+  public Object pendingForm;
+
+  public LambdaExp curMethodLambda;
 
   /** Return true if decl is lexical and not fluid. */
   public boolean isLexical (Declaration decl)
@@ -152,8 +160,17 @@ public class Translator extends Compilation
   }
 
   /** True iff a form matches a literal symbol. */
-  public boolean matches(Object form, String literal)
+  public final boolean matches(Object form, String literal)
   {
+    return matches(form, null, literal);
+  }
+
+  public boolean matches(Object form, SyntaxForm syntax, String literal)
+  {
+    if (syntax != null)
+      {
+        // FIXME
+      }
     if (form instanceof SyntaxForm)
       {
 	// FIXME
@@ -171,17 +188,18 @@ public class Translator extends Compilation
   public Declaration lookup(Object name, int namespace)
   {
     Declaration decl = lexical.lookup(name, namespace);
-    if (decl != null
-	&& (getLanguage().getNamespaceOf(decl) & namespace) != 0)
+    if (decl != null && getLanguage().hasNamespace(decl, namespace))
       return decl;
-    return lookupGlobal(name, namespace);
+    return currentModule().lookup(name, getLanguage(), namespace);
   }
 
+  /** Find global Declaration, creating one if not found. */
   public Declaration lookupGlobal(Object name)
   {
     return lookupGlobal(name, -1);
   }
 
+  /** Find global Declaration, creating one if not found. */
   public Declaration lookupGlobal(Object name, int namespace)
   {
     ModuleExp module = currentModule();
@@ -214,7 +232,11 @@ public class Translator extends Compilation
                   = ((ReferenceExp) decl.getValue()).contextDecl();
                 if (context != null)
                   macroContext = context;
+                else if (current_scope instanceof TemplateScope)
+                  macroContext = ((TemplateScope) current_scope).macroContext;
               }
+            else if (current_scope instanceof TemplateScope)
+              macroContext = ((TemplateScope) current_scope).macroContext;
             Object obj = dval.eval(env);
             return obj instanceof Syntax ? (Syntax) obj : null;
           }
@@ -227,7 +249,7 @@ public class Translator extends Compilation
     return null;
   }
 
-  public Expression rewrite_pair (Pair p)
+  public Expression rewrite_pair (Pair p, boolean function)
   {
     if (p.car instanceof Syntax)
       return apply_rewrite((Syntax) p.car, p);
@@ -298,10 +320,11 @@ public class Translator extends Compilation
     if (cdr_length < 0)
       return syntaxError("dotted list ["+cdr+"] is not allowed after "+p.car);
 
-    Expression[] args = new Expression[cdr_length];
+    boolean mapKeywordsToAttributes = false;
+    Stack vec = new Stack();
 
     ScopeExp save_scope = current_scope;
-    for (int i = 0; i < cdr_length; i++)
+    for (int i = 0; i < cdr_length;)
       {
 	if (cdr instanceof SyntaxForm)
 	  {
@@ -310,15 +333,75 @@ public class Translator extends Compilation
 	    setCurrentScope(sf.scope);
 	  }
 	Pair cdr_pair = (Pair) cdr;
-	args[i] = rewrite_car (cdr_pair, false);
+	Expression arg = rewrite_car (cdr_pair, false);
+        i++;
+
+        if (mapKeywordsToAttributes)
+          {
+            if ((i & 1) == 0) // Previous iteration was a keyword
+              {
+                Expression[] aargs = new Expression[2];
+                aargs[0] = (Expression) vec.pop();
+                aargs[1] = arg;
+                arg = new ApplyExp(gnu.kawa.xml.MakeAttribute.makeAttribute, aargs);
+              }
+            else
+              {
+                Object value;
+                if (arg instanceof QuoteExp
+                    && (value = ((QuoteExp) arg).getValue()) instanceof Keyword
+                    && i < cdr_length)
+                  arg = new QuoteExp(((Keyword) value).asSymbol());
+                else
+                  mapKeywordsToAttributes = false;
+              }
+          }
+
+        vec.addElement(arg);
 	cdr = cdr_pair.cdr;
       }
+    Expression[] args = new Expression[vec.size()];
+    vec.copyInto(args);
+
     if (save_scope != current_scope)
       setCurrentScope(save_scope);
 
     return ((LispLanguage) getLanguage()).makeApply(func, args);
   }
 
+  public Symbol namespaceResolve (Expression context, Expression member)
+  {
+    if (context instanceof ReferenceExp && member instanceof QuoteExp)
+      {
+        ReferenceExp rexp = (ReferenceExp) context;
+        Declaration decl = rexp.getBinding();
+        Object val;
+        if (decl == null || decl.getFlag(Declaration.IS_UNKNOWN))
+          {
+            Object rsym = rexp.getSymbol();
+            Symbol sym = rsym instanceof Symbol ? (Symbol) rsym
+              : env.getSymbol(rsym.toString());
+            val = env.get(sym, null);
+          }
+        else if (decl.isNamespaceDecl())
+          {
+            val = decl.getConstantValue();
+          }
+        else
+          val = null;
+        if (val instanceof Namespace)
+          {
+            Namespace ns = (Namespace) val;
+            String uri = ns.getName();
+            if (uri != null && uri.startsWith("class:"))
+              return null; 
+            String mem = ((QuoteExp) member).getValue().toString().intern();
+            return ns.getSymbol(mem);
+          }
+      }
+    return null;
+  }
+
   public static Object stripSyntax (Object obj)
   {
     while (obj instanceof SyntaxForm)
@@ -417,123 +500,28 @@ public class Translator extends Compilation
     return rewrite(exp, false);
   }
 
-  public Object namespaceResolve (String str)
+  public Object namespaceResolve (Object name)
   {
-    int colon = str.indexOf(':');
-    if (colon <= 0 || colon >= str.length() - 1)
-      return str;
-    String prefix = str.substring(0, colon);
-    String local = str.substring(colon + 1);
-    String xprefix = (Language.NAMESPACE_PREFIX+prefix).intern();
-    Object uri_decl = lexical.lookup(xprefix, Language.VALUE_NAMESPACE);
-    if (uri_decl instanceof Declaration)
+    if (! (name instanceof String))
       {
-	Declaration decl = Declaration.followAliases((Declaration) uri_decl);
-	Expression dval = decl.getValue();
-	if (dval instanceof QuoteExp)
-	  {
-	    Object val = ((QuoteExp) dval).getValue();
-	    if (val instanceof String || val instanceof FString
-		|| (val instanceof Symbol
-		    && ((Symbol) val).hasEmptyNamespace()))
-	      return Symbol.make(val.toString(), local);
-	  }
-      }
-    return str;
-  }
-
-  private Expression namespaceResolve (String str, boolean function)
-  {
-    int colon = str.indexOf(':');
-    if (colon <= 0 || colon >= str.length() - 1)
-      return null;
-    String prefix = str.substring(0, colon);
-    String local = str.substring(colon + 1);
-    if ("*".equals(prefix))
-      return ClassMethodProc.makeExp(QuoteExp.nullExp, new QuoteExp(local));
-    String xprefix = (Language.NAMESPACE_PREFIX+prefix).intern();
-    Object uri_decl = lexical.lookup(xprefix, Language.VALUE_NAMESPACE);
-    Symbol sym;
-    if (uri_decl instanceof Declaration)
-      {
-	Declaration decl = Declaration.followAliases((Declaration) uri_decl);
-	Expression dval = decl.getValue();
-
-	if (dval instanceof ReferenceExp)
-	  {
-	    ReferenceExp ref = (ReferenceExp) dval;
-	    Declaration d = Declaration.followAliases(ref.getBinding());
-	    if (d != null)
-	      {
-		dval = d.getValue();
-		if (dval instanceof ClassExp)
-                  return ClassMethodProc.makeExp(ref, new QuoteExp(local));
-	      }
-	    else
-	      {
-		String name = ref.getName();
-		int nlen = name.length();
-		if (nlen > 2 && name.charAt(0) == '<'
-		    && name.charAt(nlen-1) == '>')
-                  return ClassMethodProc.makeExp(ref, new QuoteExp(local));
-	      }
-	  }
-
-	if (dval instanceof QuoteExp)
-	  {
-	    Object val = ((QuoteExp) dval).getValue();
-	    String uri;
-	    if (val instanceof ClassType)
-              return ClassMethodProc.makeExp(dval, new QuoteExp(local));
-	    else
-	      uri = val.toString();
-            if (uri.startsWith("class:"))
-              return ClassMethodProc.makeExp(rewrite("<"+uri.substring(6)+">"),
-                                             new QuoteExp(local));
-	    sym = Symbol.make(uri, local);
-	  }
-	else
-	  return null;
-      }
-    else
-      {
-	Object v = env.get(xprefix, null);
-        if (v instanceof ClassType)
-          return ClassMethodProc.makeExp(new QuoteExp(v),
-                                         new QuoteExp(local));
-        if (v != null)
+        Pair p;
+        if (name instanceof Pair
+            && safeCar(p = (Pair) name) == LispLanguage.lookup_sym
+            && p.cdr instanceof Pair
+            && (p = (Pair) p.cdr).cdr instanceof Pair)
           {
-            String uri = v.toString();
-            if (uri.startsWith("class:"))
-              return ClassMethodProc.makeExp(rewrite("<"+uri.substring(6)+">"),
-                                             new QuoteExp(local));
-            sym = Symbol.make(uri, local);
+            Expression part1 = rewrite(p.car);
+            Expression part2 = rewrite(((Pair) p.cdr).car);
+
+            Symbol sym = namespaceResolve(part1, part2);
+            if (sym != null)
+              return sym;
+            String combinedName = GetNamedPart.combineName(part1, part2);
+            if (combinedName != null)
+              return combinedName;
           }
-        else if (prefix.length() > 2 && prefix.charAt(0) == '<'
-                 && prefix.charAt(prefix.length()-1) == '>')
-          {
-            return ClassMethodProc.makeExp(rewrite(prefix), new QuoteExp(local));
-          }
-	else
-	  {
-            try
-              {
-                /* #ifdef JAVA2 */
-                Class cl = Class.forName(prefix, false,
-                                         getClass().getClassLoader());
-                /* #else */
-                // Class cl = Class.forName(prefix);
-                /* #endif */
-                return ClassMethodProc.makeExp(new QuoteExp(Type.make(cl)),
-                                               new QuoteExp(local));
-              }
-            catch (Throwable ex)
-              {
-                return null;
-              }
-	  }
       }
-    return rewrite(sym, function);
+    return name;
   }
 
   public void setCurrentScope (ScopeExp scope)
@@ -567,11 +555,63 @@ public class Translator extends Compilation
     if (exp instanceof PairWithPosition)
       return rewrite_with_position (exp, function, (PairWithPosition) exp);
     else if (exp instanceof Pair)
-      return rewrite_pair ((Pair) exp);
+      return rewrite_pair((Pair) exp, function);
     else if (exp instanceof String
 	     || (exp instanceof Symbol && ! selfEvaluatingSymbol(exp)))
       {
 	Declaration decl = lexical.lookup(exp, function);
+        Declaration cdecl = null;
+
+        // If we're nested inside a class (in a ClassExp) then the field
+        // and methods names of this class and super-classes/interfaces
+        // need to be searched.
+        ScopeExp scope = current_scope;
+        int decl_nesting = decl == null ? -1 : ScopeExp.nesting(decl.context);
+        String dname;
+        if (exp instanceof String
+            || (exp instanceof Symbol && ((Symbol) exp).hasEmptyNamespace()))
+          dname = exp.toString();
+        else
+          {
+            dname = null;
+            scope = null;
+          }
+        for (;scope != null; scope = scope.outer)
+          {
+            if (scope instanceof LambdaExp
+                && scope.outer instanceof ClassExp // redundant? FIXME
+                && ((LambdaExp) scope).isClassMethod())
+              {
+                if (decl_nesting >= ScopeExp.nesting(scope.outer))
+                  break;
+                LambdaExp caller = (LambdaExp) scope;
+                ClassExp cexp = (ClassExp) scope.outer;
+                ClassType ctype = (ClassType) cexp.getType();
+                Object part = SlotGet.lookupMember(ctype, dname, ctype);
+                boolean contextStatic
+                  = (caller == cexp.clinitMethod
+                     || (caller != cexp.initMethod 
+                         && caller.nameDecl.isStatic()));
+                if (part == null)
+                  {
+                    char mode = contextStatic ? 'S' : 'V';
+                    PrimProcedure[] methods
+                      = ClassMethods.getMethods(ctype, dname,
+                                                mode, ctype, language);
+                    if (methods.length == 0)
+                      continue;
+                  }
+                Expression part1;
+                // FIXME We're throwing away 'part', which is wasteful.
+                if (contextStatic)
+                  part1 = new ReferenceExp(((ClassExp) caller.outer).nameDecl);
+                else
+                  part1 = new ThisExp(caller.firstDecl());
+                return GetNamedPart.makeExp(part1,
+                                            QuoteExp.getInstance(dname));
+              }
+          }
+
 	Object nameToLookup;
 	Symbol symbol = null;
 	if (decl != null)
@@ -593,44 +633,63 @@ public class Translator extends Compilation
 	  {
 	    nameToLookup = exp;
 	  }
-	if (nameToLookup instanceof String && decl == null)
-	  {
-	    Expression ss = namespaceResolve((String) nameToLookup, function);
-	    if (ss != null)
-	      return ss;
-	  }
 	symbol = exp instanceof String ? env.getSymbol((String) exp)
 	  : (Symbol) exp;
-	Object value = resolve(symbol, function);
 	boolean separate = getLanguage().hasSeparateFunctionNamespace();
         if (decl != null)
           {
             if (! isLexical(decl)
                 || (separate && decl.isProcedureDecl()))
               decl = null;
-          }
-        else if (value instanceof Named)
-          {
-            if (value instanceof AutoloadProcedure)
+            else if (current_scope instanceof TemplateScope && decl.needsContext())
+              cdecl = ((TemplateScope) current_scope).macroContext;
+            else if (decl.getFlag(Declaration.FIELD_OR_METHOD)
+                     && ! decl.isStatic())
               {
-                try
-                  {
-                    value = ((AutoloadProcedure) value).getLoaded();
-                  }
-                catch (RuntimeException ex)
+                scope = currentScope();
+                for (;;)
                   {
+                    if (scope == null)
+                      throw new Error("internal error: missing "+decl);
+                    if (scope.outer == decl.context) // I.e. same class.
+                      break;
+                    scope = scope.outer;
                   }
+                cdecl = scope.firstDecl();
               }
-            Named proc = (Named) value;
+          }
+        else
+          {
 	    Location loc
 	      = env.lookup(symbol,
 			   function && separate ? EnvironmentKey.FUNCTION
 			   : null);
 	    if (loc != null)
 	      loc = loc.getBase();
-            if (loc instanceof FieldLocation && inlineOk(null))
+            if (loc instanceof FieldLocation)
               {
-                decl = ((FieldLocation) loc).getDeclaration();
+                FieldLocation floc = (FieldLocation) loc;
+                try
+                  {
+                    decl = floc.getDeclaration();
+                    if (! inlineOk(null)
+                        // A kludge - we get a bunch of testsuite failures
+                        // if we don't inline $lookup$.  FIXME.
+                        && decl != kawa.standard.Scheme.getNamedPartDecl)
+                      decl = null;
+                    if (decl != null && ! decl.isStatic())
+                      {
+                        cdecl = new Declaration("(module-instance)");
+                        cdecl.setValue(new QuoteExp(floc.getInstance()));
+                      }
+                  }
+                catch (Throwable ex)
+                  {
+                    error('e',
+                          "exception loading '" + exp
+                          + "' - " + ex.getMessage());
+                    decl = null;
+                  }
               }
 	    /*
             else if (Compilation.inlineOk && function)
@@ -639,22 +698,15 @@ public class Translator extends Compilation
 		// which just call rewrite_car on the lhs,
 		// if we don't require function to be true.  FIXME.
                 decl = Declaration.getDeclaration(proc);
-		System.err.println("getDecl "+proc+"->"+decl);
               }
-	    */
+            */
           }
-	if (decl != null && decl.getFlag(Declaration.FIELD_OR_METHOD)
-	    && decl.isProcedureDecl() && ! function)
-	  return syntaxError("not implemented: variable reference to a method");
 	if (decl != null && decl.getContext() instanceof PatternScope)
 	  return syntaxError("reference to pattern variable "+decl.getName()+" outside syntax template");
 
 	ReferenceExp rexp = new ReferenceExp (nameToLookup, decl);
-	if (current_scope instanceof TemplateScope
-            && decl != null && decl.needsContext())
-	  rexp.setContextDecl(((TemplateScope) current_scope).macroContext);
-        else
-          checkMemberContext(rexp, decl);
+        rexp.setContextDecl(cdecl);
+        rexp.setLine(this);
 	if (function && separate)
 	  rexp.setFlag(ReferenceExp.PREFER_BINDING2);
 	return rexp;
@@ -667,41 +719,16 @@ public class Translator extends Compilation
       return QuoteExp.getInstance(Quote.quote(exp, this));
   }
 
-  public void checkMemberContext (AccessExp exp, Declaration decl)
+  public static void setLine(Expression exp, Object location)
   {
-    if (decl == null || ! decl.getFlag(Declaration.FIELD_OR_METHOD)
-        || decl.isStatic())
-      return;
-    ScopeExp scope = currentScope();
-    for (;;)
-      {
-        if (scope == null)
-          throw new Error("internal error: missing "+decl);
-        if (scope.outer == decl.context) // I.e. same class.
-          break;
-        scope = scope.outer;
-      }
-    exp.setContextDecl(scope.firstDecl());
+    if (location instanceof SourceLocator)
+      exp.setLocation((SourceLocator) location);
   }
 
-  public static void setLine(Expression exp, Object pair)
+  public static void setLine(Declaration decl, Object location)
   {
-    if (pair instanceof PairWithPosition)
-      {
-	PairWithPosition expPos = (PairWithPosition) pair;
-	exp.setFile(expPos.getFile());
-	exp.setLine(expPos.getLine(), expPos.getColumn());
-      }
-  }
-
-  public static void setLine(Declaration decl, Object pair)
-  {
-    if (pair instanceof PairWithPosition)
-      {
-	PairWithPosition declPos = (PairWithPosition) pair;
-	decl.setFile(declPos.getFile());
-	decl.setLine(declPos.getLine(), declPos.getColumn());
-      }
+    if (location instanceof SourceLocator)
+      decl.setLocation((SourceLocator) location);
   }
 
   PairWithPosition positionPair;
@@ -718,12 +745,11 @@ public class Translator extends Compilation
     PairWithPosition ppair = (PairWithPosition) pair;
     Object saved;
     if (positionPair == null
-	|| positionPair.getFile() != getFile()
-	|| positionPair.getLine() != getLine()
-	|| positionPair.getColumn() != getColumn())
+	|| positionPair.getFileName() != getFileName()
+	|| positionPair.getLineNumber() != getLineNumber()
+	|| positionPair.getColumnNumber() != getColumnNumber())
       {
-	saved = PairWithPosition.make(Special.eof, positionPair,
-				      getFile(), getLine(), getColumn());
+        saved = new PairWithPosition(this, Special.eof, positionPair);
       }
     else
       saved = positionPair;
@@ -746,13 +772,10 @@ public class Translator extends Compilation
   }
 
 
-  public void setLine (Object pair)
+  public void setLine (Object location)
   {
-    if (pair instanceof PairWithPosition)
-      {
-	PairWithPosition pos = (PairWithPosition) pair;
-	setLine(pos.getFile(), pos.getLine(), pos.getColumn());
-      }
+    if (location instanceof SourceLocator)
+      setLocation((SourceLocator) location);
   }
 
   /** Set the line position of the argument to the current position. */
@@ -761,21 +784,20 @@ public class Translator extends Compilation
   {
     if (exp instanceof QuoteExp)
       return;
-    if (exp.getFile () == null)
-      exp.setFile(getFile());
-    if (exp.getLine () == 0)
-      exp.setLine (getLine(), getColumn());
+    exp.setLocation(this);
   }
 
-  /** Extract a type from the car fo a pair. */
+  /** Extract a type from the car of a pair. */
   public Type exp2Type(Pair typeSpecPair)
   {
     Object saved = pushPositionOf(typeSpecPair);
     try
       {
 	Expression texp = rewrite_car(typeSpecPair, false);
+        texp = new InlineCalls(this).walk(texp);
 	if (texp instanceof ErrorExp)
 	  return null;
+        texp = new InlineCalls(this).walk(texp);
 	Type type = getLanguage().getTypeFor(texp);
 	 if (type == null)
 	   {
@@ -803,7 +825,7 @@ public class Translator extends Compilation
     try
       {
 	if (exp == pair)
-	  result = rewrite_pair (pair);  // To avoid a cycle
+	  result = rewrite_pair(pair, function);  // To avoid a cycle
 	else
 	  result = rewrite (exp, function);
 	setLineOf(result);
@@ -888,19 +910,38 @@ public class Translator extends Compilation
                 setCurrentScope(sf.scope);
                 obj = sf.form;
               }
+            Pair p;
+            if (obj instanceof Pair
+                && (p = (Pair) obj).car == LispLanguage.lookup_sym
+                && p.cdr instanceof Pair
+                && (p = (Pair) p.cdr).cdr instanceof Pair)
+              {
+                Expression part1 = rewrite(p.car);
+                Expression part2 = rewrite(((Pair) p.cdr).car);
+                obj = namespaceResolve(part1, part2);
+              }
             if (obj instanceof String
                 || (obj instanceof Symbol && ! selfEvaluatingSymbol(obj)))
               {
-                Declaration decl = lexical.lookup(obj, true);
-                if (decl != null)
-                  syntax = check_if_Syntax(decl);
-                else
+                Expression func = rewrite(obj, true);
+                if (func instanceof ReferenceExp)
                   {
-                    obj = resolve(obj, true);
-                    if (obj instanceof Syntax)
-                      syntax = (Syntax) obj;
+                    Declaration decl = ((ReferenceExp) func).getBinding();
+                    if (decl != null)
+                      syntax = check_if_Syntax(decl);
+                    else
+                      {
+                        obj = resolve(obj, true);
+                        if (obj instanceof Syntax)
+                          syntax = (Syntax) obj;
+                      }
                   }
               }
+            // Recognize deferred begin created in scanBody for pendingForms.
+            // A seemingly-cleaned (obj instanceof Syntax) causes problems
+            // with some Syntax forms, such as define.
+            else if (obj == kawa.standard.begin.begin)
+              syntax = (Syntax) obj;
           }
         finally
           {
@@ -909,9 +950,9 @@ public class Translator extends Compilation
           }
 	if (syntax != null)
 	  {
-	    String save_filename = getFile();
-	    int save_line = getLine();
-	    int save_column = getColumn();
+	    String save_filename = getFileName();
+	    int save_line = getLineNumber();
+	    int save_column = getColumnNumber();
 	    try
 	      {
 		setLine(st_pair);
@@ -951,16 +992,16 @@ public class Translator extends Compilation
 	      {
 		setCurrentScope(sf.scope);
 		int first = formStack.size();
-		scanBody(sf.form, defs, false);
-		Object f = wrapSyntax(popForms(first), sf);
-		if (makeList)
-		  {
-		    if (lastPair == null)
+		Object f = scanBody(sf.form, defs, makeList);
+                if (makeList)
+                  {
+                    f = wrapSyntax(f, sf);
+                    if (lastPair == null)
 		      return f;
-		    lastPair.cdr = f;
-		    return list;
-		  }
-		formStack.add(f);
+                    lastPair.cdr = f;
+                    return list;
+                  }
+		formStack.add(wrapSyntax(popForms(first), sf));
 		return null;
 	      }
 	    finally
@@ -973,6 +1014,16 @@ public class Translator extends Compilation
 	    Pair pair = (Pair) body;
 	    int first = formStack.size();
 	    scanForm(pair.car, defs);
+            if (getState() == Compilation.PROLOG_PARSED)
+              {
+                // We've seen a require form during the initial pass when
+                // we're looking module names.  Defer the require and any
+                // following forms in this body.
+                if (pair.car != pendingForm)
+                  pair = makePair(pair, pendingForm, pair.cdr);
+                pendingForm = new Pair(kawa.standard.begin.begin, pair);
+                return LList.Empty;
+              }
 	    int fsize = formStack.size();
 	    if (makeList)
 	      {
@@ -1036,7 +1087,6 @@ public class Translator extends Compilation
 	setLineOf(body);
 	if (ndecls == 0)
 	  return body;
-	mustCompileHere();
 	defs.body = body;
 	setLineOf(defs);
 	return defs;
@@ -1138,14 +1188,13 @@ public class Translator extends Compilation
       setCurrentScope(saveScope);
   }
 
-  public void finishModule(ModuleExp mexp, int first)
+  public void finishModule(ModuleExp mexp)
   {
     boolean moduleStatic = mexp.isStatic();
     for (Declaration decl = mexp.firstDecl();
 	 decl != null;  decl = decl.nextDecl())
       {
-	if (decl.getFlag(Declaration.NOT_DEFINING)
-	    && ! decl.getFlag(Declaration.IS_UNKNOWN))
+	if (decl.getFlag(Declaration.NOT_DEFINING))
 	  {
 	    String msg1 = "'";
 	    String msg2
@@ -1181,6 +1230,30 @@ public class Translator extends Compilation
       }
     if (! moduleStatic)
       mexp.declareThis(null);
+  }
+
+  public void resolveModule(ModuleExp mexp)
+  {
+    int numPending = pendingImports == null ? 0 : pendingImports.size();
+    for (int i = 0;  i < numPending;  )
+      {
+        ModuleInfo info = (ModuleInfo) pendingImports.elementAt(i++);
+        ScopeExp defs = (ScopeExp) pendingImports.elementAt(i++);
+        Expression posExp = (Expression) pendingImports.elementAt(i++);
+        if (mexp == defs)
+          {
+            // process(BODY_PARSED);
+            Expression savePos = new ReferenceExp((Object) null);
+            savePos.setLine(this);
+            setLine(posExp);
+            kawa.standard.require.importDefinitions(null, info, null,
+                                                    formStack, defs, this);
+            setLine(savePos);
+            pendingImports.setElementAt(null, i-3);
+            pendingImports.setElementAt(null, i-2);
+            pendingImports.setElementAt(null, i-1);
+          }
+      }
 
     processAccesses();
 
@@ -1189,7 +1262,7 @@ public class Translator extends Compilation
     try
       {
 	Compilation.setCurrent(this);
-	mexp.body = makeBody(first, mexp);
+	mexp.body = makeBody(firstForm, mexp);
 	lexical.pop(mexp);
       }
     finally
