@@ -4,16 +4,16 @@ import java.io.*;
 import gnu.mapping.*;
 import gnu.expr.*;
 import gnu.text.SourceMessages;
+import gnu.text.SyntaxException;
 import gnu.lists.*;
 import java.util.Vector;
+import gnu.bytecode.ClassType;
 
 /** Start a "Read-Eval-Print-Loop" for the Kawa Scheme evaluator. */
 
 public class repl extends Procedure0or1
 {
-  public static String compilationDirectory = null;
   public static String compilationTopname = null;
-  public static String compilationPrefix = null;
 
   Language language;
   static Language previousLanguage;
@@ -73,6 +73,8 @@ public class repl extends Procedure0or1
     printOption(out, "--debug-dump-zip", "Compiled interactive expressions to a zip archive");
     printOption(out, "--debug-print-expr", "Print generated internal expressions");
     printOption(out, "--debug-print-final-expr", "Print expression after any optimizations");
+    printOption(out, "--debug-error-prints-stack-trace", "Print stack trace with errors");
+    printOption(out, "--debug-warning-prints-stack-trace", "Print stack trace with warnings");
     printOption(out,"--[no-]full-tailcalls", "(Don't) use full tail-calls");
     printOption(out, "-C <filename> ...", "Compile named files to Java class files");
     printOption(out, "--output-format <format>", "Use <format> when printing top-level output");
@@ -167,7 +169,7 @@ public class repl extends Procedure0or1
 	  scmHomeDirectory = Boolean.FALSE;
 	Environment.getCurrent().put("home-directory", scmHomeDirectory);
 	if (initFile != null && initFile.exists())
-	  Shell.runFile(initFile.getPath());
+	  Shell.runFile(initFile.getPath(), 0);
       }
   }
 
@@ -246,9 +248,34 @@ public class repl extends Procedure0or1
 	    getLanguageFromFilenameExtension(filename);
 	    setArgs (args, iArg+1);
 	    checkInitFile();
-	    Shell.runFile (filename);
+	    Shell.runFile (filename, 0);
 	    something_done = true;
 	  }
+	else if (arg.startsWith("--script"))
+	  {
+            String count = arg.substring(8);
+	    iArg++;
+            int skipLines = 0;
+            if (count.length() > 0)
+              {
+                try
+                  {
+                    skipLines = Integer.parseInt(count);
+                  }
+                catch (Throwable ex)
+                  {
+                    iArg = maxArg; // force bad_option.
+                  }
+              }
+	    if (iArg == maxArg)
+	      bad_option (arg);
+	    String filename = args[iArg];
+	    getLanguageFromFilenameExtension(filename);
+	    setArgs (args, iArg+1);
+	    checkInitFile();
+	    Shell.runFile(filename, skipLines);
+            return -1;
+	  }
 	else if (arg.equals("\\"))
 	  {
 	    // Scsh-like "meta-arg".  See Kawa manual (SOON-FIXME).
@@ -322,10 +349,10 @@ public class repl extends Procedure0or1
 		// FIXME adjust line number
 		setArgs(args, iArg+1);
 		checkInitFile();
-		kawa.standard.load.loadSource(freader, Environment.user());
+		kawa.standard.load.loadSource(freader, Environment.user(), null);
 		return -1;
 	      }
-	    catch (gnu.text.SyntaxException ex)
+	    catch (SyntaxException ex)
 	      {
 		ex.printAll(OutPort.errDefault(), 20);
 	      }
@@ -374,14 +401,15 @@ public class repl extends Procedure0or1
 	    iArg++;
 	    if (iArg == maxArg)
 	      bad_option (arg);
-	    compilationDirectory = args[iArg];
+            ModuleManager manager = ModuleManager.getInstance();
+	    manager.setCompilationDirectory(args[iArg]);
 	  }
 	else if (arg.equals ("-P"))
 	  {
 	    iArg++;
 	    if (iArg == maxArg)
 	      bad_option (arg);
-	    compilationPrefix = args[iArg];
+	    Compilation.classPrefixDefault = args[iArg];
 	  }
 	else if (arg.equals ("-T"))
 	  {
@@ -395,32 +423,7 @@ public class repl extends Procedure0or1
 	    ++iArg;
 	    if (iArg == maxArg)
 	      bad_option (arg);
-	    for ( ; iArg < maxArg;  iArg++)
-	      {
-		arg = args[iArg];
-		getLanguageFromFilenameExtension(arg);
-		try
-		  {
-		    System.err.println("(compiling "+arg+")");
-		    SourceMessages messages = new SourceMessages();
-
-		    CompileFile.compile_to_files(arg,
-						 compilationDirectory,
-						 compilationPrefix,
-						 compilationTopname,
-						 messages);
-		    boolean sawErrors = messages.seenErrors();
-		    messages.checkErrors(System.err, 50);
-		    if (sawErrors)
-		      System.exit(-1);
-		  }
-		catch (Throwable ex)
-		  {
-		    System.err.println("Internal error while compiling "+arg);
-		    ex.printStackTrace(System.err);
-		    System.exit(-1);
-		  }
-	      }
+            compileFiles(args, iArg, maxArg);
 	    return -1;
 	  }
 	else if (arg.equals("--output-format")
@@ -527,12 +530,20 @@ public class repl extends Procedure0or1
 	  }
 	else if (arg.equals("--debug-print-expr"))
 	  {
-	    gnu.expr.ModuleExp.debugPrintExpr = true;
+            Compilation.debugPrintExpr = true;
 	  }
 	else if (arg.equals("--debug-print-final-expr"))
 	  {
 	    Compilation.debugPrintFinalExpr = true;
 	  }
+	else if (arg.equals("--debug-error-prints-stack-trace"))
+	  {
+            SourceMessages.debugStackTraceOnError = true;
+	  }
+	else if (arg.equals("--debug-warning-prints-stack-trace"))
+	  {
+            SourceMessages.debugStackTraceOnWarning = true;
+	  }
 	else if (arg.equals("--module-static"))
 	  {
 	    gnu.expr.Compilation.moduleStatic = 1;
@@ -570,6 +581,10 @@ public class repl extends Procedure0or1
 	    Compilation.defaultCallConvention
 	      = Compilation.CALL_WITH_RETURN;
 	  }
+        else if (arg.equals("--pedantic"))
+          {
+            Language.requirePedantic = true;
+          }
 	else if (arg.equals("--help"))
 	  {
 	    printOptions(System.out);
@@ -585,7 +600,7 @@ public class repl extends Procedure0or1
 	    System.out.print("Kawa ");
 	    System.out.print(Version.getVersion());
 	    System.out.println();
-	    System.out.println("Copyright (C) 2005 Per Bothner");
+	    System.out.println("Copyright (C) 2007 Per Bothner");
 	    something_done = true;
 	  }
 	else if (arg.length () > 0 && arg.charAt(0) == '-')
@@ -646,12 +661,12 @@ public class repl extends Procedure0or1
 	  }
 	else
           {
-            int eq = arg.indexOf('=');
-            if (eq <= 0)
+            int ci = arg.indexOf('=');
+            if (ci <= 0)
               return iArg;
-            String key = arg.substring(0, eq);
-            String value = arg.substring(eq+1);
-            String uri, local;
+            String key = arg.substring(0, ci);
+            String value = arg.substring(ci+1);
+            String uri, local, prefix = "";
             for (int i = 0; ; i++)
               {
                 String[] propertyField = propertyFields[i];
@@ -667,7 +682,7 @@ public class repl extends Procedure0or1
                         ThreadLocation loc = (ThreadLocation)
                           clas.getDeclaredField(fname).get(null);
                         loc.setGlobal(value);
-                        continue;
+                        break;
                       }
                     catch (Throwable ex)
                       {
@@ -677,23 +692,7 @@ public class repl extends Procedure0or1
                       }
                   }
               }
-            if (key.charAt(0) == '{')
-              {
-                int rbrace = key.lastIndexOf('}');
-                if (rbrace <= 0)
-                  {
-                    System.err.println("missing '}' in property name '"+key+"'");
-                    System.exit (-1);
-                  }
-                uri = key.substring(1, rbrace);
-                local = key.substring(rbrace+1);
-              }
-            else
-              {
-                uri = "";
-                local = key;
-              }
-            Symbol symbol = Symbol.make(uri, local);
+            Symbol symbol = Symbol.parse(key);
             // Run Language's static initializer.
             Language.getDefaultLanguage();
             Environment current = Environment.getCurrent();
@@ -703,6 +702,107 @@ public class repl extends Procedure0or1
     return something_done ? -1 : iArg;
   }
 
+  public static void compileFiles (String[] args, int iArg, int maxArg)
+  {
+    ModuleManager manager = ModuleManager.getInstance();
+    Compilation[] comps = new Compilation[maxArg-iArg];
+    ModuleInfo[] infos = new ModuleInfo[maxArg-iArg];
+    SourceMessages messages = new SourceMessages();
+    for (int i = iArg; i < maxArg;  i++)
+      {
+        String arg = args[i];
+        getLanguageFromFilenameExtension(arg);
+        Language language = Language.getDefaultLanguage();
+        try
+          {
+            InPort fstream;
+            try
+              {
+                fstream = InPort.openFile(arg);
+              }
+            catch (java.io.FileNotFoundException ex)
+              {
+                System.err.println(ex);
+                System.exit(-1);
+                break; // Kludge to shut up compiler.
+              }
+            
+            Compilation comp
+              = language.parse(fstream, messages, Language.PARSE_PROLOG);
+
+            if (compilationTopname != null)
+              {
+                String cname
+                  = Compilation.mangleNameIfNeeded(compilationTopname);
+                ClassType ctype = new ClassType(cname);
+                ModuleExp mexp = comp.getModule();
+                mexp.setType(ctype);
+                mexp.setName(compilationTopname);
+                comp.mainClass = ctype;
+              }
+            
+            infos[i-iArg] = manager.find(comp);
+            comps[i-iArg] = comp;
+
+          }
+        catch (Throwable ex)
+          {
+            if (! (ex instanceof SyntaxException)
+                || ((SyntaxException) ex).getMessages() != messages)
+              {
+                System.err.println("Internal error while compiling "+arg);
+                ex.printStackTrace(System.err);
+                System.exit(-1);
+              }
+          }
+        if (messages.seenErrorsOrWarnings())
+          {
+            System.err.println("(compiling "+arg+')');
+            if (messages.checkErrors(System.err, 20))
+              System.exit(1);
+          }
+      }
+
+    for (int i = iArg; i < maxArg;  i++)
+      {
+        String arg = args[i];
+        Compilation comp = comps[i-iArg];
+        try
+          {
+            System.err.println("(compiling "+arg+" to "+comp.mainClass.getName()+')');
+
+            infos[i-iArg].loadByStages(Compilation.CLASS_WRITTEN);
+            boolean sawErrors = messages.seenErrors();
+            messages.checkErrors(System.err, 50);
+            if (sawErrors)
+              System.exit(-1);
+            comps[i-iArg] = comp; 
+            sawErrors = messages.seenErrors();
+            messages.checkErrors(System.err, 50);
+            if (sawErrors)
+              System.exit(-1);
+          }
+        catch (Throwable ex)
+          {
+            StringBuffer sbuf = new StringBuffer();
+            String file = comp.getFileName();
+            int line = comp.getLineNumber();
+            if (file != null && line > 0)
+              {
+                sbuf.append(file);
+                sbuf.append(':');
+                sbuf.append(line);
+                sbuf.append(": ");
+              }
+            sbuf.append("internal error while compiling ");
+            sbuf.append(arg);
+            System.err.println(sbuf.toString());
+            ex.printStackTrace(System.err);
+            System.exit(-1);
+          }
+      }
+  }
+
   /** A list of standard command-line fluid names to map to static fields.
    * For each entry:
    * element 0 is a property name (before the '=' in the comamnd-line);
@@ -718,6 +818,7 @@ public class repl extends Procedure0or1
       { "out:right-margin", "gnu.text.PrettyWriter", "lineLengthLoc" },
       { "out:miser-width", "gnu.text.PrettyWriter", "miserWidthLoc" },
       { "out:xml-indent", "gnu.xml.XMLPrinter", "indentLoc" },
+      { "display:toolkit", "gnu.kawa.models.Display", "myDisplay" },
       null
     };
 
@@ -734,7 +835,7 @@ public class repl extends Procedure0or1
 	    getLanguageFromFilenameExtension(filename);
 	    setArgs (args, iArg+1);
 	    checkInitFile();
-	    Shell.runFile (filename);
+	    Shell.runFile(filename, 0);
 	  }
 	else
 	  {
