@@ -1,5 +1,4 @@
 package kawa.standard;
-import kawa.lang.*;
 import java.io.*;
 import gnu.bytecode.ZipLoader;
 import gnu.mapping.*;
@@ -7,7 +6,7 @@ import gnu.expr.*;
 import gnu.text.SourceMessages;
 import gnu.text.SyntaxException;
 import gnu.lists.*;
-import gnu.kawa.functions.BaseUri;
+import gnu.text.URI_utils;
 import java.net.URL;
 
 public class load extends Procedure1 {
@@ -40,17 +39,10 @@ public class load extends Procedure1 {
 	if (inst instanceof Runnable)
 	  ((Runnable)inst).run();
       }
-    catch (ClassNotFoundException ex)
+    catch (Throwable ex)
       {
-	throw new RuntimeException ("class not found in load");
-      }
-    catch (InstantiationException ex)
-      {
-	throw new RuntimeException ("class not instantiable: in load");
-      }
-    catch (IllegalAccessException ex)
-      {
-	throw new RuntimeException ("class illegal access: in load");
+        throw new WrappedException("exception during load of \""+name+'\"',
+                                   ex);
       }
     finally
       {
@@ -73,27 +65,24 @@ public class load extends Procedure1 {
 	if (!zfile.canRead ())
 	  throw new RuntimeException ("load: "+name+" - not readable");
 	ZipLoader loader = new ZipLoader (name);
-	loader.loadAllClasses();
-	Class clas = loader.loadClass (LambdaExp.fileFunctionName, true);
-	Object proc = clas.newInstance ();
+        Class clas = loader.loadAllClasses();
+        Object proc = clas == null ? null : clas.newInstance();
+        if (! (proc instanceof ModuleBody))
+          throw new RuntimeException("load: "+name+" - no module in archive");
 	gnu.kawa.reflect.ClassMemberLocation.defineAll(proc, env);
 	((ModuleBody) proc).run();
       }
     catch (java.io.IOException ex)
       {
-	throw new RuntimeException ("load: "+name+" - "+ex.toString ());
-      }
-    catch (ClassNotFoundException ex)
-      {
-	throw new RuntimeException ("class not found in load");
+	throw new WrappedException ("load: "+name+" - "+ex.toString (), ex);
       }
     catch (InstantiationException ex)
       {
-	throw new RuntimeException ("class not instantiable: in load");
+	throw new WrappedException ("class not instantiable: in load", ex);
       }
     catch (IllegalAccessException ex)
       {
-	throw new RuntimeException ("class illegal access: in load");
+	throw new WrappedException ("class illegal access: in load", ex);
       }
     finally
       {
@@ -102,38 +91,7 @@ public class load extends Procedure1 {
       }
   }
 
-  private static void loadSource (String name, Environment env)
-    throws SyntaxException
-  {
-    try
-      {
-	InPort fstream;
-	if (BaseUri.hasScheme(name))
-	  {
-	    URL url = new URL(name);
-	    InputStream in = url.openConnection().getInputStream();
-	    fstream = InPort.openFile(new BufferedInputStream(in), name);
-	  }
-	else
-	  fstream = InPort.openFile(name);
-	loadSource (fstream, env);
-	fstream.close();
-      }
-    catch (java.io.FileNotFoundException e)
-      {
-	throw new RuntimeException ("load: file not found: " + name);
-      }
-    catch (java.io.IOException ex)
-      {
-	throw new RuntimeException("failed to close \""+name+"\" after loading");
-      }
-    catch (Throwable ex)
-      {
-	throw new WrappedException(ex);
-      }
-  }
-
-  public final static void loadSource (InPort port, Environment env)
+  public final static void loadSource (InPort port, Environment env, URL url)
     throws SyntaxException, Throwable
   {
     boolean print = ModuleBody.getMainPrintValues();
@@ -146,7 +104,7 @@ public class load extends Procedure1 {
     // So instead, we read and evaluate each line individually.
     if (true)
       {
-	kawa.Shell.run(language, env, port, out, OutPort.errDefault());
+	kawa.Shell.run(language, env, port, out, OutPort.errDefault(), url);
       }
     else
       {
@@ -154,14 +112,13 @@ public class load extends Procedure1 {
 	Compilation comp
           = language.parse(port, messages, Language.PARSE_IMMEDIATE);
 	ModuleExp mexp = comp.getModule();
-	mexp.setName(Symbols.make(LambdaExp.fileFunctionName));
 	CallContext ctx = CallContext.getInstance();
 	Consumer save = ctx.consumer;
 	try
 	  {
 	    ctx.consumer = out;
 	    ctx.values = Values.noArgs;
-	    ModuleExp.evalModule(env, ctx, comp);
+	    ModuleExp.evalModule(env, ctx, comp, url, null);
 	    if (messages.seenErrors())
 	      throw new SyntaxException(messages);
 	  }
@@ -185,7 +142,7 @@ public class load extends Procedure1 {
     try
       {
 	Environment env = (Environment) arg2;
-	apply (name, env, relative);
+	apply (name, env, relative, 0);
 	return Values.empty;
       }
     catch (java.io.FileNotFoundException e)
@@ -200,113 +157,70 @@ public class load extends Procedure1 {
   }
 
   public static final void apply (String name, Environment env,
-				  boolean relative)
+				  boolean relative, int skipLines)
     throws Throwable
   {
     CallContext ctx = CallContext.getInstance();
-    String savedBaseUri = ctx.getBaseUri();
-    try
+    boolean isUri = InPort.uriSchemeSpecified(name);
+    // Resolve a relative URI.  However, if the base uri matches the
+    // default base uri (i.e. the current directory) just leave it as
+    // a filename, rather than coercing it to a URI.
+    String resolved = name;
+    if (! isUri)
       {
-        // Resolve a relative URI.  However, if the base uri matches the
-        // default base uri (i.e. the current directory) just leave it as
-        // a filename, rather than coercing it to a URI.
-        if (! BaseUri.hasScheme(name)
-            && relative
-            && ! savedBaseUri.equals(ctx.getBaseUriDefault()))
-          name = BaseUri.resolve(name, savedBaseUri);
-	ctx.setBaseUri(name);
-	if (name.endsWith (".zip") || name.endsWith(".jar"))
-	  {
-	    loadCompiled (name, env);
-	    return;
-	  }
-	if (name.endsWith (".scm"))
-	  {
-	    loadSource (name, env);
-	    return;
-	  }
-	char file_separator = System.getProperty ("file.separator").charAt(0);
-
-	if (name.endsWith (".class"))
-	  {
-	    name = name.substring (0, name.length () - 6);
-	    name = name.replace ('/', '.');
-	    if (file_separator != '/')
-	      name = name.replace (file_separator, '.');
-	    loadClassFile (name, env);
-	    return;
-	  }
-	File file = new File (name);
-	if (file.exists ())
-	  {
-	    InputStream fs = new BufferedInputStream (new FileInputStream (name));
-	    fs.mark(5);
-	    int char0 = fs.read ();
-	    if (char0 == -1)
-	      return; // Sequence.eofValue;
-	    if (char0 == 'P')
-	      {
-		int char1 = fs.read ();
-		if (char1 == 'K')
-		  {
-		    int char2 = fs.read ();
-		    if (char2 == '\003')
-		      {
-			int char3 = fs.read ();
-			if (char3 == '\004')
-			  {
-			    fs.close ();
-			    loadCompiled (name, env);
-			    return;
-			  }
-		      }
-		  }
-	      }
-	    fs.reset();
-	    InPort src = InPort.openFile(fs, name);
-	    loadSource (src, env);
-	    src.close();
-	    return;
-	  }
-	else
-	  {
-	    String fname = name.replace ('.', file_separator);
-	    String xname = fname + ".zip";
-	    file = new File (xname);
-	    if (file.exists ())
-	      {
-		loadCompiled (xname, env);
-		return;
-	      }
-	    xname = fname + ".jar";
-	    file = new File (xname);
-	    if (file.exists ())
-	      {
-		loadCompiled (xname, env);
-		return;
-	      }
-
-	    xname = fname + ".class";
-	    file = new File (xname);
-	    if (file.exists ())
-	      {
-		loadClassFile (name, env);
-		return;
-	      }
-
-	    xname = fname + ".scm";
-	    file = new File (xname);
-	    if (file.exists ())
-	      {
-		loadSource (xname, env);
-		return;
-	      }
-	  }
-	throw new java.io.FileNotFoundException(name);
+        String baseUri = ctx.getBaseUri();
+        Object base = baseUri;
+        if (! URI_utils.isAbsolute(base))
+          base = URI_utils.resolve(base, ctx.getBaseUriDefault());
+        resolved = URI_utils.resolve(name, base).toString();
+        if (relative && ! baseUri.equals(ctx.getBaseUriDefault()))
+          name = resolved;
       }
-    finally
+    if (name.endsWith (".zip") || name.endsWith(".jar"))
       {
-	ctx.setBaseUri(savedBaseUri);
+        loadCompiled (name, env);
+        return;
       }
+    URL url = new URL(resolved);
+    char file_separator = System.getProperty ("file.separator").charAt(0);
+
+    if (name.endsWith (".class"))
+      {
+        name = name.substring (0, name.length () - 6);
+        name = name.replace ('/', '.');
+        if (file_separator != '/')
+          name = name.replace (file_separator, '.');
+        loadClassFile (name, env);
+        return;
+      }
+    InputStream fs = new BufferedInputStream(URI_utils.getInputStream(url));
+    fs.mark(5);
+    int char0 = fs.read ();
+    if (char0 == -1)
+      return; // Sequence.eofValue;
+    if (char0 == 'P')
+      {
+        int char1 = fs.read ();
+        if (char1 == 'K')
+          {
+            int char2 = fs.read ();
+            if (char2 == '\003')
+              {
+                int char3 = fs.read ();
+                if (char3 == '\004')
+                  {
+                    fs.close ();
+                    loadCompiled (name, env);
+                    return;
+                  }
+              }
+          }
+      }
+    fs.reset();
+    InPort src = InPort.openFile(fs, name);
+    while (--skipLines >= 0)
+      src.skipRestOfLine();
+    loadSource (src, env, url);
+    src.close();
   }
 }
