@@ -1,5 +1,4 @@
 package kawa.standard;
-import kawa.lang.*;
 import gnu.bytecode.Type;
 import gnu.bytecode.CodeAttr;
 import gnu.mapping.*;
@@ -9,16 +8,26 @@ import gnu.expr.*;
 
 public class not extends Procedure1 implements Inlineable
 {
-  Interpreter interpreter;
+  Language language;
+  public QuoteExp trueExp;
+  public QuoteExp falseExp;
 
-  public not(Interpreter interpreter)
+  public not(Language language)
   {
-    this.interpreter = interpreter;
+    this.language = language;
+    trueExp = new QuoteExp(language.booleanObject(true));
+    falseExp = new QuoteExp(language.booleanObject(false));
+  }
+
+  public not(Language language, String name)
+  {
+    this(language);
+    setName(name);
   }
 
   public Object apply1 (Object arg1)
    {
-     return interpreter.booleanObject(! interpreter.isTrue(arg1));
+     return language.booleanObject(! language.isTrue(arg1));
    }
 
   public void compile (ApplyExp exp, Compilation comp, Target target)
@@ -28,7 +37,7 @@ public class not extends Procedure1 implements Inlineable
       {
 	ConditionalTarget ctarget = (ConditionalTarget) target;
 	ConditionalTarget sub_target
-	  = new ConditionalTarget(ctarget.ifFalse, ctarget.ifTrue);
+	  = new ConditionalTarget(ctarget.ifFalse, ctarget.ifTrue, language);
 	sub_target.trueBranchComesFirst = ! ctarget.trueBranchComesFirst;
 	arg.compile(comp, sub_target);
 	return;
@@ -42,12 +51,12 @@ public class not extends Procedure1 implements Inlineable
       }
     else
       {
-	IfExp.compile(arg, QuoteExp.falseExp, QuoteExp.trueExp, comp, target);
+	IfExp.compile(arg, falseExp, trueExp, comp, target);
       }
   }
 
   public Type getReturnType (Expression[] args)
   {
-    return Scheme.booleanType;
+    return language.getTypeFor(Boolean.TYPE);
   }
 }
