@@ -34,17 +34,14 @@ public class object extends Syntax
     if (! (form.cdr instanceof Pair))
       return tr.syntaxError("missing superclass specification in object");
     Pair pair = (Pair) form.cdr;
-    String clname = null;
+    ObjectExp oexp = new ObjectExp();
     if (pair.car instanceof FString)
       {
-	clname = pair.car.toString();
+        // oexp.setName(pair.car.toString());
 	if (! (pair.cdr instanceof Pair))
 	  return tr.syntaxError("missing superclass specification after object class name");
 	pair = (Pair) pair.cdr;
       }
-    ObjectExp oexp = new ObjectExp();
-    oexp.setSimple(true);
-    // if (clname != null) oexp.setName(clname);
     Object[] saved = scanClassDef(pair, oexp, tr);
     if (saved != null)
       rewriteClassDef(saved, tr);
@@ -57,26 +54,7 @@ public class object extends Syntax
   public Object[] scanClassDef (Pair pair, ClassExp oexp, Translator tr)
   {
     tr.mustCompileHere();
-    int num_supers = Translator.listLength(pair.car);
-    if (num_supers < 0)
-      {
-	tr.error('e', "object superclass specification not a list");
-	return null;
-      }
-    Expression[] supers = new Expression[num_supers];
     Object superlist = pair.car;
-    for (int i = 0;  i < num_supers;  i++)
-      {
-	while (superlist instanceof SyntaxForm)
-	  {
-	    // FIXME - need to pass syntax.
-	    superlist = ((SyntaxForm) superlist).form;
-	  }
-	Pair superpair = (Pair) superlist;
-	supers[i] = tr.rewrite_car(superpair, false);
-	superlist = superpair.cdr;
-      }
-    oexp.supers = supers;
     Object components = pair.cdr;
     LambdaExp method_list = null;
     LambdaExp last_method = null;
@@ -291,7 +269,8 @@ public class object extends Syntax
 		tr.error('e', "missing method name");
 		return null;
 	      }
-	    Declaration decl = oexp.addDeclaration(mname);
+	    Declaration decl
+              = oexp.addDeclaration(mname, Compilation.typeProcedure);
 	    Translator.setLine(decl, mpair);
 	    LambdaExp lexp = new LambdaExp();
 	    lexp.outer = oexp;
@@ -314,7 +293,8 @@ public class object extends Syntax
       oexp,
       components,
       inits,
-      method_list
+      method_list,
+      superlist
     };
     return result;
   }
@@ -325,8 +305,32 @@ public class object extends Syntax
     Object components = saved[1];
     Vector inits = (Vector) saved[2];
     LambdaExp method_list = (LambdaExp) saved[3];
+    Object superlist = saved[4];
     oexp.firstChild = method_list;
 
+    int num_supers = Translator.listLength(superlist);
+    if (num_supers < 0)
+      {
+        tr.error('e', "object superclass specification not a list");
+        num_supers = 0;
+      }
+    Expression[] supers = new Expression[num_supers];
+    for (int i = 0;  i < num_supers;  i++)
+      {
+	while (superlist instanceof SyntaxForm)
+	  {
+	    // FIXME - need to pass syntax.
+	    superlist = ((SyntaxForm) superlist).form;
+	  }
+	Pair superpair = (Pair) superlist;
+	supers[i] = tr.rewrite_car(superpair, false);
+	superlist = superpair.cdr;
+      }
+    oexp.supers = supers;
+
+    oexp.setClassName(tr);
+    oexp.setTypes(tr);
+
     // First a pass over init-form: specifiers, since these are evaluated
     // in a scope outside the current class.
     int len = inits.size();
@@ -336,7 +340,7 @@ public class object extends Syntax
 	if (init != null)
           rewriteInit(inits.elementAt(i), oexp, (Pair) init, tr, null);
       }
-    
+
     tr.push(oexp);
 
     // Pass to rewrite method/initializer bodies.
@@ -451,12 +455,15 @@ public class object extends Syntax
                       tr.error('e', "'*init*' methods only supported for simple classes");
                   }
                 Translator.setLine(meth, pair);
+                LambdaExp saveLambda = tr.curMethodLambda;
+                tr.curMethodLambda = meth;
 		lambda.rewrite(meth, ((Pair) pair_car).cdr, pair.cdr, tr,
 			       memberCarSyntax != null
 			       && (memberSyntax == null
 				   || memberCarSyntax.scope != memberSyntax.scope)
 			       ? memberCarSyntax.scope
 			       : null);
+                tr.curMethodLambda = saveLambda;
 		if (memberSyntax != null)
 		  tr.setCurrentScope(save_scope);
 		meth = meth.nextSibling;
@@ -477,6 +484,7 @@ public class object extends Syntax
     if (oexp.clinitMethod != null)
       oexp.clinitMethod.outer = oexp;
     tr.pop(oexp);
+    oexp.declareParts(tr);
   }
 
   private static void rewriteInit (Object d, ClassExp oexp, Pair initPair,
@@ -507,18 +515,21 @@ public class object extends Syntax
         oexp.firstChild = initMethod;
       }
     tr.push(initMethod);
+    LambdaExp saveLambda = tr.curMethodLambda;
+    tr.curMethodLambda = initMethod;
     Expression initValue = tr.rewrite_car(initPair, memberSyntax);
     if (d instanceof Declaration)
       {
         Declaration decl = (Declaration) d;
         SetExp sexp = new SetExp(decl, initValue);
-        tr.setLineOf(sexp);
+        sexp.setLocation(decl);
         decl.noteValue(null);
         initValue = sexp;
       }
     else
       initValue = Convert.makeCoercion(initValue, new QuoteExp(Type.void_type));
     ((BeginExp) initMethod.body).add(initValue);
+    tr.curMethodLambda = saveLambda;
     tr.pop(initMethod);
 }
 
